
*
*
* VMU Gong
*
* A game a bit like pong or breakout.
*
* version beta 0.2
*
* by john maushammer
*    maushammer@tidalwave.net
*    www.tidalwave.net/~maushammer/vmu.html
*
* To be done:
*    finish dreamcast icon (currently same as tetris icon)
*    finish title display
*    change way english is put on the ball
*    implement bricks to knock out
*    add multiple levels
*    add special bonus bricks
*    do some sort of scoring
*    add sounds
*
* "Special effects"
*    Paddle width is variable
*    only one ball in play at a time (maybe nice to add extra balls)
*

bounce:
	mov #$a1,ocr
	mov #$09,mcr
	mov #$80,vccr
	clr1 p3int,0
	clr1 p1,7
	mov #$ff,p3

newgame:
	mov #bounce_introscreen&$ff,trl ; show the title screen
	mov #bounce_introscreen>>8,trh
	callf setscr

waitstartgame:	
	callf getkeys
	bp   acc,4,waitstartgame	;wait for A button to be pressed
	
	mov  #5,balls		; 5 balls per game
	mov  #10, paddlex	; place paddle in center
	mov  #$50, bounce_speed	; moderate speed
	
newball:	
	mov  #56,ball_fullx	; initial x position
	mov  #14,ball_x		; must be 1/4 of fullx
	mov  #6,ball_y		; start in the center of the screen
	
	mov  #1,ball_dx
	mov  #1,ball_dy		; moving diagonally (11 degrees, ratio of 1:.25)

*========== test ============
* used for lining up the ball so that it will hit the paddle at a predictable place-
* this tests the hit/miss correctness, that the paddle is drawn the correct size, and
* that the english on the ball is correct.
*	mov  #0,ball_dx		; test...
*	mov  #1,ball_dy		;
*	mov #44,ball_fullx
*	mov #11,ball_x

	mov  #12, paddlewidth	; and make it fairly wide

	mov #playscreen2&$ff,trl ; show the bricks
	mov #playscreen2>>8,trh
	callf setscr

	call drawinitialpaddle
	callf showballs
	
pregameloop1:
	call dopaddle		; allow user to move paddle
	ld   bounce_speed
.delay:	dbnz acc,.delay		; delay loop to temper the VMU's awesome speed
				; (we need this here so that this paddle will be
				;  about as fast as in the real game)
	callf getkeys
	bn   acc,4,pregameloop1	; wait for 'a' to be released

pregameloop2:	
	call dopaddle		; wait for 'a' to be pressed again to start ball
	mov  #$50,acc
.delay:	dbnz acc,.delay		; delay loop to temper the VMU's awesome speed
	callf getkeys
	bp   acc,4,pregameloop2	

	mov #playscreen&$ff,trl ; erase the ball count by redrawing screen
	mov #playscreen>>8,trh
	callf setscr
	call drawinitialpaddle
	
		
bounce_gameloop:
*	.byte $50		; this is a special debug instruction i've added to my
*                               ; copy of Marcus's emulator. It prints out a loop time
*				; and is handy in optimizing code. Email me (john) if you
*				; want a copy of my mods. It would be nice to formalize a
*				; larger set of debug instructions (perhaps two-byte)
	
 	ld   ball_x
	st   B
	ld   ball_y
	callf plotxy		; draw the ball

	push ball_x
	push ball_y		; save ball position now so we can erase it later

	call moveball
	be   #$FF,missed	; ball moved off the bottom
	
	call dopaddle
	ld   bounce_speed
.delay1	dbnz acc,.delay1
	
	call dopaddle
	ld   bounce_speed
.delay2	dbnz acc,.delay2
	
*	call dopaddle		; paddle can move 3 times as fast as the ball's default speed
	;; actually 3x seems a little too fast; commented out to provide 2x
	ld   bounce_speed
.delay3	dbnz acc,.delay3
	
* erase the ball's old position
	pop  acc		; get ball y pos
	pop  B			; get ball x pos
	callf unplotxy
		
	jmp  bounce_gameloop		; continue animating the ball and playing the game.

* ---- handle missed balls ----
		
missed:	pop  acc		; get old position off the stack (we don't erase it...)
	pop  acc
	
	dec  balls		; player lost a ball
	ld   balls
	be   #0,endgame		; game over?
	jmp  newball

endgame: jmpf bounce		; start over	

* ---------------------------------------------------------------------
* moveball - move the ball according to the laws of physics
*
* inputs:	ball_dx
*		ball_dy
*		ball_fullx
* modifies:	ball_y	
* outputs:	ball_x
*
* returns:	acc=$FF if ball was missed
*		acc=ball_y otherwise (not used by anything)

* first calculate x bounce (against side walls)
moveball: ld  ball_dx
	add  ball_fullx
	st   ball_fullx
	ror			; divide by 2 to get ball_x
	ror			; divide by 4 to get ball_x
	and  #$3f		; drop rolled-around bit (divide by 4)
*	and  #$7f		; drop rolled-around bit (divide by 2)
	st   ball_x
	bp   acc,7,.xbnc	; below 0, bounced.
	sub  #48		; greater or equal to 48"
	bp   acc,7,.noxbnc	; no underflow
.xbnc	mov  #0,acc
	sub  ball_dx		; change (dx=-dx)
	st   ball_dx
	br   moveball		; re-calc position within bounds

* now calculate y bounce, if any (against far wall or paddle)
.noxbnc	ld   ball_dy
	add  ball_y
	bp   acc,7,.ybnc	; moved above top edge (y=0)
	inc  acc		; check against screen bottom-1 (so we don't erase paddle)
	bn   acc,5,.noybnc
	ld   ball_x		; moved below bottom edge... must check against paddle
	sub  paddlex		;  acc=ball position relative to left edge of paddle
 	bp   acc,7,.missed	; ball went to left of paddle
	sub  paddlewidth	;  acc=ball position relative to right edge of paddle
 	bn   acc,7,.missed	; ball went to right of paddle
* ball hit paddle... now calculate any spin or english put on the ball (which affects dx)	
* older method- english is determined by if the paddle is moving (actually if a button is
*               pressed.) This was ok, but hard to play.
*	ld   P3			; we hit the ball- calculate if we put english on it
*	bn   acc,2,.lefteng
*	bp   acc,3,.ybnc	; no right english
*.righteng inc ball_dx
*	br   .ybnc		; english applied, now bounce...
*.lefteng bn  acc,3,.ybnc	; if both buttons left & right pressed, then no english
*	dec  ball_dx

* newer method- english is determined by position of ball against paddle- in the center does
*               not affect angle, but near edges exerts spin.
*               algorithm:          left edge = spin - 3
*                          left 1/4 of paddle = spin - 1
*                        middle 1/2 of paddle = no spin
*                         right 1/4 of paddle = spin + 1
*                                  right edge = spin + 3
*
* note: since the paddle is divided into quarters and since the paddle is fairly
*       small, the exact size of the paddle can make a significant difference.
*       A paddle that is 12 pixels wide has 3 spin-creating points on each
*       edge (2 for spin-1 and one for spin-3). A paddle that is 11 pixels wide
*       has (due to rounding down) only 2 spin creating points on each edge
*       (one for spin-1 and one for spin-3). This makes hitting the spin-1 as
*       hard as hitting the spin-3  

	ld	paddlewidth	; get width so we can calculate 1/4 sections
	ror
	ror			; divide by 4
	and	#$3F		; drop rotated bits
	st	B		; save for a bit
	ld	ball_x
	sub	paddlex		;  acc=ball position relative to left edge of paddle
	be	#0,.lleng	; left left english
	sub     B
	bp	acc,7,.leng	; in left 1/4?
* you could do add B and add B, but I'm woried about roundoff errors
	ld	ball_x
	sub	paddlex
	sub     paddlewidth	; - 1  paddlewidth
*	inc     acc
	be	#$FF,.rreng
	add	B		; +1/4 paddlewidth
				; = -3/4 paddlewidth
	bn	acc,7,.reng
.ybnc	mov  #0,acc
	sub  ball_dy		; change (dy=-dy)
	st   ball_dy
	br   .noxbnc		; re-calc position within bounds
.noybnc	dec  acc
	st   ball_y
	ret
.missed mov  #$FF,acc		; indicate paddle missed the ball
	ret

.lleng:	dec ball_dx
	dec ball_dx
.leng:  dec ball_dx
	br  .ybnc
.rreng:	inc ball_dx
	inc ball_Dx	
.reng:	inc ball_dx
	br  .ybnc
			
* ----------------------------------------------------------------------
* dopaddle - handles moving and drawing the paddle.
*
* This function could be accelerated if it uses a special plotxy that
* doesn't have to lookup the y coordinate (since it is always 31)
*
* inputs:	paddlewidth
* modifies:	paddlex	
* calls:	plotxy
*		unplotxy
*		getkeys
*
	
dopaddle:
	callf getkeys		; check for presses
	bn   acc,2,.moveleft
	bp   acc,3,.dpret	; if right button isn't pushed, then return
.moveright:
	ld   paddlex		; make sure we don't go off screen
	add  paddlewidth
	be   #48,.dpret		; already at right edge? then don't move
	ld   paddlex		; get (old) left edge of paddle
	st   B
	mov  #31,acc
	callf unplotxy		; erase this edge
	inc  paddlex		; move the paddle to the right
	ld   paddlex
	add  paddlewidth
	dec  acc		; calc right edge of paddle
	st   B
	mov  #31,acc
	callf plotxy		; plot (new) right edge of paddle
.dpret	ret
.moveleft bn acc,3,.dpret	; if both buttons pressed, then don't move at all
	ld paddlex		; check to make sure we don't go off-screen
	be   #0,.dpret		; it's already at left edge
	add  paddlewidth
	dec  acc		; subtract 1 to find coordinate of right edge
	st   B
	mov  #31,acc		; erase right edge
	callf unplotxy
	dec  paddlex		; move paddle left
	ld   paddlex
	st   B
	mov  #31,acc		; draw new left edge
	callf plotxy
	ret
				
* ---------------------------------------------------------------------------
* draw initial paddle.
*
* this does a complete draw of the paddle. It consists of many points and is
* slow, so it isn't done every time. It's slow because it relies on plotxy -
* if this handled its own graphics, it could be much faster.
*
* The paddle is defined by two variables - paddlex and paddlewidth. When drawn,
* the paddle occupies pixels in columns 'paddlex' to 'paddlex+paddlewidth-1'
*
* inputs:	paddlex
*		paddlewidth
* calls:	plotxy
	
drawinitialpaddle:	
	ld   paddlewidth	; count of number of pixels to draw
	st   C			; loop counter
	
	ld   paddlex		; get (old) left edge of paddle
	st   B
.dip	mov  #31,acc
	callf plotxy		; draw a pixel
	inc  B			; draw next pixel to the right
	dbnz C,.dip		; draw the rest of the pixels
	ret

* ---------------------------------------------------------------------------
* showballs
* 
* this routine shows the number of balls remaining.
*
* input:	balls
* calls:	plotxy
*
showballs:
	ld   balls
	st   C			; loop counter

	mov  #1,B		; start drawing first ball at left
.sb	mov  #28,acc		; draw top left of ball
	call plotxy
	mov  #29,acc		; draw bottom left of ball
	call plotxy
	inc  B
	mov  #28,acc		; draw top right of ball
	call plotxy
	mov  #29,acc		; draw bottom right of ball
	call plotxy
	inc  B
	inc  B
	dbnz C,.sb		; draw all the balls
	ret
	
* ---------------------------------------------------------------------------
* Simple plotxy, using tables to speed up operation
*
* inputs:   B = x coordinate (0-47)
*         acc = y coordinate (0-31)
*
* messes up:	acc, R3, xbnk
* preserves:	B
*	
* Average of about 28 cycles.
*
	
plotxy:	mov  #0,xbnk		; assume on top half of screen
	bn   acc,4,.pltop	; we were right
	mov  #1,xbnk		; we were wrong- learn to live with it.
.pltop	mov  #ylookup&$ff,trl
	mov  #ylookup>>8,trh
	ldc			; lookup
	st   3			; this now points to start of line we need to modify
	ld   B			; get x coord
	ror
	ror
	ror
	and  #$1F		; drop bits that we rolled around to calc bytes into line
	add  3			; add address of beginning of line (stored in register 2)
	st   3			; now point to byte we need to modify
	mov  #xlookup&$ff,trl
*	mov  #xlookup>>8,trh	; not be needed if placed correctly
	ld   B			; 48 possibilities
*	and  #7			; 8 possibilities (if ommited, need larger table)
	ldc			; get bitmask to use
; changed, sketch does it different than original
	bp game, game_sketch, .sketch	
	clr1 ocr,5
	or   @R3		; add to what's already on-screen
	st   @R3		; plot pixel
	set1 ocr,5
    ret
.sketch:
   mov #$81,ocr
   or   @R3    ; add to what's already on-screen
   st   @R3    ; plot pixel
   mov #$a3,ocr
   ret
	      	

unplotxy: mov  #0,xbnk		; assume on top half of screen
	bn   acc,4,.pltop	; we were right
	mov  #1,xbnk		; we were wrong- learn to live with it.
.pltop	mov  #ylookup&$ff,trl
	mov  #ylookup>>8,trh
	ldc			; lookup
	st   3			; this now points to start of line we need to modify
	ld   B			; get x coord
	ror
	ror
	ror
	and  #$1F		; drop bits that we rolled around to calc bytes into line
	add  3			; add address of beginning of line (stored in register 2)
	st   3			; now point to byte we need to modify
	mov  #xlookup2&$ff,trl
	mov  #xlookup2>>8,trh	; not be needed if placed correctly
	ld   B			; 48 possibilities
*	and  #7			; 8 possibilities (if ommited, need larger table)
	ldc			; get bitmask to use
; changed, sketch does it different than original
	bp game, game_sketch, .sketch
	clr1 ocr,5
	and  @R3		; erase what's already on-screen
	st   @R3		; unplot pixel
	set1 ocr,5
   	ret
.sketch:
   mov #$81,ocr
   and  @R3    ; erase what's already on-screen
   st   @R3    ; unplot pixel
   mov #$a3,ocr
   ret
	   	
	
	
*-----
* tables used by plotxy and unplotxy
*-----		
	.cnop	0,$100		; pad to an even number of pages (so that the high byte doesn't need to be reloaded between using tables)
		
ylookup:	.byte $80,$86,$90,$96,$a0,$a6,$b0,$b6,$c0,$c6,$d0,$d6,$e0,$e6,$f0,$f6
		.byte $80,$86,$90,$96,$a0,$a6,$b0,$b6,$c0,$c6,$d0,$d6,$e0,$e6,$f0,$f6

xlookup:	.byte $80,$40,$20,$10,$08,$04,$02,$01
		.byte $80,$40,$20,$10,$08,$04,$02,$01 ; These extra elements used if 
		.byte $80,$40,$20,$10,$08,$04,$02,$01 ; 'and #7' is ommited (faster) 
		.byte $80,$40,$20,$10,$08,$04,$02,$01
		.byte $80,$40,$20,$10,$08,$04,$02,$01
		.byte $80,$40,$20,$10,$08,$04,$02,$01

xlookup2:	.byte $7F,$BF,$DF,$EF,$F7,$FB,$FD,$FE
		.byte $7F,$BF,$DF,$EF,$F7,$FB,$FD,$FE ; These extra elements used if 
		.byte $7F,$BF,$DF,$EF,$F7,$FB,$FD,$FE ; 'and #7' is ommited (faster) 
		.byte $7F,$BF,$DF,$EF,$F7,$FB,$FD,$FE
		.byte $7F,$BF,$DF,$EF,$F7,$FB,$FD,$FE
		.byte $7F,$BF,$DF,$EF,$F7,$FB,$FD,$FE
	


