tetris:
	clr1 ie,7
	mov #$a1,ocr
	mov #$09,mcr
	mov #$80,vccr
	clr1 p3int,0
	clr1 p1,7
	mov #$ff,p3

	clr1 psw,1		; Get random seed from current minute and
	ld $1c			; second system variables
	xor $1d
	set1 psw,1
	st seed
	
	set1 ie,7

	callf clrscr

	mov #<introscreen,trl
	mov #>introscreen,trh
	callf setscr

waitstart:
	callf getkeys
	and #$30		; wenn 48(dez) im ACC(A+B pressed), dann ACC=0
	bnz waitstart

	mov #<mainscreen,trl
	mov #>mainscreen,trh
	callf setscr
	mov #0,acc
	st scorelo
	st scorehi
	call addscore
	call init_hitmap
	mov #100,speed
	mov #0,keyinhib

newpiece:
	mov #5,piece_x
	mov #0,piece_y
	callf random	; bis st: Modulo 7, da es 0-6 pieces gibt
	st c		; Vorbereitung der Division
	mov #0,acc
	mov #7,b
	div
	ld b		; remainder(Rest) ins ACC
	st piece_n
	mov #0,piece_r
	ld speed
	st time
	
	call getpiece
	call checkpiece
	bnz gameover
	call drawpiece

gameloop:
	call kcontrol
	bnz fall
	dbnz time,gameloop
fall:	
	ld speed
	st time
	ld piece_y
	be #18,bottom
	inc piece_y
	call checkpiece
	dec piece_y
	bnz bottom
	call drawpiece
	inc piece_y
	call drawpiece
	br gameloop
bottom:
	mov #4,acc
	call addscore
	call etchpiece
	call checklines
	br newpiece

gameover:	
	mov #<go_sign,trl
	mov #>go_sign,trh
	call show_go
	br waitstart
	
	
kcontrol:
	callf getkeys
	push acc
	or keyinhib
	st gotkeys
	pop acc
	xor #$ff
	and #$30
	st keyinhib

	bp gotkeys,2,noleft
	ld piece_x
	bz noleft

	dec piece_x
	call checkpiece
	inc piece_x
	bnz noleft
	call drawpiece
	dec piece_x
	call drawpiece

noleft:
	bp gotkeys,3,noright
	ld piece_x
	be #10,noright

	inc piece_x
	call checkpiece
	dec piece_x
	bnz noright
	call drawpiece
	inc piece_x
	call drawpiece

noright:
	bp gotkeys,4,norot

	push piece_r
	push piece_i0
	push piece_i1
	ld piece_r
	inc acc
	and #3
	st piece_r
	call getpiece
	call checkpiece
	pop piece_i1
	pop piece_i0
	bnz norot0
	pop acc
	call drawpiece
	call getpiece
	call drawpiece
	push piece_r
norot0:	pop piece_r
norot:	mov #0,acc
	bp gotkeys,1,nodown
	inc acc
nodown:	ret

	
		

	;; Function:	addscore
	;;
	;; Adds a number to player score, and displays current score
	;;
	;; Inputs:
	;;   acc = score increment

addscore:
	push acc
	push b
	push c
	push 2
	push xbnk
	add scorelo	; neue Punkte im ACC dazuaddiert
	st scorelo
	st c
	ld scorehi
	addc #0		; Uebertrag zu scorehi
	st scorehi
	mov #1,xbnk	; select lower half of dotmatrix display
	call drawdig2
	mov #0,xbnk	; select upper half of dotmatrix display
	call drawdig2
	pop xbnk
	pop 2
	pop c
	pop b
	pop acc
	ret


	;; Function:	drawdig2
	;;
	;; Draws two digits for the score indicator
	;;
	;; Inputs:
	;;   acc,c = a 16-bit number.  The two least significant
	;;           digits will be displayed.
	;;   xbnk  = XRAM bank in which to draw the digits
	;; 
	;; Outputs:
	;;   acc,c = The remaining digits (i.e. input number / 100)
	;;
	;; Scratch:
	;;   2 and b will be scratched by this function
	
drawdig2:
	mov #$c5,2
	mov #10,b
	div
	call drawdig
	mov #$85,2
	mov #10,b
	div
	; fallthrough to drawdig...

	
	;; Function:	drawdig
	;;
	;; Draws a single digit for the score indicator
	;;
	;; Inputs:
	;;   b    = digit (0-9)
	;;   2    = position in XRAM where to draw the digit
	;;   xbnk = XRAM bank in which to draw the digit
	;;
	;; Scratch:
	;;   b will be scratched by this function
		
drawdig:
	clr1 ocr,5
	push acc
	push trh
	push trl
	push 2
	mov #<dig_table,trl
	mov #>dig_table,trh
	ld b
	clr1 psw,cy
	rolc
	clr1 psw,cy
	rolc
	clr1 psw,cy
	rolc
	st b
.ddloop:ld b
	ldc 
	st @R2
	inc b
	ld 2
	add #6
	st 2
	ld b
	ldc 
	st @R2
	inc b
	ld 2
	add #10
	st 2
	ld b
	and #7
	bnz .ddloop
	pop 2
	pop trl
	pop trh
	pop acc
	set1 ocr,5
	ret


	;; Function:	getpiece
	;;
	;; Gets a prerotated piece bitmap (4x4 pixels) from ROM
	;;
	;; Inputs:
	;;   piece_n = piece number (0-6)
	;;   piece_r = piece rotation (0-3)
	;;
	;; Outputs:
	;;   piece_i0 = left half bitmap
	;;   piece_i1 = right half bitmap
	
getpiece:			
	push trh		; sichern
	push trl
	mov #<piece_table,trl	
	mov #>piece_table,trh
	ld piece_n	; Adressberechnung
	rol
	rol
	or piece_r
	rol
	push acc		; zwischenspeichern
	ldc		; left half Bitmap loaded
	st piece_i0
	pop acc
	inc acc		; naechste Adresse
	ldc
	st piece_i1	; right half Bitmap loaded
	pop trl
	pop trh
	ret


	;; Function:	drawpiece
	;;
	;; Draws/removes the current piece (XOR mode)
	;;
	;; Inputs:	
	;;   piece_i0 = left half bitmap
	;;   piece_i1 = right half bitmap
	;;   piece_x  = x coordinate (0-10)
	;;   piece_y  = y coordinate (0-18)
	
drawpiece:
	clr1 ocr,5	; enable RC-clock-mode
	push acc
	push b
	push c
	push xbnk
	push 2
	push 3
	push trl
	push trh
	ld piece_i0
	st trl
	ld piece_i1
	st trh
	ld piece_x
	add #17
	st b
	clr1 psw,cy
	rorc
	clr1 psw,cy
	rorc
	clr1 psw,cy
	rorc
	bn piece_y,0,.even
	add #6
.even:	st 2
	ld piece_y
	add #8
	mov #0,xbnk
	bn acc,4,.upper
	inc xbnk
.upper:	and #$0e
	clr1 psw,cy
	rolc
	clr1 psw,cy
	rolc
	clr1 psw,cy
	rolc
	add 2
	or #$80
	st 2
	ld b
	and #7
	st b
	inc b
	mov #1,acc
.getbit:ror
	dbnz b,.getbit
	st b
	mov #4,3
.doall:	mov #4,c
	push 2
	push xbnk
.docol:	ld trh
	rorc
	st trh
	ld trl
	rorc
	st trl
	bn psw,cy,.transp
	ld b
	xor @R2
	st @R2
.transp:ld 2
	and #$f
	be #5,.okeven
	bp psw,cy,.okeven
	ld 2
	add #10
	bn psw,cy,.okodd
	mov #1,xbnk
	or #$80
	br .okodd
.okeven:ld 2
	add #6
.okodd:	st 2
	dbnz c,.docol
	pop xbnk
	pop 2
	ld b
	clr1 psw,cy
	rorc
	st b
	bn psw,cy,.samec
	mov #$80,b
	inc 2
.samec:	dbnz 3,.doall	
	pop trh
	pop trl
	pop 3	
	pop 2
	pop xbnk
	pop c
	pop b
	pop acc
	set1 ocr,5
	ret


	;; Function:	checkpiece
	;;
	;; Checks if the current piece can be placed at the
	;; given location without hitting an obstacle
	;;
	;; Inputs:	
	;;   piece_i0 = left half bitmap
	;;   piece_i1 = right half bitmap
	;;   piece_x  = x coordinate (0-10)
	;;   piece_y  = y coordinate (0-18)
	;;   hitmap   = current obstacle map
	;;
	;; Outputs:
	;;   acc = 1 if obstacle hit, 0 if no obstacle hit
	
checkpiece:
	push b
	push c
	push 1
	push 3
	push trl
	push trh
	ld piece_i0
	st trl
	ld piece_i1
	st trh
	ld piece_y
	clr1 psw,cy
	rolc
	bn piece_x,3,.chleft
	inc acc
.chleft:add #hitmap
	st 1
	ld piece_x
	and #7
	st b
	inc b
	mov #1,acc
.chbit:	ror
	dbnz b,.chbit
	st b
	mov #4,3
.chall:	mov #4,c
	push 1
.chcol:	ld trh
	rorc
	st trh
	ld trl
	rorc
	st trl
	bn psw,cy,.noch
	ld b
	and @R1
	bnz .didhit
.noch:	ld 1
	add #2
	st 1
	dbnz c,.chcol
	pop 1
	ld b
	clr1 psw,cy
	rorc
	st b
	bn psw,cy,.samec
	mov #$80,b
	inc 1
.samec: dbnz 3,.chall
	mov #0,acc
.nohit:	pop trh
	pop trl
	pop 3
	pop 1
	pop c
	pop b
	ret

.didhit:
	pop 1
	mov #1,acc
	br .nohit

	
	;; Function:	etchpiece
	;;
	;; Adds the current piece to the obstacle map
	;;
	;; Inputs:	
	;;   piece_i0 = left half bitmap
	;;   piece_i1 = right half bitmap
	;;   piece_x  = x coordinate (0-10)
	;;   piece_y  = y coordinate (0-18)
	;;   hitmap   = previous obstacle map
	;;
	;; Outputs:
	;;   hitmap = new obstacle map

etchpiece:	
	push b
	push c
	push 1
	push 3
	push trl
	push trh
	ld piece_i0
	st trl
	ld piece_i1
	st trh
	ld piece_y
	clr1 psw,cy
	rolc
	bn piece_x,3,.etleft
	inc acc
.etleft:add #hitmap
	st 1
	ld piece_x
	and #7
	st b
	inc b
	mov #1,acc
.etbit:	ror
	dbnz b,.etbit
	st b
	mov #4,3
.etall:	mov #4,c
	push 1
.etcol:	ld trh
	rorc
	st trh
	ld trl
	rorc
	st trl
	bn psw,cy,.noet
	ld b
	or @R1
	st @R1
.noet:	ld 1
	add #2
	st 1
	dbnz c,.etcol
	pop 1
	ld b
	clr1 psw,cy
	rorc
	st b
	bn psw,cy,.sameet
	mov #$80,b
	inc 1
.sameet:dbnz 3,.etall
	mov #0,acc
	pop trh
	pop trl
	pop 3
	pop 1
	pop c
	pop b
	ret


	;; Function:	checklines
	;;
	;; Removes "full" lines, and increases score and game
	;; speed accordingly
	;;
	;; Inputs:
	;;   hitmap  = previous obstacle map
	;;   scorelo = previous score low byte
	;;   scorehi = previous score high byte
	;;   speed   = previous game speed
	;;
	;; Outputs:
	;;   hitmap  = new obstacle map with lines removed
	;;   scorelo = new score low byte
	;;   scorehi = new score high byte
	;;   speed   = new game speed
	
checklines:
	push acc
	push 0
	push 1
	mov #hitmap+2*19,0
	mov #hitmap+2*19,1
.cklp:	ld @R0
	st @R1
	inc 0
	inc 1
	and @R0
	bne #$ff,.hole
	dec 1
	mov #10,acc
	call addscore
	call incspeed
	br .filld
.hole:	ld @R0
	st @R1
	ld 1
	be 0,.nocpy
	call copyline
.nocpy:	sub #3
	st 1
.filld:	ld 0
	sub #3
	st 0
	bne #hitmap-2,.cklp
	be 1,.ckdne
.cklp2:	mov #$c0,@R1
	inc 1
	mov #$0f,@R1
	ld 1
	call clearline
	sub #3
	st 1
	bne #hitmap-2,.cklp2	
.ckdne:	pop 1
	pop 0
	pop acc
	ret


	;; Function:	incspeed
	;;
	;; Increases game speed, if possibel
	;;
	;; Inputs:
	;;   speed = previous game speed
	;;
	;; Outputs:
	;;   speed = new game speed
	
incspeed:
	push acc
	ld speed
	be #2,.maxsp
	dec acc
	st speed
.maxsp:	pop acc
	ret


	;; Function:	copyline
	;;
	;; Copy one screen line downwards as the pile collapses
	;;
	;; Inputs:
	;;   0   = source line (as pointer into hitmap table)
	;;   acc = destination line (as pointer into hitmap table)
		
copyline:
	clr1 ocr,5
	push 2
	push xbnk
	push trh
	push trl
	push acc
	ld 0
	call calcline
	ld @R2
	and #$1f
	st trl
	inc 2
	ld @R2
	and #$f8
	st trh
	pop acc
	push acc
	call calcline
	ld @R2
	and #$e0
	or trl
	st @R2
	inc 2
	ld @R2
	and #$07
	or trh
	st @R2
	pop acc
	pop trl
	pop trh
	pop xbnk
	pop 2
	set1 ocr,5	
	ret


	;; Function:	clearline
	;;
	;; Clears one screen line as the pile collapses
	;;
	;; Inputs:
	;;   acc = line (as pointer into hitmap table)

clearline:
	clr1 ocr,5
	push acc
	push 2
	push xbnk
	call calcline
	ld @R2
	and #$e0
	st @R2
	inc 2
	ld @R2
	and #$07
	st @R2
	pop xbnk
	pop 2
	pop acc
	set1 ocr,5
	ret


	;; Function:	calcline
	;;
	;; Calculates position in XRAM for a particular screen line
	;;
	;; Inputs:
	;;   acc = line (as pointer into hitmap table)
	;;
	;; Outputs:
	;;   2    = XRAM address
	;;   xbnk = XRAM bank
	
calcline:
	push acc
	mov #$82,2
	sub #hitmap
	clr1 psw,cy
	rorc
	bn acc,0,.cllft
	mov #$88,2
.cllft:	add #8
	mov #0,xbnk
	bn acc,4,.clup
	inc xbnk
.clup:	and #$0e
	clr1 psw,cy
	rolc
	clr1 psw,cy
	rolc
	clr1 psw,cy
	rolc
	add 2
	st 2
	pop acc
	ret


	;; Function:	init_hitmap
	;;
	;; Initializes the obstacle map to only include the
	;; walls and floor
	;;
	;; Outputs:
	;;   hitmap = cleared obstacle map

init_hitmap:
	push 1
	mov #hitmap,1
.ihloop:mov #$c0,@R1
	inc 1
	mov #$0f,@R1
	inc 1
	ld 1
	bne #hitmap+2*22,.ihloop
	mov #$ff,hitmap+2*20
	mov #$ff,hitmap+2*20+1
	pop 1
	ret



	;; Function:	show_go
	;;
	;; Draws a 16x13 pixel sign in the middle of the screen
	;;
	;; Inputs:
	;;   trl = low byte of sign bitmap ROM address
	;;   trh = high byte of sign bitmap ROM address

show_go:
	clr1 ocr,5
	push xbnk
	push 2
	mov #0,xbnk
	mov #$e2,2
	mov #0,c
	ld c
.shlp:	ldc
	st @R2
	inc c
	inc 2
	ld c
	ldc
	st @R2
	inc c
	ld 2
	add #5
	bp c,1,.shodd
	add #4
	bn psw,cy,.shodd
	add #$80
	mov #1,xbnk
.shodd:	st 2
	ld c
	bne #26,.shlp
	pop 2
	pop xbnk
	set1 ocr,5
	ret


	
