; A simple scrolling space fight game.

; $Id: alienfighter.s,v 1.93 2000/12/14 21:13:59 sgust Exp $

; FUTURE:
;  increase defense shot speed even more?
;  should power ups be collectable/ignored/destroyed by shooting them?
;  better life lost handling, show level, lifes, score, skip back?
;  better random number generation
;
; PROBLEMS:
;  power ups: how to start?
;    alien waves?
;    after killing alien with random propability?
;    depending on alienscript in native code?
;  alien waves: which aliens belong to one wave and how to count them
;
; BUGS
;  level line overflow after 91 hours, do batteries last so long ?

; NAMING CONVENTIONS
;  defines:
;    ALN_	alien types
;  data:
;    bm_	bitmaps full screen / level
;    fnt_	font data for aliens
;    str_	ascii strings
;    men_	menus
;    scr_	alien scripts
;    script_	level scripts
;  routines:
;    do_	macros for alienscript cmds
;    cmd_	alienscript subroutines
;    prp_	prep routines
;    key_	keyboard routines
;    snd_	sound player routines
;    lcd_	lcd printing
;    wram_	wram printing
;    screen_	complete screen drawing routines
;    alien_     alienobject handling
;    ship_      defense ship handling
;    score_     score/highscore handling

; POWER
;  titlescreen: 0.7mA
;  frameloop: 1.5-3.0mA, average at about 2.0mA
;  sleep: 0.12mA

; HOWTO implement new aliens:
;  define ALN_
;  make font
;  write prp_ routine
;  insert prp_ into alienpreptable
;  add score into alien_hit.killed

; Copyright (C) 2000 by Soeren Gust, sgust@ithh.infoserv.de

; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.

; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.

; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

; You can get the latest Version from http://soeren.infoserv.de

; parts derived from tetris.s by Marcus Comstedt

; WARNING: Parts of this code are badly written, most of the comments are
; written just for myself and several routines have been changed so often
; that they don't make much sense any more. There are many inconsistencies
; in the naming of variables, routines and label. The information about
; trashed registers in routines is mostly incomplete / inaccurate. On several
; variables some bits are used as markers for special situations. Defining
; DEBUG has some side effects, mostly sanity checks which dump important
; game variables when failed. Look at the routines "panic" for more info.
; On some places you can find short explanations marked with "NOTE:".
; PLEASE DON'T USE THIS AS A TUTORIAL ON HOW TO PROGRAM THE VM.

; WARNING2: The macros need a version of aslc86k newer than 1.9. A patch for
; version 1.9 should be included into the source code distribution.

; --- debugging defines -------------------------------------------------------

;DEBUG		equ	1	; activate for several debugging
;DEBUG_FRAME	equ	1	; show symbol if frametime overflows
;DEBUG_SINGLE	equ	1	; single step trough frames with B

; --- variables in ram --------------------------------------------------------

key_qstart	equ	$04	; first byte of keyboard queue
key_qend	equ	$0e	; last byte of keyboard queue
key_qin		equ	$0f	; pointer to insert into keyboard queue
key_qout	equ	$10	; pointer to read from keyboard queue
key_keys	equ	$11	; key compare mask for keyboard irq
key_lastevent	equ	$12	; store the latest key event for repeat
key_repeat	equ	$13	; counter/flag for key repeat
x		equ	$14	; x position for print
y		equ	$15	; y position for print
width		equ	$16	; width of character for print
height		equ	$17	; height of character for print
char		equ	$18	; used by print
mask1		equ	$19	; used by print
mask2		equ	$1a	; used by print
collision	equ	$1b	; collision flag in print
printmode	equ	$1c	; 0=or, 1=clear
menumax		equ	$1d	; number of items in menu
menunum		equ	$1e	; current item in menu
firerepeat	equ	$1f	; repeat counter for fire
randlow		equ	$20	; bits 0-7 of random seed
randhigh	equ	$21	; bits 8-14 of random seed, highest bit unused
intflag		equ	$22	; bit 0 is set at every 0.5s base time irq
bogomips	equ	$23	; used for RC speed calibration
fps		equ	$24	; how many frames per second to draw
startlevel	equ	$25	; number of first level
level		equ	$26	; current level number
levelbitmaph	equ	$27	; pointer to current level background
levelbitmapl	equ	$28	; pointer to current level background
levellineh	equ	$29	; current level line
levellinel	equ	$2a	; current level line
levelscrl	equ	$2b	; levelscript
levelscrh	equ	$2c	; levelscript
gamestate	equ	$2d	; which state the game is in
gameoption	equ	$2e	; options
lifes		equ	$2f	; how many ships are left
score1		equ	$30	; lowest score byte, BCD
score2		equ	$31	; middle score byte, BCD
score3		equ	$32	; highest score byte, BCD
shipx		equ	$33	; x position of defense ship
shipy		equ	$34	; y position of defense ship
highlander	equ	$35	; undieable counter

snd_playing	equ	$3c	; used by soundengine
snd_wait	equ	$3d	; used by soundengine
snd_qin		equ	$3e	; pointer to insert into sound queue
snd_qout	equ	$3f	; pointer read from sound queue
snd_qstart	equ	$40	; first word of sound queue
snd_qend	equ	$50	; last word of sound queue

; values in gamestate
STATE_PLAY	equ	0	; normal game play
STATE_DEMO	equ	1	; demo mode
STATE_BOSS	equ	2	; level boss handling
STATE_DIEFL	equ	7	; ship is exploding flag
STATE_DIEMASK	equ	$80	; ship is exploding

; bits in gameoption
OPTION_SOUND	equ	0

shots		equ	$50	; start of shot structure
SHOTOFF_X	equ	0	; x position of shot
SHOTOFF_Y	equ	1	; y position of shot
SHOTOFF_ENERGY	equ	2	; bit7: collision flag, rest: energy of shot
SHOTOFF_SIZE	equ	3	; size of structure
shotsend	equ	$59	; end of shot structure

alienobjects	equ	$5e
; struct alienobject
;  byte type
; HINT: only 126 possible at the moment
ALN_SHOT	equ	1	; alien shot
ALN_ASTEROID	equ	2	; rotating asteroid
ALN_TINY	equ	3	; tiny alien
ALN_SMALL	equ	4	; small alien
ALN_SMALL2	equ	5	; another small alien
ALN_SMALL3	equ	6	; another small alien
ALN_DIA1	equ	7	; diagonal fighters
ALN_DIA2	equ	8
ALN_DIA3	equ	9
ALN_DIA4	equ	10
ALN_TENTACLE	equ	11	; alien with side tentacles
ALN_CANNONL	equ	12	; cannon on left side
ALN_CANNONR	equ	13	; cannon on right side
ALN_CANNONU	equ	14	; cannon on upper border
ALN_CANNONDL	equ	15	; cannon diagonal left
ALN_CANNONDR	equ	16	; cannon diagonal right
ALN_CANNONDUL	equ	17	; cannon diagonal left mounted on top side
ALN_CANNONDUR	equ	18	; cannon diagonal right mounted on top side
ALN_BOMBERL	equ	19	; bomb thrower on left side
ALN_BOMBERLF	equ	20	; bomb thrower on left side, fireing
ALN_BOMBERR	equ	21	; bomb thrower on right side
ALN_BOMBERRF	equ	22	; bomb thrower on right side, fireing
ALN_BOMB	equ	23	; bomb from bomb thrower
ALN_BOMB2	equ	24	; bomb from tentacle alien
ALN_GRID1	equ	25	; shootable grid
ALN_MOTHER	equ	26	; mother ship
ALN_BOSS1B	equ	27	; level 1 boss back
ALN_BOSS1L	equ	28	; level 1 boss left
ALN_BOSS1R	equ	29	; level 1 boss right
ALN_BOSS1M	equ	30	; level 1 boss middle
ALN_BOSS1F	equ	31	; level 1 boss front
ALN_SMILE	equ	32	; smiling alien
ALN_SINGLE	equ	33	; single shot alien
ALN_DOUBLE	equ	34	; double shot alien
ALN_BLOCK	equ	35	; block
ALN_EXPLOSION1	equ	36	; explosion

; rest of struct alienobject
ALNOFF_FRAME	equ	1	; for animated aliens, incremented every frame
				; bit7: do not draw alien marker, used for hits
ALNOFF_X	equ	2	; current x position, bit7: nomove marker (fork)
ALNOFF_Y	equ	3	; current y position
ALNOFF_PCL	equ	4	; pointer to alienscript low
ALNOFF_PCH	equ	5	; pointer to alienscript high
ALNOFF_WAIT	equ	6	; wait counter for alienscript
ALNOFF_V	equ	7	; speed low/high nibble is x/y
ALNOFF_HITS	equ	8	; how many hits are necessary to destroy alien
ALNOFF_SIZE	equ	9	; size of struct alienobject
alienend	equ	$00	; pointer to (end of last alienobject) + 1

; struct levelscript
LVLOFF_LINEL	equ	0	; lowbyte of startline
LVLOFF_LINEH	equ	1	; highbyte of startline
LVLOFF_TYPE	equ	2	; type of new alien
LVLOFF_X	equ	3	; x offset of new alien
LVLOFF_Y	equ	4	; y offset of new alien
LVLOFF_HITS	equ	5	; hit value of new alien
LVLOFF_SCRL	equ	6	; script pointer low
LVLOFF_SCRH	equ	7	; script pointer high
LVLOFF_SIZE	equ	8	; size of structure

; key numbers
KEY_UP		equ	1
KEY_DOWN	equ	2
KEY_LEFT	equ	3
KEY_RIGHT	equ	4
KEY_A		equ	5
KEY_B		equ	6
KEY_MODE	equ	7
KEY_SLEEP	equ	8
KEY_REL		equ	$80	; flag for key released

; --- some useful parameters --------------------------------------------------

DEMOFPS		equ	20	; frames/second in demo mode >=6!!
STARTFPS	equ	12	; frames/second in inner game loop >=6!!
NEXTFPS		equ	2	; how many fps to add if all level completed
MAXFPS		equ	20	; will not get faster that this
LIFECOUNT	equ	5	; number of lifes per game

; --- useful macros ----------------------------------------------------------
; WARNING: These need aslc86k >1.9,  a patch for 1.9 should be included in
; the distribution.

; print a fixed string from rom
mprstr	.macro		; (string,x,y)
	mov	#<\1,trl
	mov	#>\1,trh
	mov	#\2,x
	mov	#\3,y
	callf	lcd_printstring
	.endm

; switch the clock mode, corrects t0prr
clkmode	.macro
	.ifc	"\1","rc"
	ld	bogomips
	st	t0prr
	clr1	ocr,subclk		; use rc osc.
	.endc
	.ifnc	"\1","rc"
	.ifc	"\1","sub"
	set1	ocr,subclk		; back to sub clock
	mov	#$fc,t0prr		; adjust timer speed
	.endc
	.ifnc	"\1","sub"
	error	"unknown clock mode"	; error is an unknown opcode
	.endc
	.endc
	.endm

; add 1st and 2nd parameter, store in acc and 3rd
add_st	.macro
	ld	\1
	add	\2
	st	\3
	.endm

; sub 2nd from 1st parameter, store in acc and 3rd
sub_st	.macro
	ld	\1
	sub	\2
	st	\3
	.endm

; add 8 bit value to 16 bit register pair
add16	.macro
	ld	\2l
	add	\1
	st	\2l
	bn	psw,cy,.enda\@
	inc	\2h
.enda\@
	.endm

; subtract 8 bit value from 16 bit register pair
sub16	.macro
	ld	\2l
	sub	\1
	st	\2l
	bn	psw,cy,.ends\@
	dec	\2h
.ends\@
	.endm


; --- macros for the alienscript commands -------------------------------------
; NOTE:
; History of this macros:
; Originally the alienscripts were done using an interpreted macro language.
; To make the commands more readable, I used macros like the levelscript
; macro. The current macros are the functional equivalents to the original
; macros.

; move alien with fixed speed for some time
do_wait	.macro
	add_st	0,#ALNOFF_WAIT,1
	mov	#\1,@r1
.wait\@	callf	cmd_automove
	callf	alien_endscript
	add_st	0,#ALNOFF_WAIT,1
	dbnz	@r1,.wait\@
	.endm

; move alien with fixed speed for some time, execute another routine
do_run	.macro
	add_st	0,#ALNOFF_WAIT,1
	mov	#\1,@r1
.run\@	callf	cmd_automove
	callf	\2
	callf	alien_endscript
	add_st	0,#ALNOFF_WAIT,1
	dbnz	@r1,.run\@
	.endm

; set speed of alien
do_spd	.macro
	add_st	0,#ALNOFF_V,1
	mov	#\1,@r1
	.endm

; set type of alien
do_type	.macro
	mov	#\1,@r0
	.endm

; start new alien
do_fork	.macro
	mov	#\1,char	; type
	mov	#\2,x   	; x
	mov	#\3,y   	; y
	mov	#\4,width	; hits
	mov	#<\5,mask1	; alienscript
	mov	#>\5,mask2	; alienscript
	callf	cmd_fork
	.endm

; remove alien
do_remove
	.macro
	mov	#0,@r0
	callf	alien_endscript
	.endm

; --- macro for the levelscript commands --------------------------------------

levelalien	.macro
		.word	\1		; start line
		.byte	\2		; type of alien
		.byte	\3,\4		; x/y position
		.byte	\5		; hits
		.word	\6		; alienscript
		.endm

; --- reset and interrupt vectors ---------------------------------------------

	.include "sfr.i"

; reset
	.org	0
	jmpf	start

; external irq 0
	.org	$3
	jmp	nop_irq

; external irq 1
	.org	$b
	jmp	nop_irq
	
; timer/counter 0 low 8 bit mode overflow: frame timer
	.org	$13
	jmpf	framesync_irq

; count clock
	.org	$1b
	jmp	clockint

; timer/counter 0 high / 16 bit mode overflow: keyboard
	.org	$23
	jmpf	key_irq

; timer/counter 1 overflow: sound
	.org	$2b
	jmpf	snd_irq

; SIO0
	.org	$33
	jmp	nop_irq

; SIO1
	.org	$3b
	jmp	nop_irq

; RFB
	.org	$43
	jmp	nop_irq

; port3 change: key pressed
	.org	$4b
	clr1	p3int,p3iflag		; clear flag
	clr1	p3int,p3ie		; disable further interrupts
	reti

nop_irq
	reti

; --- firmware calls ----------------------------------------------------------

	.org	$100
writeflash
	not1	ext,0
	jmpf	writeflash
	ret

	.org	$110
verifyflash
	not1	ext,0
	jmpf	verifyflash
	ret

	.org	$120
readflash
	not1	ext,0
	jmpf	readflash
	ret

; Clock irq: Count time, do housekeeping
	.org	$130
clockint
	push	ie
	clr1	ie,mastie
	not1	ext,0
	jmpf	clockint
	pop	ie
	callf	myclockint
	reti
		
; Leave game
	.org	$1f0
goodbye	
	not1	ext,0
	jmpf	goodbye

; --- header ------------------------------------------------------------------

	.org	$200
; Name
	.byte	"Alien   Fighter "
; Description
	.byte	"Alien Fighter 1.0 by Soeren Gust"
; Creator application
	.byte	"aslc86k         "
; Icon header: number of frames
	.word	1
; Icon header: speed
	.word	0
; Icon header: eyecatch
	.word	0
; Icon header: crc, unused in game
	.word	0
; Length of file data
	.word	end,0
; Reserved
	.byte	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0

	.include "icon.i"
	
; --- main program ------------------------------------------------------------

start
	clr1	ie,mastie		; disable all irq
	mov	#$a1,ocr		; 32KHz, RC running
	mov	#$09,mcr		; LCD on, no cursor, gfx mode
	mov	#$80,vccr		; LCD on, enable writing
	clr1	p3int,p3ie		; disable P3 irq
	clr1	p1,7			; ???
	mov	#$ff,p3			; enable pull-ups on P3 (buttons)

	; initialise random seed
	clr1	psw,rambk0		; system variable bank
	ld	$1d			; current time: seconds
	set1	psw,rambk0		; switch back to program variable bank
	st	randlow		; put seconds into low byte of random seed
	xor	#$ff		; invert it to make sure at least 2 bits in 
	st	randhigh	; the seed are different and put into high byte

; NOTE:
; init timer0 low interrupt frequency: 32768/6/4/136 = 10.04 per second
; Used as frame timer, t0lr gets adjusted to different speeds.
; A calibration value for t0prr in rc clk will be put into bogomips
	mov	#0,t0con		; turn off everything
	mov	#$fc,t0prr		; prescaler to / 4
	mov	#$78,t0lr		; timer0 low to / 136

; NOTE:
; timer 0 high interrupt frequency: 32768/6/4/64 = 21.34 per second
; Used for keyboard, does not work good enough at 10 per second, so
; can't be done by t0 low.
	mov	#$c0,t0hr		; timer0 high to / 64

	callf	key_init		; initialise keyboard
	callf	snd_init		; initialise sound

	; initialise rest
	mov	#0,printmode
	mov	#0,gameoption
	mov	#1,startlevel
	set1	gameoption,OPTION_SOUND	; sound on by default
	set1	ie,mastie		; enable all irq

	; calibrate RC clock speed
	callf	screen_calibrate
	callf	calibrate
	.ifd	DEBUG
	mov	#0,x
	mov	#24,y
	clkmode	rc			; draw in rc clk
	callf	lcd_printhex
	clkmode	sub			; back to subclk
	callf	key_waita
	.endc

mainloop
	clkmode	rc			; draw in rc clk
	mov	#<bm_title,trl
	mov	#>bm_title,trh
	callf	lcd_fromrom
	clkmode	sub			; back to subclk
	mov	#$ff,acc		; 25 seconds timeout
	callf	key_timeab
	bz	.mainmenu
	jmpf	goodbye			; timeout, quit game
.mainmenu
	; handle the main menu
	clkmode	rc
	callf	lcd_clear
	clkmode	sub
	mov	#<menu_main,trl
	mov	#>menu_main,trh
	.ifd	DEBUG
	mov	#5,acc			; handle DEBUG item
	.endc
	.ifnd	DEBUG
	mov	#4,acc			; no DEBUG mode, no DEBUG item
	.endc
	mov	#$ff,b			; wait 25 seconds
	callf	domenu
	be	#1,.dogame		; start game
	be	#2,.optmen		; options
	be	#3,.dohigh		; high score
	be	#4,.dodemo		; demo mode
	.ifd	DEBUG
	bne	#5,mainloop
	jmp	.dodbg			; DEBUG menu
	.endc
	br	mainloop		; menu aborted

.dogame	; start game selected
	callf	playgame		; play the game
	br	mainloop

.optmen	; options selected, handle options menu
	clkmode	rc
	callf	lcd_clear
	clkmode	sub
	mov	#<menu_options,trl
	mov	#>menu_options,trh
	mov	#1,acc
	mov	#$ff,b			; wait 25 seconds
	callf	domenu
	be	#1,.dosnd
	br	.mainmenu

.dosnd	; toggle sound on/off
	not1	gameoption,OPTION_SOUND
	clkmode	rc
	callf	lcd_clear
	mov	#0,x
	mov	#0,y
	mov	#<str_sound,trl
	mov	#>str_sound,trh
	callf	lcd_printstring
	mov	#0,x
	mov	#6,y
	bn	gameoption,OPTION_SOUND,.sndoff
	mov	#<str_soundon,trl
	mov	#>str_soundon,trh
	br	.sndpr
.sndoff mov	#<str_soundoff,trl
	mov	#>str_soundoff,trh
.sndpr	callf	lcd_printstring
	clkmode	sub
	mov	#10,acc
	callf	key_timea
	jmp	.mainmenu

.dohigh	; show highscores
	callf	screen_highscore
	mov	#$ff,acc		; wait max. 25 seconds
	callf	key_timea
	jmp	.mainmenu

.dodemo	; which level to demo
	clkmode	rc
	callf	lcd_clear
	clkmode	sub
	mov	#<menu_demo,trl
	mov	#>menu_demo,trh
	mov	#4,acc
	mov	#$ff,b			; wait max. 25 seconds
	callf	domenu
	be	#0,.back
	st	level
	mov	#STATE_DEMO,gamestate
	mov	#DEMOFPS,fps
	call	playlevel		; demo the level
.back	jmp	.mainmenu

	.ifd	DEBUG
.dodbg	; handle debug menu
	clkmode	rc
	callf	lcd_clear
	clkmode	sub
	mov	#<menu_debug,trl
	mov	#>menu_debug,trh
	mov	#1,acc
	mov	#$ff,b			; wait 25 seconds
	callf	domenu
	be	#1,.nextlevel
	jmp	.mainmenu
.nextlevel
	inc	startlevel
	jmp	.mainmenu
	.endc

; play the game
playgame
	; init game variables
	mov	#LIFECOUNT,lifes	; set lifes
	mov	#0,highlander		; we can die
	ld	startlevel
	st	level			; start with first level
	mov	#$00,score1		; score = 0
	mov	#$00,score2
	mov	#$00,score3
	mov	#STARTFPS,fps
.gameloop
	callf	screen_level		; show level, lifes, score
	; play sound: start level
	mov	#<tune_level,trl
	mov	#>tune_level,trh
	callf	snd_playtune
.snwai1	ld	snd_playing
	bnz	.snwai1
	mov	#20,acc
	callf	key_timea		; 2 seconds or A
	mov	#STATE_PLAY,gamestate	; really playing, no demo
	call	playlevel		; play the level
	bnz	.over			; check for gameover
	ld	lifes
	be	#9,.skiplife
	inc	lifes
.skiplife
	inc	level			; next level
	ld	level
	be	#5,.maxlvl		; check for max level
	br	.gameloop

.maxlvl	; max level reached
	mov	#1,level		; start with 1st level
	add_st	fps,#NEXTFPS,fps	; speed up
	sub	#MAXFPS
	bp	psw,cy,.gameloop	; max speed not yet reached
	mov	#MAXFPS,fps		; max speed reached
	br	.gameloop

.over	; game over
	callf	score_testhigh		; test for new highscore
	bnz	.high			; new highscore!
	callf	screen_gameover
	; play sound: game over
	mov	#<tune_over,trl
	mov	#>tune_over,trh
	callf	snd_playtune
.snwai2	ld	snd_playing
	bnz	.snwai2
	mov	#255,acc
	callf	key_timea
	ret

.high	; new highscore, enter into table
	push	acc
	callf	screen_newhigh1		; display new highscore screen
	; play sound: new highscore
	mov	#<tune_highscore,trl
	mov	#>tune_highscore,trh
	callf	snd_playtune
.snwait	ld	snd_playing
	bnz	.snwait
	callf	screen_newhigh2		; display enter name screen
	pop	acc
	callf	score_newhigh		; enter new name and write to flash
	callf	screen_highscore	; display new highscore table
	mov	#100,acc
	callf	key_timea
	ret

; play one level
; returns 0 on successfull completion
playlevel
	; clear alienobjects
	mov	#alienobjects,0
.loop1	mov	#0,@r0
	add_st	0,#ALNOFF_SIZE,0
	bne	#alienend,.loop1

	mov	#$00,levellinel		; init levelline
	mov	#$00,levellineh
	mov	#22,shipx		; init ship position
	mov	#37,shipy
	mov	#0,highlander		; clear highlander flag
	mov	#1,firerepeat		; init repeat counter for fire button

	; init shots
	mov	#shots,0
.loop2	mov	#$ff,@r0
	add_st	0,#SHOTOFF_SIZE,0
	bne	#shotsend,.loop2

	; level specific init
	ld	level
	be	#1,.level0
	be	#2,.level1
	be	#3,.level2
	be	#4,.level3
	.ifd	DEBUG
	mov	#$20,acc
	jmpf	panic
	.endc
	; unknown level: use level 0
.level0	; init levelscroll
	mov	#<(bm_level0+6*(1024-40)),levelbitmapl
	mov	#>(bm_level0+6*(1024-40)),levelbitmaph
	; init levelscript
	mov	#<script_level0,levelscrl
	mov	#>script_level0,levelscrh
	br	.afterlevel
.level1	; init levelscroll
	mov	#<(bm_level1+6*(1024-40)),levelbitmapl
	mov	#>(bm_level1+6*(1024-40)),levelbitmaph
	; init levelscript
	mov	#<script_level1,levelscrl
	mov	#>script_level1,levelscrh
	br	.afterlevel
.level2	; init levelscroll
	mov	#<(bm_level2+6*(1024-40)),levelbitmapl
	mov	#>(bm_level2+6*(1024-40)),levelbitmaph
	; init levelscript
	mov	#<script_level2,levelscrl
	mov	#>script_level2,levelscrh
	br	.afterlevel
.level3	; init levelscroll
	mov	#<(bm_level3+6*(1024-40)),levelbitmapl
	mov	#>(bm_level3+6*(1024-40)),levelbitmaph
	; init levelscript
	mov	#<script_level3,levelscrl
	mov	#>script_level3,levelscrh
	br	.afterlevel

.afterlevel
	clr1	t0con,t0hie		; disable keyboard interrupt
	clr1	t0con,t0hrun		; disable keyboard timer
	callf	snd_init		; abort all sound playing and reinit
	clkmode	rc			; draw in rc clk
	; set frame timer to match fps value
	ld	fps
	st	b
	mov	#$05,acc
	mov	#$50,c
	div				; 1360/fps
	xor	acc
	sub	c
	st	t0lr			; 256-1360/fps

frameloop
; NOTE:
; This is the main game playing frameloop. It runs in RC clock mode. Frame
; synchronisation is done with timer0l. Levellineh/l are used as a counter how
; many frames have passed since the level was started. To simplify collision
; handling the game objects are drawn in a special order, first everything that
; kills the player. Then the ship of the player testing for pixels already set
; while drawing. Next things are the defense shots, again testing for pixels
; set. In the next step the game tries to identify which aliens collided with
; the shots. As a last step all the non shootable not deadly objects (currently
; the explosions) are drawn. Up to this point all the drawing was done into
; wram with 48x40 pixel to allow soft scrolling in of bigger aliens. Now the
; lower 32 lines of wram are copied into the LCD and the frame is visible.
;
; After that all the movements of ship, shots, aliens and level are done.
; If it is not yet time for a new frame the processor is put into HALT mode
; to conserve some battery power. If DEBUG_FRAME is set and the time for a new
; frame has already come, a small star is printed into the top left corner of
; the LCD to show that the processor is too slow to keep up the full frame
; rate.

	; start frame timer timer
	clr1	t0con,t0lrun		; stop frame timer
	clr1	t0con,t0lovf		; clear overflow
	set1	t0con,t0lie		; enable interrupt to wake up framesync
	set1	t0con,t0lrun		; start frame timer

	callf	snd_play		; handle sound

	; display background
	ld	levelbitmapl
	st	trl
	ld	levelbitmaph
	st	trh
	callf	wram_fromrom		; copy background into wram

	; display unexploded aliens
	mov	#alienobjects,0
.aloop1	ld	@r0
	bz	.askip1			; unused alienobject
	add_st	0,#ALNOFF_HITS,1
	ld	@r1
	bz	.askip1			; this one is currently exploding
	call	alien_display		; draw current alien
.askip1	add_st	0,#ALNOFF_SIZE,0	; advance to next alien
	bne	#alienend,.aloop1

	callf	ship_draw		 ;draw defense ship / handle life lost
	call	shots_draw		; draw defense shots
	call	alien_kill		; collision checks alienobjects + shots

	; draw exploding aliens
	mov	#alienobjects,0
.aloop2	ld	@r0
	bz	.askip2			; empty alienobject
	add_st	0,#ALNOFF_HITS,1
	ld	@r1
	bnz	.askip2			; this one is not exploding
	call	alien_display		; draw current alien
.askip2	add_st	0,#ALNOFF_SIZE,0	; advance to next alien
	bne	#alienend,.aloop2

	; FUTURE: collision checks for power ups + defense ship
	; FUTURE: draw powerups (static 1)

	callf	wram_tolcd			; copy current screen to lcd

	; book-keeping
	ld	gamestate
	clr1	acc,STATE_DIEFL			; ignore dying
	be	#STATE_PLAY,.fwlvl
	be	#STATE_DEMO,.fwlvl
	be	#STATE_BOSS,.fixed		; don't scroll in boss mode
	; unknown gamestate, panic
	mov	#$01,acc
	jmpf	panic

.fwlvl	sub16	#6,levelbitmap			; advance scroll pointer
.fixed	add16	#1,levelline			; advance level line
	callf	alien_exec			; execute alienscripts
	call	shots_move			; move defense shots
	call	ship_move			; move defense ship / keyboard
	call	levelscript			; execute levelscript
	; test for end of level
	ld	levellinel
	sub	#225
	bnz	.skipstartboss
	ld	levellineh
	sub	#3
	bnz	.skipstartboss
	; level end reached: start boss mode / end demo
	ld	gamestate
	clr1	acc,STATE_DIEFL
	bne	#STATE_DEMO,.toboss
	; demo finished
	callf	snd_init		; abort all sound playing and reinit
	clr1	t0con,t0lrun		; stop frame timer
	clkmode	sub			; back to subclk
	callf	key_init		; reinit keyboard routines
	clkmode	rc
	mprstr	str_pressa,3,12
	clkmode	sub
	mov	#100,acc
	callf	key_timea		; wait for 10 seconds or key A
	mov	#$ff,acc		; game over mark
	ret

.toboss ; set state to boss mode
	ld	gamestate
	and	#STATE_DIEMASK		; keep DIEMASK
	or	#STATE_BOSS		; set to end game mode
	st	gamestate
	add16	#6,levelbitmap		; move background back one line

.skipstartboss
	; test for bossmode
	ld	gamestate
	clr1	acc,STATE_DIEFL
	bne	#STATE_BOSS,.framesync

	; in boss mode: check if all aliens are dead
	mov	#alienobjects,0
.aloop3	ld	@r0
	bnz	.framesync		; this one is still alive
	add_st	0,#ALNOFF_SIZE,0
	bne	#alienend,.aloop3

	; all aliens dead, level completed successfully
	callf	snd_init		; abort all sound playing and reinit
	clr1	t0con,t0lrun		; stop frame timer
	clkmode	sub			; back to subclk
	callf	key_init		; reinit keyboard routines
	mov	#$00,acc		; next level, please
	ret

.framesync
	.ifd	DEBUG_SINGLE
.wait1	bp	p3,5,.wait1		; wait for B down
.wait2	bn	p3,5,.wait2		; wait for B up
	.endc
	.ifd	DEBUG_FRAME
	bp	t0con,t0lovf,.overrun	; frametime already reached?
	.endc
.framesync2
	bp	t0con,t0lovf,.synced	; frametime already reached?
	set1	pcon,halt		; sleep till interrupt
	br	.framesync2		; wait more
.synced jmp	frameloop
	.ifd	DEBUG_FRAME
.overrun
	; frametime already reached, processor too slow to keep full frame rate
	; draw symbol
	mov	#0,xbnk
	mov	#%01010100,$180
	mov	#%00111000,$186
	mov	#%01010100,$190
	jmp	frameloop
	.endc

; --- timing routines ---------------------------------------------------------

; this is the interrupt which wakes us while in HALT mode in .framesync
framesync_irq
	bp	t0con,t0lovf,.clearie
	reti
.clearie
	clr1	t0con,t0lie		; disable interrupt
	reti

; just set a flag to indicate interrupt
myclockint
	set1	intflag,0
	ret

; calibrate the timing, assuming clock between approx. 8KHz and 2MHz
; NOTE:
; This relies on clk/6/4/170/bogomips/2 = clk/6/bogomips/136/10
; Explanation of numbers in first term:
;  6 builtin division of clock frequency
;  4 t0prr in calibration
;  170 t0lr in calibration
;  bogomips result of calibration (74 for 600KHz)
;  2 the base timer runs at 2Hz
; Explanation of additional numbers in second term:
;  136 t0lr value for 10fps
;  10 10 frames per second
;
; If you set t0prr to bogomips in RC clock mode and t0lr to 136 the timer runs
; at 10 Hz. If you set t0prr to $fc (/4) and t0lr to 136 in sub clock mode you
; get 10 Hz, too. So to get a framerate of fps frames per second t0lr has to be
; set to 256-1360/fps. In the game playing frameloop only 1 frame per timer
; overflow is drawn. This speed compensation for timer0 is not 100% accurate,
; switching back to sub clock mode in the frameloop when finished drawing gives
; a visible unstability to the timing.
calibrate
	clr1	ocr,subclk		; use rc osc.
	mov	#$fc,t0prr		; prescaler to / 4
	mov	#$56,t0lr		; timer0 low to / 170
	clr1	intflag,0
.loop1	bn	intflag,0,.loop1	; wait for next 0.5s tick
	clr1	intflag,0
	mov	#0,c			; start with 0
	set1	t0con,t0lrun
.loop2	clr1	t0con,t0lovf
.loop3	bn	t0con,t0lovf,.loop3	; count timer 0 low ticks
	inc	c
	bn	intflag,0,.loop2	; stop after 0.5s
	clr1	t0con,t0lrun
	set1	ocr,subclk		; use sub clk
	mov	#$78,t0lr		; timer0 low to / 136
	xor	acc			; t0prr is an incrementing counter
	sub	c			; so negate value
	st	bogomips		; store calibration value
	ret

; --- score handling code starts here -----------------------------------------

; prints current score at position x/y in font from trh/trl size widthxheight,
; advances cursor
lcd_printscore
	ld	score3
	call	.printscorebyte
	ld	score2
	call	.printscorebyte
	ld	score1
	call	.printscorebyte
	ret

; prints BCD byte in a at position x/y in 5x5 font, advances cursor
.printscorebyte
	push	acc
	and	#$f0
	ror
	ror
	ror
	ror
	call	.printscorenibble
	pop	acc
	and	#$0f
	call	.printscorenibble
	ret

; prints BCD nibble in acc at position x/y in 5x5 font, advances cursor
.printscorenibble
	push	y
	callf	lcd_print
	pop	y
	add_st	x,width,x
	ret

; adds the BCD byte in acc to the current score
score_add
	add	score1
	call	.adjustbcd
	st	score1
	mov	#0,acc
	addc	score2
	call	.adjustbcd
	st	score2
	mov	#0,acc
	addc	score3
	call	.adjustbcd
	st	score3
	bp	psw,cy,.scoreover	; test for overflow
	ret
.scoreover
	; overflow of highest byte, set score to maximum
	mov	#$99,score1
	mov	#$99,score2
	mov	#$99,score3
	ret

; adjust the value in acc/ac/cy so, that the last addition was a bcd one
; there must be a much simpler solution than this one, but i can't find it
; now i know why Z80 emulator engines use a table for DAA...
.adjustbcd
	mov	#0,c			; clear carry marker
	bn	psw,ac,.adj1
	; ac overflow, lower nibble in [0,2]
	push	psw
	add	#$06
	pop	psw
.adj1	bn	psw,cy,.adj2
	; cy overflow, higher nibble in [0,2]
	add	#$60
	inc	c			; set carry marker
.adj2	push	acc
	and	#$f0
	sub	#160
	pop	acc
	bp	psw,cy,.adj3
	; higher nibble is in [$a,$f]
	add	#$60
	bn	psw,cy,.adj3
	inc	c			; set carry marker
.adj3	push	acc
	and	#$0f
	sub	#10
	pop	acc
	bp	psw,cy,.adj4
	; lower nibble is in [$a,$f]
	add	#$06
	bn	psw,cy,.adj4
	inc	c			; set carry marker
.adj4	; test again for high nibble == $a
	push	acc
	and	#$f0
	sub	#$a0
	bnz	.adj5
	pop	acc
	add	#$60
	push	acc
	bn	psw,cy,.adj5
	inc	c			; set carry marker
.adj5	pop	acc
	bn	c,0,.nocy		; test for carry
	set1	psw,cy
	ret
.nocy	clr1	psw,cy
	ret

; test for a new highscore, returns pointer to entry in ram or 0
score_testhigh
	; copy flash rom into ram
	push	ie
	clr1	ie,mastie		; disable interrupts
	clkmode	rc			; flash access needs rc clock
	mov	#0,$7d
	mov	#>highscores,$7e
	mov	#<highscores,$7f
	call	readflash
	clkmode	sub
	pop	ie

	; now test the entries
	mov	#$80,0
.loop	; test highest byte
	ld	0
	st	1
	ld	@r1
	sub	score3
	bp	psw,cy,.high		; highest byte of new score is bigger
	bnz	.next			; highest byte of old score is bigger
	; highest byte identical, test middle byte
	inc	1
	ld	@r1
	sub	score2
	bp	psw,cy,.high		; middle byte of new score is bigger
	bnz	.next			; middle byte of old score is bigger
	; middle byte identical, test lowest byte
	inc	1
	ld	@r1
	sub	score1
	bp	psw,cy,.high		; lowest byte of new score is bigger
.next	add_st	0,#6,0
	bne	#$80+30,.loop
	mov	#0,acc			; no new score
	ret
.high	ld	0			; return pointer to new score
	ret

; enter new highscore, acc: pointer to score entry in ram
; trashes acc, c, 0, 1, $7c-$ff
score_newhigh
	st	c
	; shift other entries
	mov	#$80+29,0		; end of last entry
	mov	#$80+23,1		; end of second last entry
	mov	#$80+30,acc
	sub	c			; difference between new entry and end
	sub	#6			; -1 entry = number of entries to copy
	bz	.insert			; last entry, do not copy anything
	st	b
.copylp	ld	@r1
	st	@r0
	dec	1
	dec	0
	dbnz	b,.copylp

.insert	; insert new score into table
	ld	c
	st	0
	ld	score3
	st	@r0
	ld	score2
	inc	0
	st	@r0
	ld	score1
	inc	0
	st	@r0
	; enter name
	inc	0
	mov	#'A',@r0
	ld	0
	st	1
	inc	1
	mov	#'A',@r1
	callf	key_clear		; clear keyboard queue
	push	key_repeat
	set1	key_repeat,7		; activate repeat
.firstname
	ld	@r0
	mov	#30,x
	mov	#24,y
	callf	score_enterchar
	bz	.firstname
	st	@r0
.secondname
	ld	@r1
	mov	#36,x
	mov	#24,y
	callf	score_enterchar
	bz	.firstname
	st	@r1
.keylp	callf	key_get
	be	#KEY_A,.cont
	be	#KEY_B,.secondname
	br	.keylp
.cont	pop	key_repeat
	; write back to flash
	mov	#1,$7c			; wait to stabilize
	mov	#0,$7d			; highest nibble: game bank
	mov	#>highscores,$7e
	mov	#<highscores,$7f
	push	ie			; save registers
	clr1	ie,mastie		; disable interrupts
	clkmode	rc			; flash access needs rc clock
	mov	#0,acc			; softvms <= 1.8 only
	call	writeflash		; write to flash
	clkmode	sub
	pop	ie
	bnz	.flasherror		; test for error
	push	ie			; save registers
	clr1	ie,mastie		; disable interrupts
	clkmode	rc			; flash access needs rc clock
	call	verifyflash		; verify
	clkmode	sub
	pop	ie
	bnz	.flasherror		; test for error
	ret
.flasherror
	mov	#$30,acc
	jmpf	panic

; enter one char at position (x,y)
; choose char with D pad, end with A, back with B
; returns: acc new char / 0 (B pressed)
; trashes: menunum, width, height, trh/trl
score_enterchar
	st	menunum
	; print cursor
	clkmode	rc
	mov	#6,width
	mov	#1,height
	mov	#<font5x1,trl
	mov	#>font5x1,trh
	push	y
	add_st	y,#6,y
	xor	acc
	callf	lcd_print
	pop	y
	clkmode	sub
	mov	#6,width
	mov	#5,height
.print	mov	#<font5x5,trl
	mov	#>font5x5,trh
	clkmode	rc
	push	y
	ld	menunum
	sub	#48
	callf	lcd_print
	pop	y
	clkmode	sub
.loop	callf	key_get
	be	#KEY_B,.abort
	be	#KEY_A,.ok
	be	#KEY_UP,.up
	be	#KEY_RIGHT,.up
	be	#KEY_DOWN,.down
	be	#KEY_LEFT,.down
	br	.loop
.up	ld	menunum
	be	#'Z',.loop		; biggest one already reached
	inc	menunum
	callf	key_click
.snend1	ld	snd_playing
	bnz	.snend1
	br	.print
.down	ld	menunum
	be	#'0',.loop		; smallest one already reached
	dec	menunum
	callf	key_click
.snend2	ld	snd_playing
	bnz	.snend2
	br	.print
.abort	; B pressed, clear char,cursor and abort
	callf	key_click
.snend3	ld	snd_playing
	bnz	.snend3
	clkmode	rc
	mov	#<font5x5,trl
	mov	#>font5x5,trh
	push	y
	mov	#16,acc		; [SPACE]
	callf	lcd_print
	pop	y
	mov	#6,width
	mov	#1,height
	mov	#<font5x1,trl
	mov	#>font5x1,trh
	push	y
	add_st	y,#6,y
	mov	#1,acc		; [SPACE]
	callf	lcd_print
	pop	y
	clkmode	sub
	mov	#0,acc
	ret
.ok	; A pressed, clear cursor and return char
	callf	key_click
.snend4	ld	snd_playing
	bnz	.snend4
	clkmode	rc
	mov	#6,width
	mov	#1,height
	mov	#<font5x1,trl
	mov	#>font5x1,trh
	push	y
	add_st	y,#6,y
	mov	#1,acc		; [SPACE]
	callf	lcd_print
	pop	y
	clkmode	sub
	ld	menunum		; get char entered
	ret

; --- alien handling code starts here -----------------------------------------

; return first free alienobject in 0, set cy if error
alien_getfree
	mov	#alienobjects,0		; start with first alien
.loop	ld	@r0			; type of alien
	bz	.found			; 0 = free
	add_st	0,#ALNOFF_SIZE,0	; next alien
	bne	#alienend,.loop		; check for last
	set1	psw,cy			; last one, none found, return error
	ret
.found	clr1	psw,cy			; alien found, clear error flag
	ret

; display the alien @r0
; destroys 1
alien_display
	; test if alien should be displayed
	add_st	0,#ALNOFF_FRAME,1
	ld	@r1
	bn	acc,7,.draw		; "do not display" bit is not set, show
	clr1	acc,7			; clear "do not display bit"
	st	@r1
	ret
.draw	; set x and y position
	add_st	0,#ALNOFF_X,1
	ld	@r1
	clr1	acc,7			; clear nomove marker from forking
	st	@r1
	st	x
	add_st	0,#ALNOFF_Y,1
	ld	@r1
	st	y
	; test if left pixel is still on screen
	ld	x
	sub	#48
	bn	psw,cy,.out_of_range	; x > 47
	; test if top pixel is still on screen
	ld	y
	sub	#40
	bn	psw,cy,.out_of_range	; y > 39

	; push wram_printwsimple onto stack as next routine to call after prep
	mov	#<wram_printsimple,acc
	push	acc			; low byte of next routine after prep
	mov	#>wram_printsimple,acc
	push	acc			; high byte of next routine after prep
	; handle different type by own routine, look up which one in table
	ld	@r0			; ALNOFF_TYPE
	dec	acc
	rol
	mov	#<alienpreptable,trl
	mov	#>alienpreptable,trh
	st	c
	ldc
	push	acc			; low byte of address
	ld	c
	inc	acc
	ldc
	push	acc			; high byte of address
	ret				; jump indirect
.out_of_range
	mov	#0,@r0
	ret

; move the alien @r0: execute alienscript
; NOTE:
; The alienscripts are just native lc86k code, a pointer to the current alien
; is in 0. To stop the execution until the next frame use callf alien_endscript.
; The callf is necessary to get to the address of the PC, it is pop'ed from
; the stack and stored for later use. To make life easier I have written some
; macros for often used small code sequences and some subroutines for bigger
; code sequences.
alien_exec
	mov	#alienobjects,0
alien_exec.loop
	ld	@r0
	bz	alien_exec.skip		; test if object empty
	add_st	0,#ALNOFF_X,1
	ld	@r1
	bp	acc,7,alien_exec.skip	; test for freshly forked
	; increment frame
	add_st	0,#ALNOFF_FRAME,1
	ld	@r1
	and	#$80			; just the "no display bit"
	st	b
	inc	@r1			; next frame
	ld	@r1
	and	#$7f			; clear "no display bit"
	or	b			; put old "no display bit" back
	st	@r1
	; jump indirect to alienscript PC
	add_st	0,#ALNOFF_PCL,1
	ld	@r1
	push	acc
	inc	1
	ld	@r1
	push	acc
	ret				; jump indirect

	; back from the alien script, store new PC
alien_endscript
	add_st	0,#ALNOFF_PCL+1,1
	pop	acc
	st	@r1
	dec	1
	pop	acc
	st	@r1
alien_exec.skip
	add_st	0,#ALNOFF_SIZE,0	; advance to next alien
	bne	#alienend,alien_exec.loop
	ret				; alien_exec finished

; add speed to position
cmd_automove
	add_st	0,#ALNOFF_V,1
	ld	@r1			; speed
	push	acc
	add_st	0,#ALNOFF_X,1
	pop	acc
	push	acc
	and	#$f0			; high nibble is x speed
	ror
	ror
	ror
	ror
	bn	acc,3,.cmdexpandx
	or	#$f0			; expand to 8 bit
.cmdexpandx
	add	@r1			; new x pos = x speed + old x pos
	st	@r1
	inc	1			; @r1 is ALNOFF_Y
	pop	acc			; speed
	and	#$0f			; low nibble is y speed
	bn	acc,3,.cmdexpandy
	or	#$f0			; expand to 8 bit
.cmdexpandy
	; now test gamestate: add 1 to y speed in STATE_GAME/STATE_DEMO
	st	b
	ld	gamestate
	clr1	acc,STATE_DIEFL
	be	#STATE_BOSS,.noinc
	inc	b
.noinc	add_st	b,@r1,@r1		; new y pos = y speed + old y pos
	ret

; start new alien (char=type, x=xoff, y=yoff, width=hits, mask1,2=alienscript)
cmd_fork
	push	0			; gets trashed
	call	alien_getfree		; find free alienobject
	push	0
	pop	1
	pop	0
	.ifd	DEBUG
	bp	psw,cy,.panic10
	.endc
	.ifnd	DEBUG
	bp	psw,cy,.newalienend	; skip if no space
	.endc
	; the old alienobject is @r0, the new one is @r1
	; fill in alienobject, relies on order of elements in alienobject
	ld	char
	st	@r1			; type
	inc	1
	mov	#0,@r1			; frame
	inc	1
	; get old x/y values and add offset
	push	1
	add_st	0,#ALNOFF_X,1
	ld	@r1
	st	b
	add_st	0,#ALNOFF_Y,1
	ld	@r1
	st	c
	pop	1
	; add offsets and store new x/y
	ld	x
	add	b
	set1	acc,7			; nomove marker, cleared at 1st display
	st	@r1			; x
	inc	1
	add_st	y,c,@r1			; y
	inc	1
	ld	mask1
	st	@r1			; script low
	inc	1
	ld	mask2
	st	@r1			; script high
	inc	1
	mov	#0,@r1			; wait
	inc	1
	mov	#0,@r1			; v
	inc	1
	ld	width
	st	@r1			; hitcounter
.newalienend
	ret
.panic10
	mov	#$10,acc		; out of aliens in fork
	jmpf	panic

; check for collisions between aliens and defense shots
; lets aliens explode if necessary
alien_kill
	; for all shots
	mov	#shots,1
.loopshots
	ld	@r1			; SHOTOFF_X
	be	#$ff,.nextshot		; shot empty
	add_st	1,#SHOTOFF_ENERGY,0
	ld	@r0			; SHOTOFF_ENERGY
	bn	acc,7,.nextshot		; collision marker not set, skip shot
	; check all aliens
	mov	#alienobjects,0		; start with first alien
.loopalien
	ld	@r0
	bz	.nextalien		; alien empty
	; check for explosions / invincibility
	push	1
	add_st	0,#ALNOFF_HITS,1
	ld	@r1
	pop	1
	bz	.nextalien		; this one is already exploding
	be	#$ff,.nextalien		; this one is invincible
	; get position of alien
	push	1
	add_st	0,#ALNOFF_X,1
	ld	@r1
	st	x			; get x position of current alien
	add_st	0,#ALNOFF_Y,1
	ld	@r1
	st	y			; get y position of current alien
	pop	1

.check8x8
	; check first for shot and alien within 8x8 pixels
	ld	@r1			; SHOTOFF_X
	inc	acc			; shot is 2 pixels wide
	sub	x
	bp	psw,cy,.nextalien	; missed on left side
	sub	#9			; assume max width==8
	bn	psw,cy,.nextalien	; missed on right side
	inc	1
	ld	@r1			; SHOTOFF_Y
	dec	1
	inc	acc			; shot is 2 pixel high
	sub	y
	bp	psw,cy,.nextalien	; missed: shot before alien
	sub	#9			; assume max height==8
	bn	psw,cy,.nextalien	; missed: alien before shot
	call	alien_hitnear

.nextalien
	add_st	0,#ALNOFF_SIZE,0
	bne	#alienend,.loopalien

.nextshot
	add_st	1,#SHOTOFF_SIZE,1
	bne	#shotsend,.loopshots

	ret

; alien and shot are within 8x8 pixel range, now check bit pattern
; 0 pointer to alien
; x/y position of alien
; 1 pointer to shot
alien_hitnear
	; get pointer to current alien bitmap
	ld	@r0
	push	1			; will be destroyed by same aliens
	; push .check onto stack as next routine to call after prep
	mov	#<.check,acc
	push	acc			; low byte of next routine after prep
	mov	#>.check,acc
	push	acc			; high byte of next routine after prep
	; handle different type by own routine, look up which one in table
	ld	@r0			; ALNOFF_TYPE
	dec	acc
	rol
	mov	#<alienpreptable,trl
	mov	#>alienpreptable,trh
	st	c
	ldc
	push	acc			; low byte of address
	ld	c
	inc	acc
	ldc
	push	acc			; high byte of address
	ret

.check	pop	1			; restore from above prep...
	; trh/trl points to first line of alien font, acc is the font offset
	; width/height contain the size of the character in pixel
	; normalise charpointer offset
	add	trl			; lowbyte of pointer + offset
	st	trl			; store it
	bn	psw,cy,.noinctrh
	inc	trh
.noinctrh
	; trh/trl now points to first byte of current alien bitmap
	inc	1
	ld	@r1			; SHOTOFF_Y
	dec	1
	sub	y			; y offset between alien and shot
	push	acc			; used later as pointer to bitmap line
	; calculate x distance
	sub_st	@r1,x,c			; SHOTOFF_X
	bn	psw,cy,.dodouble
	; special situation: only 1 pixel of shot overlaps
	mov	#%10000000,b
	br	.shiftend
	; shift bitmap of shot
.dodouble
	mov	#%11000000,b
	bz	.shiftend		; no shifting needed
	ld	b
.shift
	ror
	dbnz	c,.shift
	clr1	acc,7			; if c==7 the highest bit is set again
	st	b
.shiftend
	pop	acc			; pointer to overlapping bitmap line
	st	c			; needed later more than once
	; test if y offset is within size of alien
	sub	height
	bp	psw,cy,.heightok1
	ret				; alien too small
.heightok1
	; now compare the bitmaps
	ld	c			; offset to overlapping bitmap line
	ldc				; bitmap line of alien
	and	b			; compare with shot
	bnz	alien_hit		; alien hit
	; check second line of shot, although the aliens hit by this one
	; have really strange forms
	inc	c			; next line
	; test if y offset for 2nd line is within size of alien
	ld	c
	sub	height
	bp	psw,cy,.heightok2
	ret				; alien too small
.heightok2
	; now compare the bitmaps for second line
	ld	c			; offset to bitmap line
	ldc				; bitmap line of alien
	and	b			; compare with shot
	bnz	alien_hit		; alien hit
	ret				; missed alien

; alien @r0 is hit by shot @r1
alien_hit
	push	1
	add_st	1,#SHOTOFF_ENERGY,1
	ld	@r1
	clr1	acc,7			; clear collision flag
	st	b
	pop	1
	mov	#$ff,@r1		; mark shot as unused
	push	0
	add_st	0,#ALNOFF_HITS,0
	ld	@r0
	sub	b
	st	@r0
	bz	alien_killed		; hits = 0, alien dead
	bp	psw,cy,alien_killed	; hit < 1, alien dead
	; mark alien as hit, don't display next time
	pop	0
	push	1
	add_st	0,#ALNOFF_FRAME,1
	ld	@r1
	set1	acc,7
	st	@r1
	pop	1
	; play sound: shot hit alien without killing
	mov	#<tune_hit,trl
	mov	#>tune_hit,trh
	callf	snd_frameplaytune
	ret				; the shot was non deadly

; the alien @r0(stack) got hit deadly, let it explode
alien_killed
	mov	#0,@r0			; hitcounter could be negative
	pop	0
	push	1			; needed for outer loops
	ld	@r0			; old alien type
	be	#ALN_ASTEROID,.score1
	be	#ALN_CANNONL,.score1
	be	#ALN_CANNONR,.score1
	be	#ALN_CANNONU,.score1
	be	#ALN_CANNONDL,.score1
	be	#ALN_CANNONDUL,.score1
	be	#ALN_CANNONDR,.score1
	be	#ALN_CANNONDUR,.score1
	be	#ALN_BOMBERL,.score1
	be	#ALN_BOMBERLF,.score1
	be	#ALN_BOMBERR,.score1
	be	#ALN_BOMBERRF,.score1
	be	#ALN_TINY,.score5
	be	#ALN_SMALL,.score5
	be	#ALN_SMALL2,.score5
	be	#ALN_SMALL3,.score5
	be	#ALN_DIA1,.score5
	be	#ALN_DIA2,.score5
	be	#ALN_DIA3,.score5
	be	#ALN_DIA4,.score5
	be	#ALN_SMILE,.score5
	be	#ALN_BOSS1B,.score5
	be	#ALN_BOSS1L,.score5
	be	#ALN_BOSS1R,.score5
	be	#ALN_BOSS1M,.score5
	be	#ALN_BOSS1F,.score5
	be	#ALN_SINGLE,.score10
	be	#ALN_DOUBLE,.score10
	be	#ALN_TENTACLE,.score10
	be	#ALN_MOTHER,.score20
	mov	#0,b			; score
	br	.score
.score1
	mov	#1,b
	br	.score
.score5
	mov	#5,b
	br	.score
.score10
	mov	#$10,b
	br	.score
.score20
	mov	#$20,b
	br	.score
.score
	ld	b			; get score
	call	score_add		; add it
	mov	#ALN_EXPLOSION1,@r0	; set type
	add_st	0,#ALNOFF_FRAME,1
	mov	#0,@r1			; start with new frame
	add_st	0,#ALNOFF_PCL,1
	mov	#<scr_explosion1,@r1	; explosionscript low
	add_st	0,#ALNOFF_PCH,1
	mov	#>scr_explosion1,@r1	; explosionscript high
	add_st	0,#ALNOFF_V,1
	mov	#0,@r1			; speed = 0
	add_st	0,#ALNOFF_WAIT,1
	mov	#0,@r1			; wait = 0
	pop	1			; restore from above
	; play sound: alien killed
	mov	#<tune_kill,trl
	mov	#>tune_kill,trh
	callf	snd_frameplaytune
	ret

; --- defense ship and shot handling code starts here -------------------------

; draw the defense shots
shots_draw
	mov	#shots,0
.loop	ld	@r0
	st	x			; xposition
	inc	0
	ld	@r0
	st	y			; y position
	inc	0
	inc	0			; ignore energy in drawing
	ld	x
	be	#$ff,.next		; shot empty, do not print
	mov	#<fnt_shot,trl
	mov	#>fnt_shot,trh
	mov	#2,width
	mov	#2,height
	mov	#0,acc
	mov	#0,printmode		; or mode
	callf	wram_print
	; check for collision
	ld	collision
	bz	.next			; not collided
	dec	0			; SHOTOFF_ENERGY
	ld	@r0
	set1	acc,7			; collision marker
	st	@r0
	inc	0
.next	ld	0
	bne	#shotsend,.loop
	ret

; move the defense shots
shots_move
	mov	#shots,0
.loop	ld	@r0
	inc	0			; SHOTOFF_Y
	be	#$ff,.next		; shot not used
	dec	@r0			; one step up
	dec	@r0			; two steps up
	ld	@r0
	be	#$06,.remove		; reached top border
	be	#$05,.remove		; reached top border
	inc	0			; SHOTOFF_ENERGY
	ld	@r0
	dec	0			; SHOTOFF_Y
	bn	acc,7,.next		; not collided
.remove	; top border, delete current shot
	dec	0			; SHOTOFF_X
	mov	#$ff,@r0		; clear x
	inc	0			; SHOTOFF_Y
.next	inc	0			; SHOTOFF_ENERGY
	inc	0			; next shot
	ld	0
	bne	#shotsend,.loop
	ret

; draw the defense ship / explosion, handle game over
ship_draw
	ld	gamestate
	bp	acc,STATE_DIEFL,.dying	; handle dying differently
	clr1	acc,STATE_DIEFL
	bne	#STATE_DEMO,.nodemo
	ret				; no ship handling in demo mode

.nodemo	; draw the ship
	mov	#<fnt_ship,trl
	mov	#>fnt_ship,trh
	mov	#4,width
	mov	#3,height
	ld	shipx
	st	x
	ld	shipy
	st	y
	; blink in highlander mode
	ld	highlander
	ror
	and	#1
	st	printmode
	mov	#0,acc
	callf	wram_print
	; check for undieability
	ld	highlander
	bz	.candie			; we can die
	dec	highlander
	ret				; still undieable, ship finished

.candie	; check for collision
	ld	collision
	bnz	.die
	ret				; not collided, ship finished

.die	; defense ship collided with something, die
	mov	#0,highlander		; used as framecounter
	set1	gamestate,STATE_DIEFL
	mov	#<tune_explode,trl
	mov	#>tune_explode,trh
	callf	snd_frameplaytune	; play ship explode sound
.dying	mov	#<fnt_expl1,trl
	mov	#>fnt_expl1,trh
	mov	#8,width
	mov	#8,height
	ld	shipx
	st	x
	; shift position as needed to show explosion full screen
	sub	#41
	bp	psw,cy,.skipx
	mov	#40,x
.skipx	ld	shipy
	st	y
	sub	#33
	bp	psw,cy,.skipy
	mov	#32,y
.skipy	ld	highlander
	and	#$18			; 4 frames, time 8, size 8
	mov	#0,printmode		; or mode
	callf	wram_print
	inc	highlander
	ld	highlander
	be	#32,.newshp		; check for last frame of explosion
	ret

.newshp	; explosion finished, use next ship, highlander is still 32
	clr1 	gamestate,STATE_DIEFL	; no longer dieing
	dec	lifes			; reduce number of lifes
	ld	lifes
	bz	.over
	ret				; still lifes left

.over	; all ships destroyed, GAME OVER
	callf	snd_init		; abort all sound playing and reinit
	clr1	t0con,t0lrun		; stop frame timer
	clkmode	sub			; back to subclk
	callf	key_init		; reinit keyboard routines
	pop	acc			; need to end routine one level up
	pop	acc			; so remove one address from stack
	mov	#$ff,acc		; game over mark
	ret

; move the defense ship and handle keyboard events
ship_move
	ld	gamestate
	bp	acc,STATE_DIEFL,.move3	; exploding ships dont move
	clr1	ACC,STATE_DIEFL
	bne	#STATE_DEMO,.move0

	; STATE_DEMO, only test quit/sleep/end
	bp	p3,4,.move0		; A not pressed
.waita	bn	p3,4,.waita
	; demo finished
	pop	acc			; remove subroutine address
	pop	acc			; from stack, ugly
	callf	snd_init		; abort all sound playing and reinit
	clr1	t0con,t0lrun		; stop frame timer
	clkmode	sub			; back to subclk
	; reinit keyboard routines
	callf	key_init
	mov	#$ff,acc		; game over mark
	ret

.move0	; STATE_PLAY or STATE_BOSS
	bn	p3,0,.up
	bn	p3,1,.down
.move1	bn	p3,2,.left
	bn	p3,3,.right
.move2	bn	p3,4,.shota
	.ifd	DEBUG
	.ifnd	DEBUG_SINGLE
	bn	p3,5,.shotb		; test B if DEBUG, not if DEBUG_SINGLE
	.endc
	.endc
	mov	#1,firerepeat		; button A released, clear repeat
.move3	bn	p3,6,.quit		; mode
	bn	p3,7,.sleep		; sleep
.move4	bp	p7,0,.quit		; docked?
	ret

	.ifd	DEBUG
.shotb	mov	#$ff,acc
	jmpf	panic			; panic if button B pressed in DEBUG
	.endc

.up	ld	shipy
	be	#8,.move1		; top border reached
	dec	shipy
	br	.move1			; continue button check

.down	ld	shipy
	be	#37,.move1		; bottom border reached
	inc	shipy
	br	.move1			; continue button check

.left	ld	shipx
	bz	.move2			; left border reached
	dec	shipx
	br	.move2			; continue button check

.right	ld	shipx
	be	#44,.move2		; right border reached
	inc	shipx
	br	.move2			; continue button check

.shota	; A button is pressed, check for repeat
	dbnz	firerepeat,.move3	; repeat timeout not reached
	mov	#5,firerepeat		; how many frames to wait between repeats
	; search a free shot
	mov	#shots,0
.shotfind
	ld	@r0			; x position
	be	#$ff,.shotastart	; marked empty?
	add_st	0,#SHOTOFF_SIZE,0	; next shot
	be	#shotsend,.move3	; all shots used up, ignore A button
	br	.shotfind
.shotastart
	; start shot
	mov	#<tune_shot,trl
	mov	#>tune_shot,trh
	callf	snd_frameplaytune
	ld	shipx
	inc	acc			; adjust x position to ship center
	st	@r0			; is always on screen
	ld	shipy
	sub	#3			; adjust y position to before ship
	bp	psw,cy,.shotabort	; ship too near to top of screen
	inc	0
	st	@r0
	inc	0
	; FUTURE: higher energy shots
	mov	#1,@r0
	br	.move3
.shotabort
	; abort shot, out of screen
	mov	#$ff,acc
	st	@r0
	br	.move3

.quit	jmpf	goodbye

.sleep	push	ocr
	push	ie
	push	t1cnt
	mov	#0,t1cnt		; stop sound
	mov	#0,vccr			; LCD off
	set1	ie,mastie		; enable all irq
	mov	#$23,ocr		; 16KHz only subclock runnning
.sleep1	; wait until sleep released
	bp	p7,0,.quit		; docked?
	bn	p3,7,.sleep1
.sleep2	; wait until sleep pressed again
	set1	pcon,halt
	bp	p7,0,.quit		; docked?
	bp	p3,7,.sleep2
.sleep3	bp	p7,0,.quit		; docked?
	bn	p3,7,.sleep3
	; wake up again
	mov	#$80,vccr		; LCD on
	pop	t1cnt
	pop	ie
	pop	ocr
	ret

; --- level handling code starts here -----------------------------------------

; execute the levelscript
levelscript
	; test if position for next command reached
	ld	levelscrl
	st	trl
	ld	levelscrh
	st	trh
	mov	#LVLOFF_LINEL,acc
	ldc			; line low byte
	sub	levellinel
	bz	.cont1		; = low byte of current line
	ret			; line not reached, end

.cont1	mov	#LVLOFF_LINEH,acc
	ldc			; line high byte
	sub	levellineh
	bz	.cont2		; = high byte of current line
	ret			; line not reached, end

.cont2	; ok, insert new alien
	callf	alien_startnew
	add16	#LVLOFF_SIZE,levelscr	; advance levelscript
	br	levelscript

; insert the aliendata at trh/trl into a new alien
; trashes 0
alien_startnew
	; find free alienobject
	callf	alien_getfree
	.ifd	DEBUG
	bp	psw,cy,.panic11
	.endc
	.ifnd	DEBUG
	bp	psw,cy,.newalienend	; skip if no space
	.endc
	; fill in alienobject, relies on order of elements in alienobject
	mov	#LVLOFF_TYPE,acc
	ldc
	st	@r0			; type
	inc	0
	mov	#0,@r0			; frame
	inc	0
	mov	#LVLOFF_X,acc
	ldc
	st	@r0			; x
	inc	0
	mov	#LVLOFF_Y,acc
	ldc
	st	@r0			; y
	inc	0
	mov	#LVLOFF_SCRL,acc
	ldc
	st	@r0			; script low
	inc	0
	mov	#LVLOFF_SCRH,acc
	ldc
	st	@r0			; script high
	inc	0
	mov	#0,@r0			; wait
	inc	0
	mov	#0,@r0			; v
	inc	0
	mov	#LVLOFF_HITS,acc
	ldc
	st	@r0			; hitcounter
.newalienend
	ret
.panic11
	mov	#$11,acc
	jmpf	panic

; --- routines drawing directly to lcd ----------------------------------------

; clear the screen
; trashes: 2, c, xbnk
lcd_clear
	mov	#0,xbnk		; start in lcd bank 0
	mov	#$80,2		; init pointer to lcd
.loop	mov	#12,c		; 12 bytes = 2 lines = 1 double-line
.line	mov	#0,@r2		; clear LCD byte
	inc	2		; advance lcd pointer
	dbnz	c,.line		; next byte
	add_st	2,#4,2		; advance to start of next double-line
	bn	psw,cy,.loop	; next line
	mov	#$80,2		; reinit pointer to lcd
	inc	xbnk		; next lcd bank
	bn	xbnk,1,.loop	; continue copying if in lcd bank 0/1
	ret

; copy a bitmap to the screen
; trl/trh pointer to bitmap
; destroys acc, b, c, 2, trl/trh, xbnk
lcd_fromrom
	mov	#0,b		; rom pointer offset
	mov	#0,xbnk		; start in lcd bank 0
	mov	#$80,2		; init pointer to lcd
.cpyloop
	mov	#12,c		; 12 bytes = 2 lines = 1 double-line
.cpyline
	ld	b		; get rom-offset
	ldc			; get byte from rom
	st	@r2		; put into screen
	inc	2		; advance lcd pointer
	inc	b		; advance rom-offset
	dbnz	c,.cpyline	; next byte

	add_st	2,#4,2		; advance to start of next double-line
	bn	psw,cy,.cpyloop	; next line
	mov	#$80,2		; reinit pointer to lcd
	inc	xbnk		; next lcd bank
	bn	xbnk,1,.cpyloop	; continue copying if in lcd bank 0/1
	ret

; prints a char into the LCD
; trl/trh pointer to font
; x/y position
; width of font
; height of font
; acc char

; 0 <= x <= 47 and 0 <= y <= 31
; this version tolerates left and bottom parts leaving screen

; mask1, mask2: bitmasks for 1st, 2nd byte
; char: pointer to fonttable

; trashes b, y, 2
lcd_print
	push	c
	push	xbnk
	push	trl
	push	trh
	push	height

	; calculate charpointer offset
	st	c		; 1st operand: number of char
	ld	height		; height of one char...
	st	b		; ...is 2nd operand
	xor	acc		; highbyte=0
	mul			; charpointer offset now in B/ACC/C
	push	acc		; highbit needed later
	add_st	trl,c,trl	; add lowbyte of offset
	pop	acc		; get highbyte of offset
	addc	trh		; add carry and highbyte of pointer
	st	trh		; store it
	mov	#0,char		; initial offset

	; calculate LCD line
	ld	y
	rol
	rol
	rol
	and	#$70
	add	#$80
	bn	y,0,.print_line
	add	#6
.print_line
	st	2
	; add x position
	ld	x
	ror
	ror
	ror
	and	#$7
	add	2
	st	2

	; calculate LCD bank
	ld	y
	ror
	ror
	ror
	ror
	and	#1
	st	xbnk

	; calculate bitmasks
	ld	width		; width of chars
	st	c		; counter
	xor	acc		; acc=0
.setmask
	set1	acc,0		; set lowest bit
	ror			; rotate lowest bit to highest
	dbnz	c,.setmask
	; the highest width bits of acc are now set
	st	mask1
	mov	#0,mask2
	ld	x
	and	#7
	st	c
	bz	.print_loop	; bits are aligned
	ld	mask1		; get mask for 1st byte
	clr1	psw,cy		; clear carry
.shiftmask
	rorc			; shift mask1
	xch	mask2
	rorc			; shift mask2
	xch	mask2
	dbnz	c,.shiftmask
	st	mask1

	; now check if 2nd byte is off screen = 1st byte is at end of line
	ld	x
	and	#%00111000
	; this test is only valid if x < 48, bad things can happen otherwise
	bne	#%00101000,.print_loop	; x / 8 != 5, everything on screen
	mov	#0,mask2		; mark 2nd byte as unused

	; loop for every line in char
.print_loop
	ld	x		; get x position
	and	#7
	st	c		; x % 8, number of rotations
	inc	c		; needed for test if x % 8 == 0
	ld	char		; get char data
	ldc			; from rom
	dbnz	c,.rot1		; test if x % 8 == 0
	br	.rot1end	; no need to rotate
.rot1	ror			; rotate it
	dbnz	c,.rot1
.rot1end

	; do the printing: 1st byte
	push	acc		; (1) store char bitmap for 2nd byte
	and	mask1
	st	c
	ld	mask1
	xor	#$ff
	and	@r2		; clear width bits in LCD
	or	c		; insert char bitmap
	st	@R2		; put into LCD
	; handle 2nd byte
	pop	c		; pop shifted font from stack (1)
	ld	mask2		; get mask
	bz	.print_no2nd	; skip 2nd byte completely if possible
	and	c		; mask font
	st	c
	inc	2
	ld	mask2
	xor	#$ff
	and	@R2
	or	c
	st	@R2		; put into LCD
	dec	2		; previous LCD byte
.print_no2nd
	; advance to next LCD line
	ld	2			; get LCD position
	add	#6			; next line
	bn	y,0,.print_lineskip
	add	#4			; current row was even
.print_lineskip
	st	2			; store LCD position
	bp	2,7,.print_nobank	; test for bankswitch
	inc	xbnk			; next bank
	bp	xbnk,1,.print_end	; next bank is 2: out of normal screen
	add_st	2,#$80,2		; correct position for new bank
.print_nobank
	; advance pointer to font
	inc	char
	; position tracking
	inc	y
	dbnz	height,.print_loop

.print_end
	pop	height
	pop	trh
	pop	trl
	pop	xbnk
	pop	c
	ret

; prints a \0 terminated string pointed to trl/trh in 6x5 font
; x/y startposition of string
; restriction: string must be shorter than 256 chars, string must fit onto screen
lcd_printstring
	mov	#6,width
	mov	#5,height
	mov	#0,c
.loop	ld	c
	ldc
	bnz	.noret
	ret			; \0, end of string
.noret	push	trl
	push	trh
	push	y
	mov	#<font5x5,trl
	mov	#>font5x5,trh
	sub	#48
	callf	lcd_print
	pop	y
	pop	trh
	pop	trl
	inc	c
	add_st	x,width,x
	br	.loop

; print a hex value in acc at (x,y) in font5x5
; trashes: trl, trh, acc, width, height
lcd_printhex
	mov	#5,width
	mov	#5,height
	mov	#<font5x5,trl
	mov	#>font5x5,trh
	push	acc
	rol
	rol
	rol
	rol
	and	#$0f
	callf	.nibble2char
	push	y
	callf	lcd_print
	pop	y
	add_st	x,#6,x
	pop	acc
	and	#$0f
	callf	.nibble2char
	callf	lcd_print
	ret
; convert a nibble from $00-$0f into a char usable for print
.nibble2char
	be	#$0a,.nonnumeric
	be	#$0b,.nonnumeric
	be	#$0c,.nonnumeric
	be	#$0d,.nonnumeric
	be	#$0e,.nonnumeric
	be	#$0f,.nonnumeric
	ret
.nonnumeric
	add	#$07
	ret

; --- full screen drawing routines to lcd -------------------------------------

; announce the calibration
screen_calibrate
	clr1	ocr,subclk		; use rc osc, not calibrated yet
	callf	lcd_clear
	mprstr	str_calibrate1,0,0
	mprstr	str_calibrate2,6,6
	mprstr	str_calibrate3,0,12
	mprstr	str_calibrate4,0,18
	set1	ocr,subclk		; use sub clk, not calibrated yet
	ret

; show the highscores
screen_highscore
	clkmode	rc			; draw in rc clk
	callf	lcd_clear
	mov	#0,menunum		; used as pointer to highscore entry
.loop	mov	#<highscores,trl
	mov	#>highscores,trh
	ld	menunum
	ldc				; highest score byte
	st	score3
	ld	menunum
	add	#1
	ldc				; middle score byte
	st	score2
	ld	menunum
	add	#2
	ldc				; lowest score byte
	st	score1
	mov	#0,x
	ld	menunum			; one entry is 6 bytes = height of font
	st	y
	mov	#<font5x5,trl
	mov	#>font5x5,trh
	mov	#6,width
	mov	#5,height
	callf	lcd_printscore		; print score part
	mov	#>highscores,trh
	mov	#<highscores+3,acc	; offset to name string
	add	menunum			; offset of current entry
	st	trl			; now points to 2 byte string of name

	mov	#37,x			; x position of name string
	ld	menunum
	st	y
	callf	lcd_printstring

	add_st	menunum,#6,menunum
	bne	#30,.loop		; print all 5 entries
	clkmode	sub
	ret

; show new highscore screen
screen_newhigh1
	clkmode	rc
	callf	lcd_clear
	mprstr	str_newhigh,0,0		; "NEW HI-"
	mprstr	str_score,0,6		; "SCORE"
	mov	#<font5x5,trl
	mov	#>font5x5,trh
	mov	#6,width
	mov	#5,height
	mov	#13,x
	mov	#12,y
	callf	lcd_printscore		; score
	clkmode	sub
	ret

; show enter name screen
screen_newhigh2
	clkmode	rc
	mprstr	str_enter,0,18		; "ENTER"
	mprstr	str_name,0,24		; "NAME"
	clkmode	sub
	ret

; show level, lifes and score
screen_level
	clkmode	rc			; draw in rc clk
	callf	lcd_clear
	mprstr	str_level,0,0		; "LEVEL"
	; level number
	mov	#<font5x5,trl
	mov	#>font5x5,trh
	mov	#6,width
	mov	#5,height
	mov	#36,x
	ld	level
	callf	lcd_print
	mprstr	str_ships,0,9		; "SHIPS"
	; number of lifes
	mov	#<font5x5,trl
	mov	#>font5x5,trh
	mov	#6,width
	mov	#5,height
	mov	#36,x
	ld	lifes
	callf	lcd_print
	mprstr	str_sc,0,18		; "SCORE"
	; score
	mov	#<font5x5,trl
	mov	#>font5x5,trh
	mov	#6,width
	mov	#5,height
	mov	#13,x
	mov	#18,y
	callf	lcd_printscore
	clkmode	sub			; back to subclk
	ret

; display game over and the final score
screen_gameover
	clkmode	rc			; draw in rc clk
	callf	lcd_clear
	mprstr	str_over1,12,0		; "GAME"
	mprstr	str_over2,12,9		; "OVER"
	mprstr	str_score,9,18		; "SCORE"
	mov	#<font5x5,trl
	mov	#>font5x5,trh
	mov	#6,width
	mov	#5,height
	mov	#6,x
	mov	#27,y
	callf	lcd_printscore		; score
	clkmode	sub			; back to subclk
	ret

; a severe problem has occured, its so bad that its better to stop the game
; and print a diagnostic message. should only happen if DEBUG is defined.
; current panics are:
; 01 unknown gamestate
; 10 out of alien ram in cmd_fork
; 11 out of alien ram in levelscript
; 20 level out of range
; 30 write to flash failed
; 41 scr_triple shooting neither ALN_SINGLE nor ALN_DOUBLE
; ff button b pressed
panic
	push	acc
	; reinit everything
	clr1	ie,mastie		; disable all irq
	clkmode	sub
	mov	#0,t0con		; turn off everything
	mov	#$fc,t0prr		; prescaler to / 4
	mov	#$78,t0lr		; timer0 low to / 128
	mov	#$c0,t0hr		; timer0 high to / 64
	mov	#$b6,bogomips		; for 600KHz, exact timing unnecessary
	callf	key_init		; initialise keyboard
	callf	snd_init		; initialise sound
	mov	#0,printmode
	set1	ie,mastie		; enable all irq

	clkmode	rc
	callf	lcd_clear
	mprstr	str_panic,0,0		; "PANIC"
	mov	#36,x
	mov	#0,y
	pop	acc
	callf	lcd_printhex		; error code
	; dump parameters
	mov	#0,x
	mov	#6,y
	ld	levellineh
	callf	lcd_printhex
	mov	#12,x
	mov	#6,y
	ld	levellinel
	callf	lcd_printhex
	mov	#36,x
	mov	#6,y
	ld	gamestate
	callf	lcd_printhex

	mov	#0,x
	mov	#12,y
	ld	levelscrh
	callf	lcd_printhex
	mov	#12,x
	mov	#12,y
	ld	levelscrl
	callf	lcd_printhex
	mov	#36,x
	mov	#12,y
	ld	level
	callf	lcd_printhex

	mov	#0,x
	mov	#18,y
	ld	snd_qin
	callf	lcd_printhex
	mov	#12,x
	mov	#18,y
	ld	snd_qout
	callf	lcd_printhex
	mov	#36,x
	mov	#18,y
	ld	snd_playing
	callf	lcd_printhex

	mov	#0,x
	mov	#24,y
	ld	fps
	callf	lcd_printhex
	mov	#18,x
	mov	#24,y
	ld	bogomips
	callf	lcd_printhex

	clkmode	sub

	callf	key_waitab
	jmpf	start

; --- menu handling code ------------------------------------------------------

; completely handle a menu
; trh/trl: menu data
; acc: number of items
; b: timeout value

; returns
; acc: which item is selected, 0 = abort/timeout
domenu
	push	acc
	ld	b
	st	firerepeat	; register reuse: time counter
	pop	acc
	push	key_repeat
	set1	key_repeat,7
	st	menumax		; number of items
	st	b
	mov	#0,y		; start line
	clkmode	rc
.loop	mov	#6,x		; leave room for cursor
	push	b		; (1)
	callf	lcd_printstring	; print item string
	; find end of string
	mov	#0,c
.search	inc	c
	ld	c
	ldc
	bnz	.search		; end marker not yet found, try next
	inc	c		; start of next item
	add16	c,tr		; advance rom pointer to next item
	add_st	y,#6,y		; next line on screen
	pop	b		; from (1)
	dbnz	b,.loop
	clkmode	sub
	; prepare timer
	mov	#$78,t0lr	; timer0 low to / 136
	set1	t0con,t0lrun	; start timer
	mov	#0,menunum	; first item is selected
	; parameters for print
	mov	#0,x
	mov	#5,width
	mov	#5,height
.selectloop
	; print cursor
	clkmode	rc
	ld	menunum		; number of current item
	st	b
	mov	#6,c
	xor	acc
	mul
	ld	c
	st	y		; y position of current item
	mov	#<font5x5,trl
	mov	#>font5x5,trh
	mov	#10,acc
	push	y
	callf	lcd_print
	pop	y
	clkmode	sub
.keyloop
	callf	key_get
	be	#KEY_A,.select
	be	#KEY_B,.abort
	be	#KEY_UP,.up
	be	#KEY_DOWN,.down
	bn	t0con,t0lovf,.keyloop	; timer not finished yet
	dec	firerepeat		; register reuse: time counter
	ld	firerepeat
	bz	.time			; timeout
	clr1	t0con,t0lovf		; clear overflow flag
	br	.keyloop
.select	call	key_click
	clr1	t0con,t0lrun		; stop timer
	ld	menunum			; return item +1
	inc	acc
	pop	key_repeat
	ret
.abort	call	key_click
.time	clr1	t0con,t0lrun		; stop timer
	mov	#0,acc			; return 0
	pop	key_repeat
	ret
.up	ld	menunum
	bz	.keyloop		; skip if already 1st
	dec	menunum			; previous item
	call	key_click
	; clear old cursor
	clkmode	rc
	mov	#<font5x5,trl
	mov	#>font5x5,trh
	mov	#16,acc
	callf	lcd_print
	clkmode	sub
	br	.selectloop
.down	ld	menunum
	sub	menumax
	inc	acc
	bz	.keyloop		; skip if already last
	inc	menunum			; next item
	call	key_click
	; clear old cursor
	clkmode	rc
	mov	#<font5x5,trl
	mov	#>font5x5,trh
	mov	#16,acc
	callf	lcd_print
	clkmode	sub
	jmp	.selectloop

; --- work ram drawing routines starts here -----------------------------------

; clear the wram
; trashes: vrmad, vsel, acc, b
wram_clear
	mov	#0,vrmad1
	mov	#0,vrmad2
	set1	vsel,ince
	mov	#0,acc
	mov	#192,b
.loop	st	vtrbf
	dbnz	b,.loop
	ret

; copy the rom bitmap at trh/trl into wram
; trashes: vrmad, vsel, acc
wram_fromrom
	mov	#0,vrmad1
	mov	#0,vrmad2
	set1	vsel,ince
	mov	#0,acc
.loop	push	acc
	ldc
	st	vtrbf
	pop	acc
	inc	acc
	bne	#192+48,.loop
	ret

; complex print char to wram routine with collision check, or/clear mode
; trl/trh pointer to font
; x/y position
; width of font
; height of font
; printmode: 0=or, 1=clear
; acc char

; returns:
; collision == $ff if there is already an object in wram

; (char+1)*height < 256, only for small fonts
; 0 <= x <= 47 and 0 <= y <= 39
; this version tolerates left and bottom parts leaving screen
; bottom parts just trash more wram

; mask1, mask2: bitmasks for 1st, 2nd byte
; char: pointer to fonttable

; trashes vrmad, vsel, acc, b, c, mask1, mask2, char
wram_print
	; save registers
	push	trl
	push	trh
	push	height

	mov	#$00,collision	; init collision flag
	st	char

	; calculate wram position
	clr1	vsel,ince	; no auto increment
	mov	#0,vrmad2	; high bit always 0
	ld	y
	st	b
	mov	#6,c
	xor	acc
	mul
	
	ld	x
	ror
	ror
	ror
	and	#7		; x/8 = byte offset to start of line
	add	c		; add line offset
	st	vrmad1		; start byte for printing

	; calculate bitmasks
	ld	width		; width of chars
	st	c		; counter
	xor	acc		; acc=0
.setmask
	set1	acc,0		; set lowest bit
	ror			; rotate lowest bit to highest
	dbnz	c,.setmask
	; the highest width bits of acc are now set
	st	mask1		; initial value of mask1
	mov	#0,mask2	; mask2 starts with 0
	ld	x
	and	#7
	st	c		; x % 8
	bz	.loop		; bits are aligned
	ld	mask1		; get mask for 1st byte
	clr1	psw,cy		; clear carry to shifting in 0
.shiftmask
	rorc			; shift mask1
	xch	mask2
	rorc			; shift mask2
	xch	mask2
	dbnz	c,.shiftmask
	st	mask1

	; now check if 2nd byte is off screen = 1st byte is at end of line
	ld	x
	and	#%00111000
	; this test is only valid if x < 48, bad things can happen otherwise
	bne	#%00101000,.loop	; x / 8 != 5, everything on screen
	mov	#0,mask2		; mark 2nd byte as unused

	; loop for every line in char
.loop	ld	x
	and	#7
	st	c		; x % 8, number of rotations
	inc	c		; needed for test if x % 8 == 0
	ld	char		; get char data
	ldc			; from rom
	dbnz	c,.rot1		; test if x % 8 == 0
	br	.rot1end	; no need to rotate
.rot1	ror			; rotate it
	dbnz	c,.rot1
.rot1end

	; do the printing: 1st byte
	push	acc		; (1) store char bitmap for 2nd byte
	and	mask1		; mask out for first byte
	push	acc		; (2) store masked bitmap for later
	and	vtrbf		; collision test
	bz	.nocollision1	; no collision
	mov	#$ff,collision	; collision!
.nocollision1
	ld	printmode	; test for mode: or/clear
	be	#1,.clearmode1
.ormode1			; set all pixels in char
	pop	acc		; masked bitmap from (2)
	or	vtrbf		; set bits
	st	vtrbf		; store new data
	br	.mode1end	; finished
.clearmode1			; clear all pixels in char
	pop	acc		; masked bitmap from (2)
	xor	#$ff		; invert
	and	vtrbf		; clear bits
	st	vtrbf		; store new data
.mode1end

	; do the printing: 2nd byte
	pop	c		; char bitmap from (1)
	ld	mask2		; mask for 2nd byte
	bz	.no2nd		; skip 2nd byte if not needed
	inc	vrmad1		; next byte in wram
	and	c		; mask bitmap for 2nd byte
	push	acc		; (3) store masked bitmap
	and	vtrbf		; collision test
	bz	.nocollision2	; no collision
	mov	#$ff,collision	; collision!
.nocollision2
	ld	printmode	; test for mode: or/clear
	be	#1,.clearmode2
.ormode2			; set all pixels in char
	pop	acc		; masked bitmap from (3)
	or	vtrbf		; set bits
	st	vtrbf		; store new data
	br	.mode2end	; finished
.clearmode2			; clear all pixels in char
	pop	acc		; masked bitmap from (2)
	xor	#$ff		; invert
	and	vtrbf		; clear bits
	st	vtrbf		; store new data
.mode2end
	dec	vrmad1		; back to first byte
.no2nd

	add_st	vrmad1,#6,vrmad1	; next line in wram
	inc	char			; next line in bitmap
	dbnz	height,.loop		; loop until all lines printed

	pop	height			; restore registers
	pop	trh
	pop	trl
	ret

; simple print char to wram routine, or mode
; trl/trh pointer to font
; x/y position
; width of font
; height of font
; acc char

; 0 <= x <= 47 and 0 <= y <= 39
; (char+1)*height < 256, only for small fonts
; this version tolerates left and bottom parts leaving screen
; bottom parts just trash more wram

; mask1, mask2: bitmasks for 1st, 2nd byte
; char: pointer to fonttable

; trashes vrmad, vsel, acc, b, c, mask1, mask2, char
wram_printsimple
	; save registers
	push	trl
	push	trh
	push	height

	; calculate charpointer offset
	st	char

	; calculate wram position
	clr1	vsel,ince	; no auto increment
	mov	#0,vrmad2	; high bit always 0
	ld	y
	st	b
	mov	#6,c
	xor	acc
	mul
	
	ld	x
	ror
	ror
	ror
	and	#7		; x/8 = byte offset to start of line
	add	c		; add line offset
	st	vrmad1		; start byte for printing

	; calculate bitmasks
	ld	width		; width of chars
	st	c		; counter
	xor	acc		; acc=0
.setmask
	set1	acc,0		; set lowest bit
	ror			; rotate lowest bit to highest
	dbnz	c,.setmask
	; the highest width bits of acc are now set
	st	mask1		; initial value of mask1
	mov	#0,mask2	; mask2 starts with 0
	ld	x
	and	#7
	st	c		; x % 8
	bz	.loop		; bits are aligned
	ld	mask1		; get mask for 1st byte
	clr1	psw,cy		; clear carry to shifting in 0
.shiftmask
	rorc			; shift mask1
	xch	mask2
	rorc			; shift mask2
	xch	mask2
	dbnz	c,.shiftmask
	st	mask1

	; now check if 2nd byte is off screen = 1st byte is at end of line
	ld	x
	and	#%00111000
	; this test is only valid if x < 48, bad things can happen otherwise
	bne	#%00101000,.loop	; x / 8 != 5, everything on screen
	mov	#0,mask2		; mark 2nd byte as unused

	; loop for every line in char
.loop	ld	x
	and	#7
	st	c			; x % 8, number of rotations
	inc	c			; needed for test if x % 8 == 0
	ld	char			; get char data
	ldc				; from rom
	dbnz	c,.rot1			; test if x % 8 == 0
	br	.rot1end		; no need to rotate
.rot1	ror				; rotate it
	dbnz	c,.rot1
.rot1end

	; do the printing: 1st byte
	push	acc			; (1) store char bitmap for 2nd byte
	and	mask1			; mask out for first byte
	or	vtrbf			; set bits
	st	vtrbf			; store new data

	; do the printing: 2nd byte
	pop	c			; char bitmap from (1)
	ld	mask2			; mask for 2nd byte
	bz	.no2nd			; skip 2nd byte if not needed
	inc	vrmad1			; next byte in wram
	and	c			; mask bitmap for 2nd byte
	or	vtrbf			; set bits
	st	vtrbf			; store new data
	dec	vrmad1			; back to first byte
.no2nd

	add_st	vrmad1,#6,vrmad1	; next line in wram
	inc	char			; next line in bitmap
	dbnz	height,.loop		; loop until all lines printed

	pop	height			; restore registers
	pop	trh
	pop	trl
	ret

; copy the wram into lcd
; trashes: vrmad, vsel, acc, b, c, 2, xbnk
wram_tolcd
	mov	#48,vrmad1
	mov	#0,vrmad2
	set1	vsel,ince
	mov	#0,xbnk			; start with first bank
.loop3	mov	#$80,2			; first lcd adress in bank
	mov	#8,c			; 8 double lines in one bank
.loop2	mov	#12,b			; 12 bytes in one double line
.loop1	ld	vtrbf
	st	@r2
	inc	2
	dbnz	b,.loop1
	add_st	2,#4,2
	dbnz	c,.loop2
	inc	xbnk
	ld	xbnk
	bp	xbnk,1,.ret		; if in bank 2, finish
	br	.loop3
.ret	ret

; --- misc routines start here ------------------------------------------------

; generate 1 random bit, return in carry
; uses the primitiv polynom 1 + V^14 + V^15 in GF(2)
; the 1 can be ignored if we set U=1 constant
; needs 15 bit shift register in randhigh, randlow
; repeats after 2^15 - 1 bits
; NOTE: 
; Not the best random number generator, it tends to produce quit long
; sequences of 0s/1s. So when I need to shoot every 32nd step in avarage, I
; get 5 bits and test if they are equal to $15. When testing for 0 the shots
; are not evenly distributed.
randombit
	push	acc		; store register
	ld	randhigh
	rol			; acc:6 is now V^14
	xor	randhigh	; acc:6 is now V^14 + V^15 = new bit
	rolc			; new bit in acc:7
	rolc			; new bit now in cy
	push	psw		; store cy, its our return value
	ld	randlow
	rolc			; shift new bit into seed
	st	randlow
	ld	randhigh
	rolc			; shift rest of seed
	st	randhigh
	pop	psw		; get out return value, bit in cy
	pop	acc		; restore register
	ret

; --- keyboard code starts here -----------------------------------------------

; handle periodic keyboard interrupt from timer0h
; NOTE:
; This is only used outside the game playing frameloop.
; Nearly identical to the version I used in minesweeper, except sleep mode,
; see there for more info.
key_irq
	bp	t0con,t0hovf,.doit
	reti				; skip if not timer0 high ovl
.doit	clr1	t0con,t0hovf		; clear interrupt reason
	push	psw			; save registers
	push	acc
	push	0
	push	b
	push	c
	set1	psw,rambk0		; game ram bank
	ld	p3			; p3 state
	push	acc			; save p3 state
	xor	key_keys		; keys which changed status are now 1
	pop	key_keys		; get p3 state and store for next call
	bz	.nochange		; nothing changed, finished
	mov	#1,b			; number of key
	mov	#1,c			; bitmask of key
.loop	push	acc			; store keymask
	and	c			; mask out one bit
	bz	.nopress		; key is not pressed
	ld	key_keys		; get p3 state
	and	c			; mask out one bit
	bz	.norel			; skip release marker
	mov	#$80,acc		; set release flag
.norel	or	b			; keynumber + release flag
	push	key_qin			; move queue input
	pop	0			; to pointer register
	st	@r0			; store key event in pointer
	st	key_lastevent		; store for repeat
	; advance queue pointer
	ld	key_qin			; queue input pointer
	inc	acc			; advance
	bne	#key_qend+1,.qok1	; test for end of queue
	mov	#key_qstart,acc		; end of queue, start again
.qok1	st	key_qin			; new queue input pointer
	; reset repeat counter
	ld	key_repeat
	and	#$80			; mask out repeat flag, set rest to 0
	st	key_repeat
.nopress
	inc	b			; next keynumber
	ld	c
	clr1	psw,cy			; new bit to shift in is 0
	rolc				; next bit to test
	st	c
	pop	acc			; get keymask of all keys
	bn	psw,cy,.loop		; repeat until all bits tested

.nochange
	; test for repeat
	bn	key_repeat,7,.end
	inc	key_repeat		; count interrupts until repeat
	ld	key_repeat
	bne	#$86,.end		; 1 char every 6 ints
	mov	#$80,key_repeat		; restart
	; do repeat
	ld	key_lastevent		; latest key event
	bp	acc,7,.end		; don't repeat key release event
	be	#KEY_SLEEP,.end		; no repeat on SLEEP
	be	#KEY_MODE,.end		; no repeat on MODE
	be	#KEY_A,.end		; no repeat on A
	be	#KEY_B,.end		; no repeat on B
	; insert into queue
	push	key_qin			; move queue input
	pop	0			; to pointer register
	st	@r0			; store key event in pointer
	; advance queue pointer
	ld	key_qin			; queue input pointer
	inc	acc			; advance
	bne	#key_qend+1,.qok2	; test for end of queue
	mov	#key_qstart,acc		; end of queue, start again
.qok2	st	key_qin			; new queue input pointer
.end	pop	c			; restore registers
	pop	b
	pop	0
	pop	acc
	pop	psw
.reti	reti

; returns one pressed keynumber or 0 in acc
; changes keyboard queue pointers
key_getone
	clr1	t0con,t0hie		; disable keyboard irq
	ld	key_qin			; get input pointer
	xor	key_qout		; compare with output pointer
	bz	.end			; no event in queue
	push	0			; save pointer register
	push	key_qout		; move queue output
	pop	0			; to pointer register
	ld	@r0			; store key event in pointer
	pop	0			; restore pointer register
	push	acc			; save key event
	; advance queue pointer
	ld	key_qout		; queue output pointer
	inc	acc			; advance
	bne	#key_qend+1,.qok	; test for end of queue
	mov	#key_qstart,acc		; end of queue, start again
.qok	st	key_qout
	pop	acc			; restore key event
.end	set1	t0con,t0hie		; restore keyboard irq state
	ret

; return one pressed keynumber or 0 in acc
; handles sleep, mode, docked
; changes keyboard queue pointers, vccr
; NOTE:
; This is changed since minesweeper, this version only consumes half as much
; power in sleep mode as the old one. The reason for this is that the VM does
; not wake up at every key interrupt, only at the 0.5 second base timer tick.
; And if the VM wakes up this routine goes back to sleep much earlier than the
; old one.
key_get
	call	key_getone
	bp	p7,0,.quit		; docked?
	be	#KEY_MODE,.quit		; MODE pressed, quit
	be	#KEY_SLEEP,.sleep	; SLEEP pressed, go to sleep
	ret

.quit	jmpf	goodbye
.sleep	mov	#0,vccr			; blank LCD
	callf	snd_suspend
	push	ocr
	push	t0con
	mov	#$23,ocr		; 16KHz only subclock runnning
	clr1	t0con,t0lrun		; stop timer
	clr1	t0con,t0hie		; stop key int
	clr1	t0con,t0hrun		; stop key int timer
.sleep1	; wait until sleep released
	bp	p7,0,.quit		; docked?
	bn	p3,7,.sleep1
.sleep2	; now sleep until sleep pressed again
	set1	pcon,halt
	bp	p7,0,.quit		; docked?
	bp	p3,7,.sleep2
.sleep3	; wait until sleep released again
	bp	p7,0,.quit		; docked?
	bn	p3,7,.sleep3
	; wake up again
	pop	t0con
	pop	ocr
	callf	snd_restart
	mov	#$80,vccr		; reenable LCD
	br	key_get			; check again for keys

; clear the keyboard queue
; destroys acc
key_clear
	clr1	t0con,t0hie		; disable repeat irq
	ld	key_qin			; get input pointer
	st	key_qout		; set to output pointer
	mov	#$80,key_lastevent	; clear key event
	set1	t0con,t0hie		; disable repeat irq
	ret

; wait until A is pressed and released again
key_waita
	call	key_get
	bne	#KEY_A,key_waita	; wait until A down
	call	key_click
.waita2	call	key_get
	bne	#KEY_A|KEY_REL,.waita2	; wait until A up again
	ret

; wait until A and B are pressed
key_waitab
	call	key_get
	be	#KEY_A,.keya			; A pressed
	be	#KEY_B,.keyb			; B pressed
	br	key_waitab
.keya	call	key_get
	be	#KEY_B,.ret			; A+B pressed, finished
	be	#KEY_A|KEY_REL,key_waitab	; A released again, restart
	br	.keya
.keyb	call	key_get
	be	#KEY_A,.ret			; B+A pressed, finished
	be	#KEY_B|KEY_REL,key_waitab	; B released again, restart
	br	.keyb
.ret	call	key_click
	ret

; wait for some time or until A+B pressed
; trashes b, t0lr, t0con
key_timeab
	st	b			; number of ints to wait (1/10 second)
	mov	#$78,t0lr		; timer0 low to / 136
	set1	t0con,t0lrun		; start timer
.loop	clr1	t0con,t0lovf		; clear overflow flag
.loop2	call	key_get
	be	#KEY_A,.keya		; A pressed
	be	#KEY_B,.keyb		; B pressed
	bn	t0con,t0lovf,.loop2	; timer not finished yet
	dbnz	b,.loop
	clr1	t0con,t0lrun		; stop timer
	mov	#$ff,acc		; timeout marker
	ret
	; handle button A down
.keya0	clr1	t0con,t0lovf		; clear overflow flag
.keya	call	key_get
	be	#KEY_B,.end		; A+B pressed, finished
	be	#KEY_A|KEY_REL,.loop2	; A released again, restart
	bn	t0con,t0lovf,.keya	; timer not finished yet
	dbnz	b,.keya0
	clr1	t0con,t0lrun		; stop timer
	mov	#$ff,acc		; timeout marker
	ret
	; handle button B down
.keyb0	clr1	t0con,t0lovf		; clear overflow flag
.keyb	call	key_get
	be	#KEY_A,.end		; B+A pressed, finished
	be	#KEY_B|KEY_REL,.loop2	; B released again, restart
	bn	t0con,t0lovf,.keyb	; timer not finished yet
	dbnz	b,.keyb0
	clr1	t0con,t0lrun		; stop timer
	mov	#$ff,acc		; timeout marker
	ret
.end	call	key_click
	xor	acc			; keys pressed marker
	ret

; wait for some time or until A pressed
; acc: number of ints to wait
; trashes: b, t0con, t0lr
key_timea
	st	b			; number of ints to wait (1/10 second)
	mov	#$78,t0lr		; timer0 low to / 136
	set1	t0con,t0lrun		; start timer
.loop	clr1	t0con,t0lovf		; clear overflow flag
.loop2	call	key_get
	be	#KEY_A,.keya		; A pressed, stop
	bn	t0con,t0lovf,.loop2	; timer not finished yet
	dbnz	b,.loop			; count down
	clr1	t0con,t0lrun		; stop timer
	ret
.keya	call	key_click
	ret

; initialise all keyboard variables and start the interrupt
key_init
	clr1	t0con,t0hie		; disable key interrupt
	clr1	t0con,t0hrun		; stop key timer
	mov	#$ff,key_keys		; no key pressed
	mov	#key_qstart,key_qin	; init queue pointer
	mov	#key_qstart,key_qout	; init queue pointer
	mov	#$80,key_lastevent	; set latest key event to release
	mov	#$00,key_repeat		; don't repeat
	set1	t0con,t0hie		; enable key interrupt
	set1	t0con,t0hrun		; start key timer
	ret

; play the key click
key_click
	mov	#<tune_keyclick,trl
	mov	#>tune_keyclick,trh
	call	snd_playtune		; play keyclick
.sound	ld	snd_playing
	bnz	.sound			; wait until sound finished
	ret

; --- sound engine code starts here -------------------------------------------
; available as seperate package at http://soeren.infoserv.de/vm/sound.html
; this version is enhanced in many ways:
;   it can queue tunes and plays them when the ones before are finished
;   the player is seperated from the irq routine to allow calling in frameloop

; initialise all sound variables and stop all tunes
snd_init
	mov	#0,t1cnt		; stop sound
	mov	#0,snd_playing		; mark sound as stopped
	mov	#snd_qstart,snd_qin	; init sound queue
	mov	#snd_qstart,snd_qout	; init sound queue
	ret

; start to play tune pointed by trh/trl
; trashes acc
snd_playtune
	bn	gameoption,OPTION_SOUND,.ret
	call	snd_queue
	ld	snd_playing		; check for sound already playing
	bnz	.ret
	; start irq for new sound
	mov	#$ff,t1hr		; timer rolls over immediately
	mov	#0,snd_wait		; next event immediately
	mov	#$ff,snd_playing	; we are playing
	mov	#$94,t1cnt		; start timerh, irq, rest is done in irq
.ret	ret

; play tune pointed by trh/trl, for use inside frameloop
; trashes acc
snd_frameplaytune
	bn	gameoption,OPTION_SOUND,.ret
	call	snd_queue
	ld	snd_playing
	bnz	.ret
	; start playing in next frame
	mov	#0,snd_wait		; next event immediately
	mov	#$ff,snd_playing	; we are playing
.ret	ret

; insert sound into queue
; trashes acc
snd_queue
	; test for overflow
	ld	snd_qin
	sub	snd_qout
	bnz	.ok
	; queue is either full or empty, if no sound plays its empty
	ld	snd_playing
	bz	.ok
	; sound queue overflow: show symbol if debugging
	.ifd	DEBUG
	mov	#0,xbnk
	mov	#%01010100,$185
	mov	#%00111000,$18b
	mov	#%01010100,$195
	.endc
	ret
.ok	; insert into queue
	push	0
	ld	snd_qin
	st	0
	ld	trl			; low byte of event pointer
	st	@r0
	inc	0
	ld	trh			; high byte of event pointer
	st	@r0
	pop	0
	; advance queue pointer
	ld	snd_qin
	add	#2
	bne	#snd_qend,.noqend
	mov	#snd_qstart,acc
.noqend	st	snd_qin
	ret

; temporary stop the tune, only for irq driven
snd_suspend
	mov	#0,t1cnt		; stop timer
	mov	#0,snd_playing		; we are no longer playing
	ret

; restart the tune with the next tone, only for irq driven
snd_restart
	mov	#$ff,t1hr		; timer rolls over immediately
	mov	#0,snd_wait		; next event immediately
	mov	#$ff,snd_playing	; we are playing
	mov	#$94,t1cnt		; start timerh irq, rest is done in irq
	ret

; handle irq requests related to playing a tune
snd_irq
	bn	t1cnt,t1hovf,.reti	; skip if not timer1 high ovl
	clr1	t1cnt,t1hovf		; clear overflow, important!
	push	psw			; store registers
	set1	psw,rambk0		; game ram bank
	push	acc			; store registers
	push	0
	push	trl
	push	trh
	call	snd_play		; all work done here
	; set timer period to maximum, may be set to minimum if first sound
	mov	#$00,t1hr		; timer1 high period to maximum
	set1	t1cnt,t1hrun		; start timer1 high (wait)
	clr1	t1cnt,t1hovf		; clear overflow again
	set1	t1cnt,t1hie		; start timer1 irq (wait)
	pop	trh			; restore registers
	pop	trl
	pop	0
	pop	acc
	pop	psw
.reti	reti

; this routine is called once for every frame or from the timer interrupt
; it plays sound events in the soundqueue
; trashes acc, trl, trh, 0
snd_play
	ld	snd_playing
	bz	.ret			; no sound to play
	ld	snd_wait
	bz	.event			; its time for the next sound event
	dec	snd_wait
.ret	ret

.event	; process sound event
	ld	snd_qout		; queue read pointer
	st	0
	ld	@r0
	st	trl			; low byte of event pointer
	inc	0
	ld	@r0
	st	trh			; high byte of event pointer
	mov	#0,acc			; offset for pointer
	ldc				; command byte of sound event
	be	#1,.note		; 1 = set note
	be	#2,.wait		; 2 = wait
	; unknown command or 0, current tune is finished
	ld	snd_qout
	add	#2			; advance queue pointer
	bne	#snd_qend,.noqend
	mov	#snd_qstart,acc		; queue reached end, start again
.noqend	st	snd_qout		; store new queue pointer
	; check if queue empty
	sub	snd_qin
	bnz	.event			; process next sound in queue
	; no next sound in queue, stop playing
	xor	acc
	st	snd_playing		; clear flag
	st	t1cnt			; stop timer
	ret
	; play note
.note	clr1	t1cnt,t1lrun		; stop timer1 low (sound)
	mov	#1,acc			; offset to value for t1lr
	ldc
	st	t1lr
	mov	#2,acc			; offset to value for t1lc
	ldc
	st	t1lc
	set1	t1cnt,t1lrun		; start timer1 low (sound)
	set1	t1cnt,eldt1c		; set compare value
	clr1	t1cnt,t1long		; use 2 8 bit timers
	add16	#3,tr			; advance soundpointer
	; store trl/trh in queue
	ld	snd_qout
	st	0
	ld	trl			; low byte of event pointer
	st	@r0
	inc	0
	ld	trh			; high byte of event pointer
	st	@r0
	br	.event			; note is started, next event

.wait	; wait
	mov	#1,acc			; offset to number of wait cycles
	ldc
	st	snd_wait
	add16	#2,tr			; advance soundpointer
	; store trl/trh in queue
	ld	snd_qout
	st	0
	ld	trl			; low byte of event pointer
	st	@r0
	inc	0
	ld	trh			; high byte of event pointer
	st	@r0
	ret

; --- textfonts ---------------------------------------------------------------

; the normal text font, 5x5 pixel used on normal chars, 6x5 pixel drawn
font5x5
	; 0
	.byte	%01110000
	.byte	%10011000
	.byte	%10101000
	.byte	%11001000
	.byte	%01110000
	; 1
	.byte	%00100000
	.byte	%01100000
	.byte	%00100000
	.byte	%00100000
	.byte	%01110000
	; 2
	.byte	%01110000
	.byte	%10001000
	.byte	%00110000
	.byte	%01000000
	.byte	%11111000
	; 3
	.byte	%01110000
	.byte	%10001000
	.byte	%00110000
	.byte	%10001000
	.byte	%01110000
	; 4
	.byte	%00110000
	.byte	%01010000
	.byte	%10010000
	.byte	%11111000
	.byte	%00010000
	; 5
	.byte	%11111000
	.byte	%10000000
	.byte	%11110000
	.byte	%00001000
	.byte	%11110000
	; 6
	.byte	%01110000
	.byte	%10000000
	.byte	%11110000
	.byte	%10001000
	.byte	%01110000
	; 7
	.byte	%11111000
	.byte	%00001000
	.byte	%00010000
	.byte	%00100000
	.byte	%00100000
	; 8
	.byte	%01110000
	.byte	%10001000
	.byte	%01110000
	.byte	%10001000
	.byte	%01110000
	; 9
	.byte	%01110000
	.byte	%10001000
	.byte	%01111000
	.byte	%00001000
	.byte	%01110000
	; cursor = 10 = :
	.byte	%11000000
	.byte	%11100000
	.byte	%11110000
	.byte	%11100000
	.byte	%11000000
	; . = 11 = ;
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00100000
	; < = 12 = <
	.byte	%00010000
	.byte	%00100000
	.byte	%01000000
	.byte	%00100000
	.byte	%00010000
	; = = 13 = -
	.byte	%00000000
	.byte	%00000000
	.byte	%01110000
	.byte	%00000000
	.byte	%00000000
	; > = 14 = >
	.byte	%01000000
	.byte	%00100000
	.byte	%00010000
	.byte	%00100000
	.byte	%01000000
	; ? = 15 = ?
	.byte	%01110000
	.byte	%10001000
	.byte	%00110000
	.byte	%00000000
	.byte	%00100000
	; [SPACE] = 16
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	; A = 17
	.byte	%01110000
	.byte	%10001000
	.byte	%11111000
	.byte	%10001000
	.byte	%10001000
	; B = 18
	.byte	%11110000
	.byte	%10001000
	.byte	%11110000
	.byte	%10001000
	.byte	%11110000
	; C
	.byte	%01111000
	.byte	%10000000
	.byte	%10000000
	.byte	%10000000
	.byte	%01111000
	; D
	.byte	%11110000
	.byte	%10001000
	.byte	%10001000
	.byte	%10001000
	.byte	%11110000
	; E
	.byte	%11111000
	.byte	%10000000
	.byte	%11100000
	.byte	%10000000
	.byte	%11111000
	; F
	.byte	%11111000
	.byte	%10000000
	.byte	%11100000
	.byte	%10000000
	.byte	%10000000
	; G
	.byte	%01111000
	.byte	%10000000
	.byte	%10111000
	.byte	%10001000
	.byte	%01110000
	; H
	.byte	%10001000
	.byte	%10001000
	.byte	%11111000
	.byte	%10001000
	.byte	%10001000
	; I
	.byte	%11111000
	.byte	%00100000
	.byte	%00100000
	.byte	%00100000
	.byte	%11111000
	; J
	.byte	%11111000
	.byte	%00001000
	.byte	%00001000
	.byte	%10001000
	.byte	%01110000
	; K
	.byte	%10001000
	.byte	%10010000
	.byte	%11100000
	.byte	%10010000
	.byte	%10001000
	; L
	.byte	%10000000
	.byte	%10000000
	.byte	%10000000
	.byte	%10000000
	.byte	%11111000
	; M
	.byte	%10001000
	.byte	%11011000
	.byte	%10101000
	.byte	%10001000
	.byte	%10001000
	; N
	.byte	%10001000
	.byte	%11001000
	.byte	%10101000
	.byte	%10011000
	.byte	%10001000
	; O
	.byte	%01110000
	.byte	%10001000
	.byte	%10001000
	.byte	%10001000
	.byte	%01110000
	; P
	.byte	%11110000
	.byte	%10001000
	.byte	%11110000
	.byte	%10000000
	.byte	%10000000
	; Q
	.byte	%01110000
	.byte	%10001000
	.byte	%10101000
	.byte	%10010000
	.byte	%01101000
	; R
	.byte	%11110000
	.byte	%10001000
	.byte	%11110000
	.byte	%10010000
	.byte	%10001000
	; S
	.byte	%01111000
	.byte	%10000000
	.byte	%01110000
	.byte	%00001000
	.byte	%11110000
	; T
	.byte	%11111000
	.byte	%00100000
	.byte	%00100000
	.byte	%00100000
	.byte	%00100000
	; U
	.byte	%10001000
	.byte	%10001000
	.byte	%10001000
	.byte	%10001000
	.byte	%01110000
	; V
	.byte	%10001000
	.byte	%10001000
	.byte	%10001000
	.byte	%01010000
	.byte	%00100000
	; W
	.byte	%10001000
	.byte	%10001000
	.byte	%10001000
	.byte	%10101000
	.byte	%01010000
	; X
	.byte	%10001000
	.byte	%01010000
	.byte	%00100000
	.byte	%01010000
	.byte	%10001000
	; Y
	.byte	%10001000
	.byte	%01010000
	.byte	%00100000
	.byte	%00100000
	.byte	%00100000
	; Z
	.byte	%11111000
	.byte	%00010000
	.byte	%00100000
	.byte	%01000000
	.byte	%11111000

; just a line used as cursor and a space to delete the cursor
font5x1
	.byte	%11111000

	.byte	%00000000

; --- prep routines for the aliens --------------------------------------------

; NOTE:
; These routines are called from alien_draw and alien_kill to get information
; about the bitmap data of the current alien in 0. Those who have more than
; one frame are optimized into unreadability, just look at what the comments
; say, divide the frame counter by "time", and out the bits needed for "frames"
; frames and multiply by "size". Do the divisions and multiplications with
; ror and rol if possible. Now optimize by removing the rotations pair wise
; while adapting the and. You should get to the same code as I did, otherwise
; one of us has made a mistake.

; table with pointers to individual routines
alienpreptable
	.word	prp_shot
	.word	prp_asteroid
	.word	prp_tiny
	.word	prp_small
	.word	prp_small2
	.word	prp_small3
	.word	prp_dia1
	.word	prp_dia2
	.word	prp_dia3
	.word	prp_dia4
	.word	prp_tentacle
	.word	prp_cannonl
	.word	prp_cannonr
	.word	prp_cannonu
	.word	prp_cannondl
	.word	prp_cannondr
	.word	prp_cannondul
	.word	prp_cannondur
	.word	prp_bomberl
	.word	prp_bomberlf
	.word	prp_bomberr
	.word	prp_bomberrf
	.word	prp_bomb
	.word	prp_bomb2
	.word	prp_grid1
	.word	prp_mother
	.word	prp_boss1b
	.word	prp_boss1l
	.word	prp_boss1r
	.word	prp_boss1m
	.word	prp_boss1f
	.word	prp_smile
	.word	prp_single
	.word	prp_double
	.word	prp_block
	.word	prp_expl1

; 1 frame ALN_SHOT
prp_shot
	mov	#<fnt_shot,trl
	mov	#>fnt_shot,trh
	mov	#2,width
	mov	#2,height
	mov	#0,acc
	ret

; 4-frame asteroid ALN_ASTEROID
prp_asteroid
	add_st	0,#ALNOFF_FRAME,1
	ld	@r1			; get frame
	ror				; / 2
	and	#3			; only 4 frames
	; convert char number to offset
	st	b
	mov	#6,c
	xor	acc
	mul	
	ld	c
	mov	#<fnt_asteroid,trl
	mov	#>fnt_asteroid,trh
	mov	#6,width
	mov	#6,height
	ret

; 1 frame ALN_TINY
prp_tiny
	mov	#<fnt_tiny,trl
	mov	#>fnt_tiny,trh
	mov	#4,width
	mov	#2,height
	mov	#0,acc
	ret

; 2-frame ALN_SMALL
prp_small
	add_st	0,#ALNOFF_FRAME,1
	ld	@r1			; get frame
	ror				; / 2
	and	#1			; only 2 frames
	; convert char number to offset
	st	b
	mov	#6,c
	xor	acc
	mul	
	ld	c
	mov	#<fnt_small,trl
	mov	#>fnt_small,trh
	mov	#6,width
	mov	#6,height
	ret

; 2-frame ALN_SMALL2
prp_small2
	add_st	0,#ALNOFF_FRAME,1
	ld	@r1			; get frame
	ror				; / 2
	and	#1			; only 2 frames
	; convert char number to offset
	st	b
	mov	#8,c
	xor	acc
	mul	
	ld	c
	mov	#<fnt_small2,trl
	mov	#>fnt_small2,trh
	mov	#6,width
	mov	#8,height
	ret

; 4-frame ALN_SMALL
prp_small3
	add_st	0,#ALNOFF_FRAME,1
	ld	@r1			; get frame
	ror				; / 2
	and	#$3
	; convert char number to offset
	st	b
	mov	#7,c
	xor	acc
	mul	
	ld	c
	mov	#<fnt_small3,trl
	mov	#>fnt_small3,trh
	mov	#6,width
	mov	#7,height
	ret

; 2-frame ALN_TENTACLE
prp_tentacle
	add_st	0,#ALNOFF_FRAME,1
	ld	@r1			; get frame
	and	#$4			; 2 frames, time 4, size 4
	mov	#<fnt_tentacle,trl
	mov	#>fnt_tentacle,trh
	mov	#8,width
	mov	#4,height
	ret

; 1 frame ALN_SMILE
prp_smile
	mov	#<fnt_smile,trl
	mov	#>fnt_smile,trh
	mov	#8,width
	mov	#6,height
	mov	#0,acc
	ret

; 2-frame ALN_SINGLE
prp_single
	add_st	0,#ALNOFF_FRAME,1
	ld	@r1			; get frame
	ror				; / 2
	and	#1			; only 2 frames
	; convert char number to offset
	st	b
	mov	#6,c
	xor	acc
	mul	
	ld	c
	mov	#<fnt_single,trl
	mov	#>fnt_single,trh
	mov	#8,width
	mov	#6,height
	ret

; 2-frame ALN_DOUBLE
prp_double
	add_st	0,#ALNOFF_FRAME,1
	ld	@r1			; get frame
	ror				; / 2
	and	#1			; only 2 frames
	; convert char number to offset
	st	b
	mov	#8,c
	xor	acc
	mul	
	ld	c
	mov	#<fnt_double,trl
	mov	#>fnt_double,trh
	mov	#8,width
	mov	#8,height
	ret

; 1 frame ALN_BLOCK
prp_block
	mov	#<fnt_block,trl
	mov	#>fnt_block,trh
	mov	#8,width
	mov	#8,height
	mov	#0,acc
	ret

; 1 frame ALN_BOSS1B
prp_boss1b
	mov	#<fnt_boss1b,trl
	mov	#>fnt_boss1b,trh
	mov	#5,width
	mov	#8,height
	mov	#0,acc
	ret

; 1 frame ALN_BOSS1L
prp_boss1l
	mov	#<fnt_boss1l,trl
	mov	#>fnt_boss1l,trh
	mov	#8,width
	mov	#7,height
	mov	#0,acc
	ret

; 1 frame ALN_BOSS1R
prp_boss1r
	mov	#<fnt_boss1r,trl
	mov	#>fnt_boss1r,trh
	mov	#8,width
	mov	#7,height
	mov	#0,acc
	ret

; 1 frame ALN_BOSS1M
prp_boss1m
	mov	#<fnt_boss1m,trl
	mov	#>fnt_boss1m,trh
	mov	#5,width
	mov	#7,height
	mov	#0,acc
	ret

; 1 frame ALN_BOSS1F
prp_boss1f
	mov	#<fnt_boss1f,trl
	mov	#>fnt_boss1f,trh
	mov	#7,width
	mov	#6,height
	mov	#0,acc
	ret

; 4-frame ALN_EXPLOSION1
prp_expl1
	add_st	0,#ALNOFF_FRAME,1
	ld	@r1			; get frame
	rol
	and	#$18			; 4 frames, size 8, time 4
	mov	#<fnt_expl1,trl
	mov	#>fnt_expl1,trh
	mov	#8,width
	mov	#8,height
	ret

; 1 frame ALN_CANNONL
prp_cannonl
	mov	#<fnt_cannonl,trl
	mov	#>fnt_cannonl,trh
	mov	#5,width
	mov	#4,height
	mov	#0,acc
	ret

; 1 frame ALN_CANNONR
prp_cannonr
	mov	#<fnt_cannonr,trl
	mov	#>fnt_cannonr,trh
	mov	#5,width
	mov	#4,height
	mov	#0,acc
	ret

; 1 frame ALN_CANNONU
prp_cannonu
	mov	#<fnt_cannonu,trl
	mov	#>fnt_cannonu,trh
	mov	#4,width
	mov	#5,height
	mov	#0,acc
	ret

; 1 frame ALN_CANNONDL
prp_cannondl
	mov	#<fnt_cannondl,trl
	mov	#>fnt_cannondl,trh
	mov	#3,width
	mov	#4,height
	mov	#0,acc
	ret

; 1 frame ALN_CANNONDR
prp_cannondr
	mov	#<fnt_cannondr,trl
	mov	#>fnt_cannondr,trh
	mov	#3,width
	mov	#4,height
	mov	#0,acc
	ret

; 1 frame ALN_CANNONDUL
prp_cannondul
	mov	#<fnt_cannondul,trl
	mov	#>fnt_cannondul,trh
	mov	#4,width
	mov	#4,height
	mov	#0,acc
	ret

; 1 frame ALN_CANNONDUR
prp_cannondur
	mov	#<fnt_cannondur,trl
	mov	#>fnt_cannondur,trh
	mov	#4,width
	mov	#4,height
	mov	#0,acc
	ret

; 1 frame ALN_BOMBERL
prp_bomberl
	mov	#<fnt_bomberl,trl
	mov	#>fnt_bomberl,trh
	mov	#8,height
	mov	#8,width
	mov	#0,acc
	ret

; 6 frame ALN_BOMBERLF fire animation
prp_bomberlf
	mov	#<fnt_bomberl,trl
	mov	#>fnt_bomberl,trh
	mov	#8,height
	mov	#8,width
	add_st	0,#ALNOFF_FRAME,1
	ld	@r1			; get frame
	sub	#6
	bn	psw,cy,.first
	ld	@r1			; frame 0-5 show firing anim
	rol
	rol
	rol				; 6 frames once, size 8, time 1
	ret
.first	mov	#0,acc			; frame >6, show inactive bitmap
	ret

; 1 frame ALN_BOMBERR
prp_bomberr
	mov	#<fnt_bomberr,trl
	mov	#>fnt_bomberr,trh
	mov	#8,height
	mov	#8,width
	mov	#0,acc
	ret

; 6 frame ALN_BOMBERRF fire animation
prp_bomberrf
	mov	#<fnt_bomberr,trl
	mov	#>fnt_bomberr,trh
	mov	#8,height
	mov	#8,width
	add_st	0,#ALNOFF_FRAME,1
	ld	@r1			; get frame
	sub	#6
	bn	psw,cy,.first
	ld	@r1			; frame 0-5 show firing anim
	; convert char number to offset
	rol
	rol
	rol				; 6 frames once, size 8, time 1
	ret
.first	mov	#0,acc			; frame >6, show inactive bitmap
	ret

; 1 frame ALN_BOMB
prp_bomb
	mov	#<fnt_bomb,trl
	mov	#>fnt_bomb,trh
	mov	#3,width
	mov	#3,height
	mov	#0,acc
	ret

; 1 frame ALN_BOMB
prp_bomb2
	mov	#<fnt_bomb2,trl
	mov	#>fnt_bomb2,trh
	mov	#4,width
	mov	#4,height
	mov	#0,acc
	ret

; 1 frame GRID1
prp_grid1
	mov	#<fnt_grid1,trl
	mov	#>fnt_grid1,trh
	mov	#5,height
	mov	#8,width
	mov	#0,acc
	ret

; 1 frame ALN_DIA1
prp_dia1
	mov	#<fnt_dia1,trl
	mov	#>fnt_dia1,trh
	mov	#5,width
	mov	#5,height
	mov	#0,acc
	ret

; 1 frame ALN_DIA2
prp_dia2
	mov	#<fnt_dia2,trl
	mov	#>fnt_dia2,trh
	mov	#5,width
	mov	#5,height
	mov	#0,acc
	ret

; 1 frame ALN_DIA3
prp_dia3
	mov	#<fnt_dia3,trl
	mov	#>fnt_dia3,trh
	mov	#5,width
	mov	#5,height
	mov	#0,acc
	ret

; 1 frame ALN_DIA4
prp_dia4
	mov	#<fnt_dia4,trl
	mov	#>fnt_dia4,trh
	mov	#5,width
	mov	#5,height
	mov	#0,acc
	ret

; 2 frame ALN_MOTHER
prp_mother
	add_st	0,#ALNOFF_FRAME,1
	ld	@r1			; get frame
	ror				; / 2
	and	#1			; only 2 frames
	; convert char number to offset
	st	b
	mov	#5,c
	xor	acc
	mul	
	ld	c
	mov	#<fnt_mother,trl
	mov	#>fnt_mother,trh
	mov	#8,width
	mov	#5,height
	ret

; --- fonts for the aliens ----------------------------------------------------

fnt_shot	; ALN_SHOT
	.byte	%11000000
	.byte	%11000000

fnt_cannondl	; ALN_CANNONDL
	.byte	%10000000
	.byte	%11000000
	.byte	%11100000
	.byte	%01000000

fnt_cannondr	; ALN_CANNONDR
	.byte	%00100000
	.byte	%01100000
	.byte	%11100000
	.byte	%01000000

fnt_tiny	; ALN_TINY
	.byte	%01100000
	.byte	%11110000

fnt_cannondul	; ALN_CANNONDUL
	.byte	%11000000
	.byte	%11100000
	.byte	%01110000
	.byte	%00100000

fnt_cannondur	; ALN_CANNONDUR
	.byte	%00110000
	.byte	%01110000
	.byte	%11100000
	.byte	%01000000

fnt_cannonu	; ALN_CANNONU
	.byte	%11110000
	.byte	%11110000
	.byte	%11110000
	.byte	%01100000
	.byte	%01100000

fnt_cannonl	; ALN_CANNONL
	.byte	%11100000
	.byte	%11111000
	.byte	%11111000
	.byte	%11100000

fnt_cannonr	; ALN_CANNONR
	.byte	%00111000
	.byte	%11111000
	.byte	%11111000
	.byte	%00111000

fnt_bomberl	; ALN_BOMBERL frame 0
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00010000
	.byte	%00111000
	.byte	%01110000
	.byte	%11100000
	.byte	%11000000
		; ALN_BOMBERL frame 1
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00010000
	.byte	%00111000
	.byte	%11110000
	.byte	%11100000
	.byte	%11100000
		; ALN_BOMBERL frame 2
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00010000
	.byte	%01111000
	.byte	%01110000
	.byte	%11110000
	.byte	%11000000
		; ALN_BOMBERL frame 3
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00111000
	.byte	%00111000
	.byte	%01111000
	.byte	%11100000
	.byte	%11000000
		; ALN_BOMBERL frame 4
	.byte	%00000000
	.byte	%00000000
	.byte	%00011100
	.byte	%00011100
	.byte	%00111100
	.byte	%01110000
	.byte	%11100000
	.byte	%11000000
		; ALN_BOMBERL frame 5
	.byte	%00000000
	.byte	%00001110
	.byte	%00001110
	.byte	%00011110
	.byte	%00111000
	.byte	%01110000
	.byte	%11100000
	.byte	%11000000

fnt_bomberr	; ALN_BOMBERR frame 0
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00001000
	.byte	%00011100
	.byte	%00001110
	.byte	%00000111
	.byte	%00000011
		; ALN_BOMBERR frame 1
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00001000
	.byte	%00011100
	.byte	%00001111
	.byte	%00000111
	.byte	%00000111
		; ALN_BOMBERR frame 2
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00001000
	.byte	%00011110
	.byte	%00001110
	.byte	%00001111
	.byte	%00000011
		; ALN_BOMBERR frame 3
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00011100
	.byte	%00011100
	.byte	%00011110
	.byte	%00000111
	.byte	%00000011
		; ALN_BOMBERR frame 4
	.byte	%00000000
	.byte	%00000000
	.byte	%00111000
	.byte	%00111000
	.byte	%00111100
	.byte	%00001110
	.byte	%00000111
	.byte	%00000011
		; ALN_BOMBERR frame 5
	.byte	%00000000
	.byte	%01110000
	.byte	%01110000
	.byte	%01111000
	.byte	%00011100
	.byte	%00001110
	.byte	%00000111
	.byte	%00000011

fnt_bomb	; ALN_BOMB
	.byte	%11100000
	.byte	%11100000
	.byte	%11100000

fnt_bomb2	; ALN_BOMB2
	.byte	%01100000
	.byte	%11110000
	.byte	%11110000
	.byte	%01100000

fnt_dia1	; ALN_DIA1
	.byte	%11000000
	.byte	%01100000
	.byte	%01110000
	.byte	%11111000
	.byte	%11001000

fnt_dia2	; ALN_DIA2
	.byte	%00011000
	.byte	%00110000
	.byte	%01110000
	.byte	%11111000
	.byte	%10011000

fnt_dia3	; ALN_DIA3
	.byte	%10011000
	.byte	%11111000
	.byte	%01110000
	.byte	%00110000
	.byte	%00011000

fnt_dia4	; ALN_DIA4
	.byte	%11001000
	.byte	%11111000
	.byte	%01110000
	.byte	%01100000
	.byte	%11000000

fnt_smile	; ALN_SMILE frame0
	.byte	%00100100
	.byte	%00111100
	.byte	%01011010
	.byte	%11111111
	.byte	%10100101
	.byte	%00011000

fnt_boss1b	; ALN_BOSS1B frame0
	.byte	%01110000
	.byte	%11111000
	.byte	%11111000
	.byte	%00100000
	.byte	%00100000
	.byte	%00100000
	.byte	%00100000
	.byte	%00100000

fnt_boss1l	; ALN_BOSS1L frame0
	.byte	%01100000
	.byte	%11110000
	.byte	%11111111
	.byte	%11110000
	.byte	%01100000
	.byte	%01100000
	.byte	%01100000

fnt_boss1m	; ALN_BOSS1M frame0
	.byte	%00100000
	.byte	%00100000
	.byte	%01110000
	.byte	%11111000
	.byte	%01110000
	.byte	%00100000
	.byte	%00100000

fnt_boss1r	; ALN_BOSS1R frame0
	.byte	%00000110
	.byte	%00001111
	.byte	%11111111
	.byte	%00001111
	.byte	%00000110
	.byte	%00000110
	.byte	%00000110

fnt_boss1f	; ALN_BOSS1F frame0
	.byte	%00010000
	.byte	%00111000
	.byte	%01111100
	.byte	%01111100
	.byte	%11111110
	.byte	%01000100

fnt_asteroid	; ALN_ASTEROID frame 0
	.byte	%01110000
	.byte	%11111100
	.byte	%11111100
	.byte	%01111100
	.byte	%01111000
	.byte	%00011000
		; ALN_ASTEROID frame 1
	.byte	%00011000
	.byte	%01111100
	.byte	%01111100
	.byte	%11111100
	.byte	%11111000
	.byte	%00111000
		; ALN_ASTEROID frame 2
	.byte	%01100000
	.byte	%01111000
	.byte	%11111000
	.byte	%11111100
	.byte	%11111100
	.byte	%00111000
		; ALN_ASTEROID frame 3
	.byte	%01110000
	.byte	%01111100
	.byte	%11111100
	.byte	%11111000
	.byte	%11111000
	.byte	%01100000

fnt_small	; ALN_SMALL frame 0
	.byte	%00110000
	.byte	%01111000
	.byte	%11111100
	.byte	%11111100
	.byte	%01001000
	.byte	%10010000
		; ALN_SMALL frame 1
	.byte	%00110000
	.byte	%01111000
	.byte	%11111100
	.byte	%11111100
	.byte	%01001000
	.byte	%00100100

fnt_small2	; ALN_SMALL2 frame 0
	.byte	%10000100
	.byte	%01001000
	.byte	%00110000
	.byte	%01111000
	.byte	%11111100
	.byte	%01111000
	.byte	%01001000
	.byte	%10010000
		; ALN_SMALL2 frame 1
	.byte	%00110000
	.byte	%01001000
	.byte	%00110000
	.byte	%01111000
	.byte	%11111100
	.byte	%01111000
	.byte	%01001000
	.byte	%00100100

fnt_small3	; ALN_SMALL3 frame 0
	.byte	%10000100
	.byte	%01001000
	.byte	%00110000
	.byte	%11111100
	.byte	%10110100
	.byte	%11111100
	.byte	%01111000
		; ALN_SMALL3 frame 1
	.byte	%00000000
	.byte	%11001100
	.byte	%00110000
	.byte	%11111100
	.byte	%10110100
	.byte	%11111100
	.byte	%01111000
		; ALN_SMALL3 frame 2
	.byte	%10000100
	.byte	%01001000
	.byte	%00110000
	.byte	%11111100
	.byte	%10110100
	.byte	%11111100
	.byte	%01111000
		; ALN_SMALL3 frame 3
	.byte	%00000000
	.byte	%11001100
	.byte	%00110000
	.byte	%11111100
	.byte	%10110100
	.byte	%11111100
	.byte	%01111000

fnt_single	; ALN_SINGLE frame 0
	.byte	%10111101
	.byte	%11110111
	.byte	%01101110
	.byte	%00111100
	.byte	%00011000
	.byte	%00011000
		; ALN_SINGLE frame 1
	.byte	%10111101
	.byte	%11101111
	.byte	%01110110
	.byte	%00111100
	.byte	%00011000
	.byte	%00011000

fnt_double	; ALN_DOUBLE frame 0
	.byte	%01100110
	.byte	%10011001
	.byte	%10101101
	.byte	%01110110
	.byte	%00111100
	.byte	%01111110
	.byte	%11100111
	.byte	%01000010
		; ALN_DOUBLE frame 1
	.byte	%01100110
	.byte	%10011001
	.byte	%10110101
	.byte	%01101110
	.byte	%00111100
	.byte	%01111110
	.byte	%11100111
	.byte	%01000010

fnt_block	; ALN_BLOCK frame 0
	.byte	%11111111
	.byte	%11111111
	.byte	%11111111
	.byte	%11111111
	.byte	%11111111
	.byte	%11111111
	.byte	%11111111
	.byte	%11111111

fnt_tentacle	; ALN_TENTACLE frame 0
	.byte	%00011000
	.byte	%10111100
	.byte	%01111110
	.byte	%00100101
		; ALN_TENTACLE frame 1
	.byte	%00011000
	.byte	%00111101
	.byte	%01111110
	.byte	%10100100

fnt_mother	; ALN_MOTHER frame 0
	.byte	%00100100
	.byte	%01111110
	.byte	%11011111
	.byte	%11111011
	.byte	%01111110
		; ALN_MOTHER frame 1
	.byte	%00100100
	.byte	%01111110
	.byte	%11111011
	.byte	%11011111
	.byte	%01111110

fnt_grid1	; ALN_GRID1
	.byte	%11111111
	.byte	%10101010
	.byte	%01010101
	.byte	%10101010
	.byte	%11111111

fnt_expl1	; explosion frame 0
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00101000
	.byte	%00010000
	.byte	%00101000
	.byte	%00000000
	.byte	%00000000
		; explosion frame 1
	.byte	%00000000
	.byte	%00000000
	.byte	%01001000
	.byte	%00100100
	.byte	%01010000
	.byte	%00000100
	.byte	%01010000
	.byte	%00000000
		; explosion frame 2
	.byte	%10000010
	.byte	%00101000
	.byte	%01000100
	.byte	%00010010
	.byte	%01001000
	.byte	%00010010
	.byte	%10000000
	.byte	%00100100
		; explosion frame 3
	.byte	%10000010
	.byte	%00100000
	.byte	%00000000
	.byte	%00000010
	.byte	%01000000
	.byte	%00000010
	.byte	%10000000
	.byte	%00100100

fnt_ship	; defense ship
	.byte	%01100000
	.byte	%11110000
	.byte	%10010000

; --- leveldata level 0 -------------------------------------------------------

; NOTE:
; The level data was converted with the included ppmtolevel.pl script.

bm_level0
	.include "level0.i"

; levelscripts: first appearance of alien
; levelline, type, x, y, hits, alienscript
; levelline = 984 - pixelline
script_level0
	; first some cannons
	levelalien 6,ALN_CANNONL,0,4,1,scr_cannonl
	levelalien 26,ALN_CANNONR,39,4,1,scr_cannonr
	levelalien 46,ALN_CANNONL,3,4,1,scr_cannonl
	levelalien 66,ALN_CANNONR,39,4,1,scr_cannonr
	; several asteroids
	levelalien 97,ALN_ASTEROID,19,2,1,scr_asteroid1
	levelalien 112,ALN_ASTEROID,19,2,1,scr_asteroid1
	levelalien 123,ALN_ASTEROID,12,2,1,scr_asteroid1
	levelalien 137,ALN_ASTEROID,30,2,1,scr_asteroid1
	levelalien 152,ALN_ASTEROID,12,2,1,scr_asteroid1
	levelalien 167,ALN_ASTEROID,30,2,1,scr_asteroid1
	; cannons on first split
	levelalien 178,ALN_CANNONR,17,4,1,scr_cannonr
	levelalien 178,ALN_CANNONL,29,4,1,scr_cannonl
	levelalien 188,ALN_CANNONR,16,4,1,scr_cannonr
	levelalien 188,ALN_CANNONL,29,4,1,scr_cannonl
	levelalien 198,ALN_CANNONR,16,4,1,scr_cannonr
	levelalien 198,ALN_CANNONL,26,4,1,scr_cannonl
	levelalien 208,ALN_CANNONR,15,4,1,scr_cannonr
	levelalien 208,ALN_CANNONL,27,4,1,scr_cannonl
	; more cannons
	levelalien 236,ALN_CANNONL,2,4,1,scr_cannonl
	levelalien 252,ALN_CANNONL,2,4,1,scr_cannonl
	; another wave of asteroids
	levelalien 282,ALN_ASTEROID,7,2,1,scr_asteroid1
	levelalien 297,ALN_ASTEROID,4,2,1,scr_asteroid1
	levelalien 312,ALN_ASTEROID,4,2,1,scr_asteroid1
	; 3 small aliens
	levelalien 336,ALN_SMALL,38,8,3,scr_alien1
	levelalien 350,ALN_SMALL,38,8,3,scr_alien2
	levelalien 364,ALN_SMALL,38,8,3,scr_alien3
	; asteroids again
	levelalien 467,ALN_ASTEROID,38,2,1,scr_asteroid1
	levelalien 487,ALN_ASTEROID,38,2,1,scr_asteroid1
	levelalien 507,ALN_ASTEROID,38,2,1,scr_asteroid1
	levelalien 517,ALN_ASTEROID,36,2,1,scr_asteroid1
	levelalien 526,ALN_ASTEROID,32,2,1,scr_asteroid2
	levelalien 527,ALN_ASTEROID,39,2,1,scr_asteroid1
	levelalien 536,ALN_ASTEROID,31,2,1,scr_asteroid2
	levelalien 537,ALN_ASTEROID,38,2,1,scr_asteroid1
	levelalien 546,ALN_ASTEROID,29,2,1,scr_asteroid2
	levelalien 547,ALN_ASTEROID,36,2,1,scr_asteroid1
	levelalien 556,ALN_ASTEROID,24,2,1,scr_asteroid2
	levelalien 557,ALN_ASTEROID,32,2,1,scr_asteroid1
	levelalien 566,ALN_ASTEROID,22,2,1,scr_asteroid2
	levelalien 567,ALN_ASTEROID,31,2,1,scr_asteroid1
	levelalien 576,ALN_ASTEROID,20,2,1,scr_asteroid2
	levelalien 577,ALN_ASTEROID,28,2,1,scr_asteroid1
	levelalien 586,ALN_ASTEROID,18,2,1,scr_asteroid2
	levelalien 587,ALN_ASTEROID,27,2,1,scr_asteroid1
	levelalien 596,ALN_ASTEROID,17,2,1,scr_asteroid2
	levelalien 597,ALN_ASTEROID,25,2,1,scr_asteroid1
	levelalien 606,ALN_ASTEROID,16,2,1,scr_asteroid2
	levelalien 607,ALN_ASTEROID,23,2,1,scr_asteroid1
	; cannons before split
	levelalien 634,ALN_CANNONL,2,4,1,scr_cannonl
	levelalien 654,ALN_CANNONR,43,4,1,scr_cannonr
	levelalien 673,ALN_CANNONL,4,4,1,scr_cannonl
	levelalien 688,ALN_CANNONR,41,4,1,scr_cannonr
	levelalien 703,ALN_CANNONL,3,4,1,scr_cannonl
	; cannons on second split
	levelalien 728,ALN_CANNONR,18,4,1,scr_cannonr
	levelalien 743,ALN_CANNONL,28,4,1,scr_cannonl
	levelalien 761,ALN_CANNONR,16,4,1,scr_cannonr
	; another 3 small aliens
	levelalien 812,ALN_SMALL,38,8,3,scr_alien1
	levelalien 826,ALN_SMALL,38,8,3,scr_alien2
	levelalien 840,ALN_SMALL,38,8,3,scr_alien3
	; cannons in end game
	levelalien 978,ALN_CANNONDL,2,4,1,scr_cannondl
	levelalien 978,ALN_CANNONDR,43,4,1,scr_cannondr
	levelalien 989,ALN_CANNONU,9,5,1,scr_cannonu
	levelalien 989,ALN_CANNONU,36,5,1,scr_cannonu
	; diagonal fighters in end game
	levelalien 992,ALN_DIA1,43,8,3,scr_diafighterr
	levelalien 992,ALN_DIA2,0,8,3,scr_diafighterl

	.word	$ffff

; --- leveldata level 1 -------------------------------------------------------

bm_level1
	.include "level1.i"

; levelscripts: first appearance of alien
; levelline, type, x, y, hits, alienscript
; levelline = 984 - pixelline
script_level1
	; cannons on first mast
	levelalien 1,ALN_CANNONDUL,3,8,1,scr_cannondul
	levelalien 1,ALN_CANNONDUR,41,8,1,scr_cannondur
	; shootable grid mast
	levelalien 1,ALN_GRID1,2,3,1,scr_null
	levelalien 1,ALN_GRID1,10,3,1,scr_null
	levelalien 1,ALN_GRID1,30,3,1,scr_null
	levelalien 1,ALN_GRID1,38,3,1,scr_null
	; cannons on middle grid
	levelalien 12,ALN_CANNONR,17,0,1,scr_cannonr
	levelalien 12,ALN_CANNONL,26,0,1,scr_cannonl
	levelalien 22,ALN_CANNONR,17,0,1,scr_cannonr
	levelalien 22,ALN_CANNONL,26,0,1,scr_cannonl
	levelalien 32,ALN_CANNONR,17,0,1,scr_cannonr
	levelalien 32,ALN_CANNONL,26,0,1,scr_cannonl
	levelalien 42,ALN_CANNONR,17,0,1,scr_cannonr
	levelalien 42,ALN_CANNONL,26,0,1,scr_cannonl
	levelalien 52,ALN_CANNONR,17,0,1,scr_cannonr
	levelalien 52,ALN_CANNONL,26,0,1,scr_cannonl
	; shootable grid mast
	levelalien 74,ALN_GRID1,2,0,1,scr_null
	levelalien 74,ALN_GRID1,10,0,1,scr_null
	levelalien 74,ALN_GRID1,30,0,1,scr_null
	levelalien 74,ALN_GRID1,38,0,1,scr_null
	; mothership attack
	levelalien 90,ALN_MOTHER,0,8,10,scr_mother1
	levelalien 97,ALN_CANNONDL,3,0,1,scr_cannondl
	levelalien 97,ALN_CANNONDR,42,0,1,scr_cannondr
	levelalien 107,ALN_CANNONDL,3,0,1,scr_cannondl
	levelalien 107,ALN_CANNONDR,42,0,1,scr_cannondr
	levelalien 117,ALN_CANNONDL,3,0,1,scr_cannondl
	levelalien 117,ALN_CANNONDR,42,0,1,scr_cannondr
	; cannons on grids + bomb thrower
	levelalien 270,ALN_CANNONR,30,0,1,scr_cannonr
	levelalien 275,ALN_CANNONDUL,3,0,1,scr_cannondul
	levelalien 280,ALN_CANNONL,13,0,1,scr_cannonl
	levelalien 288,ALN_BOMBERL,3,0,1,scr_bomberl
	levelalien 291,ALN_CANNONR,30,0,1,scr_cannonr
	; tentacle alien throwing bombs
	levelalien 330,ALN_TENTACLE,0,8,5,scr_tentacle
	levelalien 330+16,ALN_TENTACLE,0,8,5,scr_tentacle
	levelalien 330+32,ALN_TENTACLE,0,8,5,scr_tentacle
	levelalien 330+48,ALN_TENTACLE,0,8,5,scr_tentacle
	; shootable grid
	levelalien 424,ALN_GRID1,4,0,1,scr_null
	levelalien 424,ALN_GRID1,20,0,1,scr_null
	levelalien 424,ALN_GRID1,36,0,1,scr_null
	; cannons in grid 3 way, easyest right
	levelalien 468,ALN_CANNONL,3,0,1,scr_cannonl
	levelalien 468,ALN_CANNONR,40,0,1,scr_cannonr
	levelalien 509,ALN_CANNONU,6,0,1,scr_cannonu
	levelalien 509,ALN_CANNONU,19,0,1,scr_cannonu
	levelalien 509,ALN_CANNONU,25,0,1,scr_cannonu
	; shootable grid
	levelalien 514,ALN_GRID1,4,0,1,scr_null
	levelalien 514,ALN_GRID1,20,0,1,scr_null
	levelalien 514,ALN_GRID1,36,0,1,scr_null
	; mothership attack + cannons
	levelalien 529,ALN_MOTHER,0,8,10,scr_mother1
	levelalien 536,ALN_CANNONDL,3,0,1,scr_cannondl
	levelalien 536,ALN_CANNONDR,42,0,1,scr_cannondr
	levelalien 546,ALN_CANNONDL,3,0,1,scr_cannondl
	levelalien 546,ALN_CANNONDR,42,0,1,scr_cannondr
	levelalien 556,ALN_CANNONDL,3,0,1,scr_cannondl
	levelalien 556,ALN_CANNONDR,42,0,1,scr_cannondr
	; cannons in grid laby, perfect hit at obvious flying route
	levelalien 689,ALN_CANNONDUR,41,0,1,scr_cannondur
	levelalien 729,ALN_CANNONDUL,3,0,1,scr_cannondul
	; bomb thrower
	levelalien 742,ALN_BOMBERR,38,0,1,scr_bomberr
	; small aliens
	levelalien 790,ALN_SMALL2,47,8,3,scr_small2
	levelalien 811,ALN_SMALL2,47,8,3,scr_small2
	levelalien 832,ALN_SMALL2,47,8,3,scr_small2
	levelalien 853,ALN_SMALL2,47,8,3,scr_small2
	levelalien 874,ALN_SMALL2,47,8,3,scr_small2
	levelalien 895,ALN_SMALL2,47,8,3,scr_small2
	levelalien 916,ALN_SMALL2,47,8,3,scr_small2
	levelalien 937,ALN_SMALL2,47,8,3,scr_small2
	; level boss
	levelalien 993,ALN_BOSS1F,4,0,10,scr_boss1f
	levelalien 1000,ALN_BOSS1L,0,1,10,scr_boss1l
	levelalien 1000,ALN_BOSS1M,5,0,15,scr_boss1m
	levelalien 1000,ALN_BOSS1R,7,1,10,scr_boss1r
	levelalien 1004,ALN_BOSS1B,5,0,1,scr_boss1b
	.word	$ffff

; --- leveldata level 2 -------------------------------------------------------

bm_level2
	.include "level2.i"

; levelscripts: first appearance of alien
; levelline, type, x, y, hits, alienscript
; levelline = 984 - pixelline
script_level2
	levelalien 81,ALN_ASTEROID,42,0,1,scr_aster1shot
	levelalien 151,ALN_ASTEROID,0,0,1,scr_aster2shot
	levelalien 219,ALN_ASTEROID,42,0,1,scr_aster1shot
	levelalien 274,ALN_ASTEROID,42,3,1,scr_aster1shot
	levelalien 324,ALN_SMILE,47,8,4,scr_smile
	levelalien 364,ALN_SMILE,47,8,4,scr_smile
	levelalien 404,ALN_SMILE,47,8,4,scr_smile
	levelalien 444,ALN_SMILE,47,8,4,scr_smile
	levelalien 626,ALN_ASTEROID,0,0,1,scr_aster2shot
	levelalien 717,ALN_ASTEROID,0,0,1,scr_aster2shot
	levelalien 768,ALN_ASTEROID,42,0,1,scr_aster1shot
	levelalien 814,ALN_ASTEROID,42,0,1,scr_aster1shot
	levelalien 993,ALN_SMALL,47,8,255,scr_boss2.1
	levelalien 997,ALN_SMALL,47,8,255,scr_boss2.2
	levelalien 1001,ALN_SMALL,47,8,255,scr_boss2.3
	levelalien 1005,ALN_SMALL,47,8,255,scr_boss2.4
	levelalien 1017,ALN_SMALL2,47,8,255,scr_boss2.5
	levelalien 1021,ALN_SMALL2,47,8,255,scr_boss2.6
	levelalien 1025,ALN_SMALL2,47,8,255,scr_boss2.7
	levelalien 1029,ALN_SMALL2,47,8,255,scr_boss2.8
	levelalien 1040,ALN_SMALL3,47,8,255,scr_boss2.9
	levelalien 1044,ALN_SMALL3,47,8,255,scr_boss2.10
	levelalien 1048,ALN_SMALL3,47,8,255,scr_boss2.11
	levelalien 1052,ALN_SMALL3,47,8,255,scr_boss2.12
	.word	$ffff

; --- leveldata level 3 -------------------------------------------------------

; NOTE: This level looks like I ran out of ideas. It's true. I tried to
; compensate for this lack of ideas by making it extremly hard to survive.

bm_level3
	.include "level3.i"

; levelscripts: first appearance of alien
; levelline, type, x, y, hits, alienscript
; levelline = 984 - pixelline
script_level3
	levelalien 1,ALN_BLOCK,0,0,255,scr_blockl.b
	levelalien 1,ALN_BLOCK,16,0,255,scr_blockl.b
	levelalien 1,ALN_BLOCK,24,0,255,scr_blockl.1
	levelalien 1,ALN_BLOCK,32,0,255,scr_blockl.2
	levelalien 33,ALN_DOUBLE,6,0,15,scr_triple.d
	levelalien 37,ALN_SINGLE,0,0,10,scr_triple.s
	levelalien 37,ALN_SINGLE,12,0,10,scr_triple.s
	levelalien 209,ALN_BLOCK,0,0,255,scr_blockl.b
	levelalien 209,ALN_BLOCK,16,0,255,scr_blockl.b
	levelalien 209,ALN_BLOCK,24,0,255,scr_blockl.1
	levelalien 209,ALN_BLOCK,32,0,255,scr_blockl.2
	levelalien 219,ALN_CANNONL,8,0,1,scr_cannonlfast
	levelalien 229,ALN_CANNONR,35,0,1,scr_cannonrfast
	levelalien 239,ALN_CANNONL,8,0,1,scr_cannonlfast
	levelalien 249,ALN_CANNONR,35,0,1,scr_cannonrfast
	levelalien 259,ALN_CANNONL,8,0,1,scr_cannonlfast
	levelalien 269,ALN_CANNONR,35,0,1,scr_cannonrfast
	levelalien 279,ALN_CANNONL,8,0,1,scr_cannonlfast
	levelalien 289,ALN_CANNONR,35,0,1,scr_cannonrfast
	levelalien 307,ALN_BLOCK,24,0,255,scr_blockr.b
	levelalien 307,ALN_BLOCK,40,0,255,scr_blockr.b
	levelalien 307,ALN_BLOCK,8,0,255,scr_blockr.1
	levelalien 307,ALN_BLOCK,16,0,255,scr_blockr.2
	levelalien 325,ALN_CANNONL,0,0,1,scr_cannonlfast
	levelalien 335,ALN_CANNONR,43,0,1,scr_cannonrfast
	levelalien 345,ALN_CANNONL,0,0,1,scr_cannonlfast
	levelalien 355,ALN_CANNONR,43,0,1,scr_cannonrfast
	levelalien 365,ALN_CANNONL,0,0,1,scr_cannonlfast
	levelalien 375,ALN_CANNONR,43,0,1,scr_cannonrfast
	levelalien 385,ALN_CANNONL,0,0,1,scr_cannonlfast
	levelalien 395,ALN_CANNONR,43,0,1,scr_cannonrfast
	levelalien 405,ALN_CANNONL,0,0,1,scr_cannonlfast
	levelalien 415,ALN_CANNONR,43,0,1,scr_cannonrfast
	levelalien 425,ALN_CANNONL,0,0,1,scr_cannonlfast
	levelalien 435,ALN_CANNONR,43,0,1,scr_cannonrfast
	levelalien 456,ALN_DOUBLE,6,0,15,scr_triple.d
	levelalien 460,ALN_SINGLE,0,0,10,scr_triple.s
	levelalien 460,ALN_SINGLE,12,0,10,scr_triple.s
	levelalien 632,ALN_BLOCK,24,0,255,scr_blockr.b
	levelalien 632,ALN_BLOCK,40,0,255,scr_blockr.b
	levelalien 632,ALN_BLOCK,8,0,255,scr_blockr.1
	levelalien 632,ALN_BLOCK,16,0,255,scr_blockr.2
	levelalien 642,ALN_CANNONL,8,0,1,scr_cannonlfast
	levelalien 652,ALN_CANNONR,35,0,1,scr_cannonrfast
	levelalien 662,ALN_CANNONL,8,0,1,scr_cannonlfast
	levelalien 672,ALN_CANNONR,35,0,1,scr_cannonrfast
	levelalien 682,ALN_CANNONL,8,0,1,scr_cannonlfast
	levelalien 692,ALN_CANNONR,35,0,1,scr_cannonrfast
	levelalien 710,ALN_BLOCK,0,0,255,scr_blockl.b
	levelalien 710,ALN_BLOCK,16,0,255,scr_blockl.b
	levelalien 710,ALN_BLOCK,24,0,255,scr_blockl.1
	levelalien 710,ALN_BLOCK,32,0,255,scr_blockl.2
	levelalien 760,ALN_BLOCK,0,0,255,scr_blockl.b
	levelalien 760,ALN_BLOCK,16,0,255,scr_blockl.b
	levelalien 760,ALN_BLOCK,24,0,255,scr_blockl.1
	levelalien 760,ALN_BLOCK,32,0,255,scr_blockl.2
	levelalien 792,ALN_DOUBLE,6,0,15,scr_triple.d
	levelalien 796,ALN_SINGLE,0,0,10,scr_triple.s
	levelalien 796,ALN_SINGLE,12,0,10,scr_triple.s
	levelalien 984,ALN_CANNONDL,0,4,1,scr_cannondlslow
	levelalien 984,ALN_CANNONDR,45,4,1,scr_cannondrslow
	levelalien 989,ALN_CANNONU,7,4,1,scr_cannonu.1
	levelalien 989,ALN_CANNONU,13,4,1,scr_cannonu.2
	levelalien 989,ALN_CANNONU,19,4,1,scr_cannonu.3
	levelalien 989,ALN_CANNONU,25,4,1,scr_cannonu.3
	levelalien 989,ALN_CANNONU,31,4,1,scr_cannonu.2
	levelalien 989,ALN_CANNONU,37,4,1,scr_cannonu.1
	.word	$ffff

; --- alien movement scripts for aliens ---------------------------------------

; NOTE:
; The order of this script is just random, I tried to group similar scripts
; together, but this does not always make sense. Some of the script (especially
; the shots) may be identical, but the semantics are different making the code
; easier to maintain.

; dont move, shootable background
scr_null
	callf	cmd_automove
	callf	alien_endscript
	br	scr_null

; moving opening, border blocks
scr_blockl.b
	do_spd	$10
	do_wait	24
scr_blockr.b
	do_spd	$f0
	do_wait	24
	br	scr_blockl.b

; moving opening, first block
scr_blockl.1
	do_wait	8
	add_st	0,#ALNOFF_X,1
	sub_st	@r1,#16,@r1
	do_wait	16
scr_blockr.1
	do_wait	16
	add_st	0,#ALNOFF_X,1
	add_st	@r1,#16,@r1
	do_wait	8
	br	scr_blockl.1

; moving opening, second block
scr_blockl.2
	do_wait	16
	add_st	0,#ALNOFF_X,1
	sub_st	@r1,#16,@r1
	do_wait	8
scr_blockr.2
	do_wait	8
	add_st	0,#ALNOFF_X,1
	add_st	@r1,#16,@r1
	do_wait	16
	br	scr_blockl.2

; cannon on left side, just shooting
scr_cannonl
	do_fork	ALN_SHOT,5,2,255,scr_cannonlshot
	do_wait	16
	br	scr_cannonl

; cannon on left side, shooting fast
scr_cannonlfast
	do_fork	ALN_SHOT,5,2,255,scr_cannonlshotfast
	do_wait	11
	br	scr_cannonlfast

; cannon on right side, just shooting
scr_cannonr
	do_fork	ALN_SHOT,0,2,255,scr_cannonrshot
	do_wait	16
	br	scr_cannonr

; cannon on right side, shooting fast
scr_cannonrfast
	do_fork	ALN_SHOT,0,2,255,scr_cannonrshotfast
	do_wait	11
	br	scr_cannonrfast

; cannon on upper side
scr_cannonu
	do_fork	ALN_SHOT,1,6,255,scr_cannonushot
	do_wait	16
	br	scr_cannonu

; 8 upper cannons fireing one after the other
scr_cannonu.1
	do_wait	6
scr_cannonu.2
	do_wait	6
scr_cannonu.3
	do_fork	ALN_SHOT,1,6,255,scr_cannonushotfast
	do_wait	10
	jmp	scr_cannonu.1

; cannon on left side, shoot diagonal
scr_cannondl
	do_fork	ALN_SHOT,3,4,255,scr_cannondlshot
	do_wait	16
	br	scr_cannondl

; cannon on left side, shoot diagonal with slower repeat rate
scr_cannondlslow
	do_fork	ALN_SHOT,3,4,255,scr_cannondlshot
	do_wait	22
	br	scr_cannondlslow

; cannon on right side, shoot diagonal
scr_cannondr
	do_fork	ALN_SHOT,-2,4,255,scr_cannondrshot
	do_wait	16
	br	scr_cannondr

; cannon on right side, shoot diagonal with slower repeat rate
scr_cannondrslow
	do_fork	ALN_SHOT,-2,4,255,scr_cannondrshot
	do_wait	22
	br	scr_cannondrslow

; cannon on top side, shoot diagonal left
scr_cannondul
	do_fork	ALN_SHOT,4,4,255,scr_cannondlshot
	do_wait	16
	br	scr_cannondul

; cannon on top side, shoot diagonal right
scr_cannondur
	do_fork	ALN_SHOT,-2,4,255,scr_cannondrshot
	do_wait	16
	br	scr_cannondur

; bomb thrower left side
scr_bomberl
	do_wait	19
	do_fork	ALN_BOMBERLF,0,1,1,scr_bomberlf	; compensate movement
	do_remove

; bomb thrower left side, firing
scr_bomberlf
	do_wait	5
	do_fork	ALN_BOMB,5,1,255,scr_bombl	; compensate movement
	do_wait	255
	do_remove

; bomb thrower right side
scr_bomberr
	do_wait	19
	do_fork	ALN_BOMBERRF,0,1,1,scr_bomberrf	; compensate movement
	do_remove

; bomb thrower right side, firing
scr_bomberrf
	do_wait	5
	do_fork	ALN_BOMB,0,1,255,scr_bombr	; compensate movement
	do_wait	255
	do_remove

; asteroid, fall down, 2 different scripts to avoid framesync when at same y
scr_asteroid2
	do_spd	$0f
	callf	cmd_automove
	callf	alien_endscript
scr_asteroid1
	do_spd	$01
.loop	callf	cmd_automove
	callf	alien_endscript
	br	.loop

; explosion, fixed on moving background
scr_explosion1
	do_wait	15
	do_remove

; three small aliens massive attack
scr_alien1
	do_spd	$ff
	do_wait	14
scr_alien2
	do_spd	$ff
	do_wait	14
scr_alien3
	do_spd	$ff
	do_wait	3
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
	do_wait	3
	do_spd	$1f
	do_wait	3
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
	do_wait		3
	do_spd	$ff
	do_wait	3
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
	do_wait	3
	do_spd	$1f
	do_wait	3
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
	do_wait	3
	do_spd	$ff
	do_wait	3
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
	do_wait	3
	do_spd	$1f
	do_wait	3
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
	do_wait	3
	do_spd	$ff
	do_wait	3
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
	do_wait	3
	do_spd	$1f
	do_wait	3
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
	do_wait	3
	do_spd	$ff
	do_wait	3
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
	do_wait	3
	do_spd	$1f
	do_wait	3
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
	do_wait	3
	do_spd	$ff
	do_wait	3
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
	do_wait	3
	do_spd	$1f
	do_wait	3
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
	do_wait	3
	do_spd	$ff
	do_wait	3
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
	do_wait	3
	do_spd	$1f
	do_wait	3
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
	do_wait	255
	do_remove

; tentacle aliens
scr_tentacle
	do_spd	$1f
	do_wait	8
	do_fork	ALN_BOMB2,2,5,255,scr_tentacleshot
	do_wait	24
	do_fork	ALN_BOMB2,2,5,255,scr_tentacleshot
	do_wait	255
	do_remove

; small2 alien 
scr_small2
	do_spd	$ff
	do_wait	5
	do_fork	ALN_SHOT,2,9,255,scr_cannonushot
	do_wait	8
	do_fork	ALN_SHOT,2,9,255,scr_cannonushot
	do_wait	8
	do_fork	ALN_SHOT,2,9,255,scr_cannonushot
	do_wait	8
	do_fork	ALN_SHOT,2,9,255,scr_cannonushot
	do_wait	8
	do_fork	ALN_SHOT,2,9,255,scr_cannonushot
	do_wait	255
	do_remove

; diagonal fighter from right side in end game
scr_diafighterr
	do_spd	$f1
	do_fork	ALN_SHOT,-2,5,255,scr_dia1shot
	do_wait	16
	do_fork	ALN_SHOT,-2,5,255,scr_dia1shot
	do_wait	11
	do_type	ALN_DIA4
	do_spd	$ff
	do_fork	ALN_SHOT,-2,-2,255,scr_dia4shot
	do_wait	16
	do_type	ALN_DIA3
	do_spd	$1f
	do_fork	ALN_SHOT,5,-2,255,scr_dia3shot
	do_wait	11
	do_type	ALN_DIA2
	do_spd	$11
	do_fork	ALN_SHOT,5,5,255,scr_dia2shot
	do_wait	16
	do_fork	ALN_SHOT,5,5,255,scr_dia2shot
	do_wait	11
	do_type	ALN_DIA4
	;TURNPOINT
	do_spd	$ff
	do_fork	ALN_SHOT,-2,-2,255,scr_dia4shot
	do_wait	16
	do_fork	ALN_SHOT,-2,-2,255,scr_dia4shot
	do_wait	11
	do_type	ALN_DIA1
	do_spd	$f1
	do_fork	ALN_SHOT,-2,5,255,scr_dia1shot
	do_wait	11
	do_type	ALN_DIA2
	do_spd	$11
	do_fork	ALN_SHOT,5,5,255,scr_dia2shot
	do_wait	16
	do_type	ALN_DIA3
	do_spd	$1f
	do_fork	ALN_SHOT,5,-2,255,scr_dia3shot
	do_wait	16
	do_fork	ALN_SHOT,5,-2,255,scr_dia3shot
	do_wait	11
	do_type	ALN_DIA1
	jmpf	scr_diafighterr

; diagonal fighter from left side in end game
scr_diafighterl
	do_spd	$11
	do_fork	ALN_SHOT,5,5,255,scr_dia2shot
	do_wait	16
	do_fork	ALN_SHOT,5,5,255,scr_dia2shot
	do_wait	11
	do_type	ALN_DIA3
	do_spd	$1f
	do_fork	ALN_SHOT,5,-2,255,scr_dia3shot
	do_wait	16
	do_type	ALN_DIA4
	do_spd	$ff
	do_fork	ALN_SHOT,-2,-2,255,scr_dia4shot
	do_wait	11
	do_type	ALN_DIA1
	do_spd	$f1
	do_fork	ALN_SHOT,-2,5,255,scr_dia1shot
	do_wait	16
	do_fork	ALN_SHOT,-2,5,255,scr_dia1shot
	do_wait	11
	do_type	ALN_DIA3
	;TURNPOINT
	do_spd	$1f
	do_fork	ALN_SHOT,5,-2,255,scr_dia3shot
	do_wait	16
	do_fork	ALN_SHOT,5,-2,255,scr_dia3shot
	do_wait	11
	do_type	ALN_DIA2
	do_spd	$11
	do_fork	ALN_SHOT,5,5,255,scr_dia2shot
	do_wait	11
	do_type	ALN_DIA1
	do_spd	$f1
	do_fork	ALN_SHOT,-2,5,255,scr_dia1shot
	do_wait	16
	do_type	ALN_DIA4
	do_spd	$ff
	do_fork	ALN_SHOT,-2,-2,255,scr_dia4shot
	do_wait	16
	do_fork	ALN_SHOT,-2,-2,255,scr_dia4shot
	do_wait	11
	do_type	ALN_DIA2
	jmpf	scr_diafighterl

; mother ship 1
scr_mother1
	do_spd	$1f
	do_wait	20
	do_spd	$0f
	do_wait	5
	do_fork	ALN_TINY,2,5,2,scr_tiny1
	do_wait	10
	do_fork	ALN_TINY,2,5,2,scr_tiny2
	do_wait	10
	do_fork	ALN_TINY,2,5,2,scr_tiny3
	do_wait	10
	do_fork	ALN_TINY,2,5,2,scr_tiny4
	do_wait	10
	do_fork	ALN_TINY,2,5,2,scr_tiny5
	do_wait	10
	do_fork	ALN_TINY,2,5,2,scr_tiny6
	do_wait	25
	do_spd	$1f
	do_wait	255
	do_remove

; the daughter ships
scr_tiny1
	do_spd	$00
	do_wait	5
	do_spd	$ff
	do_wait	20
	do_spd	$0f
	do_wait	34
	jmpf	scr_tinyfire

scr_tiny2
	do_spd	$00
	do_wait	5
	do_spd	$1f
	do_wait	20
	do_spd	$0f
	do_wait	24
	jmpf	scr_tinyfire

scr_tiny3
	do_spd	$00
	do_wait	5
	do_spd	$ff
	do_wait	12
	do_spd	$0f
	do_wait	22
	jmpf	scr_tinyfire

scr_tiny4
	do_spd	$00
	do_wait	5
	do_spd	$1f
	do_wait	12
	do_spd	$0f
	do_wait	12
	jmpf	scr_tinyfire

scr_tiny5
	do_spd	$00
	do_wait	5
	do_spd	$ff
	do_wait	4
	do_spd	$0f
	do_wait	10
	jmpf	scr_tinyfire

scr_tiny6
	do_spd	$00
	do_wait	5
	do_spd	$1f
	do_wait	4
	do_spd	$0f
scr_tinyfire
	do_fork	ALN_SHOT,1,2,255,scr_cannonushot
	do_wait	10
	do_spd	$00
	do_wait	4
	do_spd	$0f
	do_fork	ALN_SHOT,1,2,255,scr_cannonushot
	do_wait	10
	do_spd	$00
	do_wait	4
	do_spd	$0f
	do_fork	ALN_SHOT,1,2,255,scr_cannonushot
	do_wait	10
	do_spd	$00
	do_wait	4
	do_spd	$0f
	do_fork	ALN_SHOT,1,2,255,scr_cannonushot
	do_wait	10
	do_spd	$00
	do_wait	4
	do_spd	$0f
	do_fork	ALN_SHOT,1,2,255,scr_cannonushot
	do_wait	10
	do_spd	$00
	do_wait	255
	do_remove

; smiling alien
scr_smile
	do_spd	$ff
	do_wait 15
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
	do_wait 8
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
	do_wait 8
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
	do_wait 8
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
	do_wait 8
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
	do_spd	$30
	do_wait 4
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
	do_wait 5
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
	do_wait 4
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
	do_spd	$d0
	do_wait 4
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
	do_wait 5
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
	do_wait 4
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
	do_spd	$30
	do_wait 255
	do_remove

; level 1 boss in end game back side
scr_boss1b
	do_spd	$01
	do_wait	8
.loop	do_spd	$10
	do_wait	33
	do_spd	$f0
	do_wait	33
	jmp	.loop

; level 1 boss in end game middle part
scr_boss1m
	do_spd	$01
	do_wait	12
.loop	do_spd	$10
	do_wait	33
	do_spd	$f0
	do_wait	33
	jmpf	.loop

; level 1 boss in end game left cannon
scr_boss1l
	do_spd	$01
	do_wait	12
.loop	do_spd	$10
	do_fork	ALN_SHOT,1,8,255,scr_fastdownshot
	do_wait	11
	do_fork	ALN_SHOT,1,8,255,scr_fastdownshot
	do_wait	11
	do_fork	ALN_SHOT,1,8,255,scr_fastdownshot
	do_wait	11
	do_spd	$f0
	do_fork	ALN_SHOT,1,8,255,scr_fastdownshot
	do_wait	11
	do_fork	ALN_SHOT,1,8,255,scr_fastdownshot
	do_wait	11
	do_fork	ALN_SHOT,1,8,255,scr_fastdownshot
	do_wait	11
	jmpf	.loop

; level 1 boss in end game right cannon
scr_boss1r
	do_spd	$01
	do_wait	12
.loop	do_spd	$10
	do_fork	ALN_SHOT,5,8,255,scr_fastdownshot
	do_wait	11
	do_fork	ALN_SHOT,5,8,255,scr_fastdownshot
	do_wait	11
	do_fork	ALN_SHOT,5,8,255,scr_fastdownshot
	do_wait	11
	do_spd	$f0
	do_fork	ALN_SHOT,5,8,255,scr_fastdownshot
	do_wait	11
	do_fork	ALN_SHOT,5,8,255,scr_fastdownshot
	do_wait	11
	do_fork	ALN_SHOT,5,8,255,scr_fastdownshot
	do_wait	11
	jmpf	.loop

; level 1 boss in end game double cannon
scr_boss1f
	do_spd	$01
	do_wait	19
.loop	do_spd	$10
	do_fork	ALN_SHOT,7,6,255,scr_fastdlshot
	do_fork	ALN_SHOT,-2,6,255,scr_fastdrshot
	do_wait	11
	do_fork	ALN_SHOT,7,6,255,scr_fastdlshot
	do_fork	ALN_SHOT,-2,6,255,scr_fastdrshot
	do_wait	11
	do_fork	ALN_SHOT,7,6,255,scr_fastdlshot
	do_fork	ALN_SHOT,-2,6,255,scr_fastdrshot
	do_wait	11
	do_spd	$f0
	do_fork	ALN_SHOT,7,6,255,scr_fastdlshot
	do_fork	ALN_SHOT,-2,6,255,scr_fastdrshot
	do_wait	11
	do_fork	ALN_SHOT,7,6,255,scr_fastdlshot
	do_fork	ALN_SHOT,-2,6,255,scr_fastdrshot
	do_wait	11
	do_fork	ALN_SHOT,7,6,255,scr_fastdlshot
	do_fork	ALN_SHOT,-2,6,255,scr_fastdrshot
	do_wait	11
	jmpf	.loop

; level 2 boss scripts: invaders style attack
; NOTE: The hit counter is set to 255 by the levelscript, so you can't kill
; them until all of them are on the screen. The hit counter is then set to
; lower values depending on the type of alien.
scr_boss2.1
	do_spd	$e0
	do_wait	4
scr_boss2.2
	do_spd	$e0
	do_wait	4
scr_boss2.3
	do_spd	$e0
	do_wait	4
scr_boss2.4
	do_spd	$e0
	do_wait	2
	do_spd	$f0
	do_wait	1
	jmpf	scr_boss2.bot

scr_boss2.5
	do_spd	$e0
	do_wait	4
scr_boss2.6
	do_spd	$e0
	do_wait	4
scr_boss2.7
	do_spd	$e0
	do_wait	4
scr_boss2.8
	do_spd	$e0
	do_wait	2
	do_spd	$f0
	do_wait	1
	jmpf	scr_boss2.mid

scr_boss2.9
	do_spd	$e0
	do_wait	4
scr_boss2.10
	do_spd	$e0
	do_wait	4
scr_boss2.11
	do_spd	$e0
	do_wait	4
scr_boss2.12
	do_spd	$e0
	do_wait	2
	do_spd	$f0
	do_wait	1
	jmp	scr_boss2.top

scr_boss2.bot
	do_spd	$01		; down, to make room for next line
	do_wait 9
	do_spd	0		; stop until next line moved in
	do_wait	15

scr_boss2.mid
	do_spd	$01		; down, to make room for next line
	do_wait 8
	do_spd	0		; stop until next line moved in
	do_wait	15

scr_boss2.top
	; make alien killable now
	add_st	0,#ALNOFF_HITS,1	; r1 points to number of hits
	ld	@r0			; alien type
	be	#ALN_SMALL,.small
	be	#ALN_SMALL2,.small2
	be	#ALN_SMALL3,.small3
	br	.start
.small	mov	#1,@r1
	br	.start
.small2	mov	#3,@r1
	br	.start
.small3	mov	#5,@r1
.start	do_spd	$e0
.loop	; left/right movement invaders loop
	add_st	0,#ALNOFF_WAIT,1
	mov	#9,@r1
.wait1	callf	cmd_automove
	callf	scr_invadershot		; shoot if necessary
	callf	alien_endscript
	add_st	0,#ALNOFF_WAIT,1
	dbnz	@r1,.wait1
	; invert speed value
	add_st	0,#ALNOFF_V,1
	xor	acc
	sub	@r1
	st	@r1
	br	.loop

; test for alignment of current alien with ship and shoot randomly
scr_invadershot
	add_st	0,#ALNOFF_X,1
	ld	shipx
	sub	@r1
	be	#0,.aligned
	be	#1,.aligned
	be	#2,.aligned
	; not aligned: shoot with propability 1/32
	xor	acc
	callf	randombit
	rolc
	callf	randombit
	rolc
	callf	randombit
	rolc
	callf	randombit
	rolc
	callf	randombit
	rolc
	be	#$15,.shoot		; see NOTE at randombit
	ret
.aligned
	; shoot with propability 1/4, 1/8, 1/16 depending on alien
	xor	acc
	callf	randombit
	rolc
	callf	randombit
	rolc
	callf	randombit
	rolc
	callf	randombit
	rolc
	st	c
	ld	@r0
	be	#ALN_SMALL,.probe
	clr1	c,3			; new propability value: 0-7
	be	#ALN_SMALL2,.probe
	set1	c,2			; new propability value: 0-3
.probe	ld	c
	be	#$05,.shoot		; see NOTE at randombit
	ret
.shoot	; do the shooting
	.ifd	DEBUG			; test for free alien in DEBUG mode,
	push	0			; the shootings are very aggressive
	callf	alien_getfree		; and we will use all alienobject
	pop	0
	bp	psw,cy,.ret
	.endc
	do_fork	ALN_SHOT,2,6,255,scr_cannonushot
.ret	ret

; 2 ALN_SINGLE and 1 ALN_DOUBLE form a triple
scr_triple.d
	do_wait	4
scr_triple.s
	do_run	8,scr_tripleshot
	do_spd	$10
	do_run	8,scr_tripleshot
	do_spd	$1f
	do_run	12,scr_tripleshot
	do_spd	$1e
	do_run	8,scr_tripleshot
	do_spd	$ff
	do_run	28,scr_tripleshot
	do_spd	0
	; next rotation
	do_spd	$10
	do_run	8,scr_tripleshot
	do_spd	$1f
	do_run	12,scr_tripleshot
	do_spd	$1e
	do_run	8,scr_tripleshot
	do_spd	$ff
	do_run	28,scr_tripleshot
	do_spd	0
	; last rotation
	do_spd	$10
	do_run	8,scr_tripleshot
	do_spd	$1f
	do_run	12,scr_tripleshot
	do_spd	$1e
	do_run	8,scr_tripleshot
	do_spd	$1f
	do_run	255,scr_tripleshot
	do_remove

; fire for triple: shoot with propability 1/32
scr_tripleshot
	xor	acc
	mov	#5,b
.getrnd	callf	randombit
	rolc
	dbnz	b,.getrnd
	be	#$15,.fire		; see NOTE at randombit
	ret
.fire
	.ifd	DEBUG			; test for free alien in DEBUG mode,
	push	0			; the shootings are very aggressive
	callf	alien_getfree		; and we could use all alienobjects
	pop	0
	bp	psw,cy,.ret
	.endc
	ld	@r0
	be	#ALN_SINGLE,.single
	be	#ALN_DOUBLE,.double
	.ifd	DEBUG
	mov	#$41,acc
	jmpf	panic
	.endc
.ret	ret
.single
	do_fork	ALN_SHOT,3,7,255,scr_cannonushot
	ret
.double
	do_fork	ALN_SHOT,-2,8,255,scr_cannondrshot
	do_fork	ALN_SHOT,8,8,255,scr_cannondlshot
	ret

; --- alien movement scripts for forked off shots -----------------------------

; cannon left shot
scr_cannonlshot
	do_spd	$10
	do_wait	$ff
	do_remove

; cannon left shot fast
scr_cannonlshotfast
	do_spd	$20
	do_wait	$ff
	do_remove

; cannon right shot
scr_cannonrshot
	do_spd	$f0
	do_wait	$ff
	do_remove

; cannon right shot fast
scr_cannonrshotfast
	do_spd	$e0
	do_wait	$ff
	do_remove

; cannon up shot
scr_cannonushot
	do_spd	$01
	do_wait	$ff
	do_remove

; cannon up shot fast
scr_cannonushotfast
	do_spd	$02
	do_wait	$ff
	do_remove

; cannon diagonal left shot
scr_cannondlshot
	do_spd	$11
	do_wait	$ff
	do_remove

; cannon diagonal right shot
scr_cannondrshot
	do_spd	$f1
	do_wait	$ff
	do_remove

; asteroid down from right
scr_aster1shot
	do_spd	$e2
	do_wait	255
	do_remove

; asteroid down from left
scr_aster2shot
	do_spd	$22
	do_wait	255
	do_remove

; diagonal fighter 1 shot
scr_dia1shot
	do_spd	$e2
	do_wait	255
	do_remove

; diagonal fighter 2 shot
scr_dia2shot
	do_spd	$22
	do_wait	255
	do_remove

; diagonal fighter 3 shot
scr_dia3shot
	do_spd	$2e
	do_wait	255
	do_remove

; diagonal fighter 4 shot
scr_dia4shot
	do_spd	$ee
	do_wait	255
	do_remove

; bomb thrown from left side
scr_bombl
	do_spd	$2e
	do_wait	7
	do_spd	$1e
	do_wait	1
	do_fork	ALN_SHOT,0,4,255,scr_shrapnell1
	do_fork	ALN_SHOT,4,4,255,scr_shrapnell2
	do_fork	ALN_SHOT,4,0,255,scr_shrapnell3
	do_fork	ALN_SHOT,4,-4,255,scr_shrapnell4
	do_fork	ALN_SHOT,0,-4,255,scr_shrapnell5
	do_fork	ALN_SHOT,-4,-4,255,scr_shrapnell6
	do_fork	ALN_SHOT,-4,0,255,scr_shrapnell7
	do_fork	ALN_SHOT,-4,4,255,scr_shrapnell8
	do_remove

; bomb thrown from right side
scr_bombr
	do_spd	$ee
	do_wait	7
	do_spd	$fe
	do_wait	1
	do_fork	ALN_SHOT,0,4,255,scr_shrapnell1
	do_fork	ALN_SHOT,4,4,255,scr_shrapnell2
	do_fork	ALN_SHOT,4,0,255,scr_shrapnell3
	do_fork	ALN_SHOT,4,-4,255,scr_shrapnell4
	do_fork	ALN_SHOT,0,-4,255,scr_shrapnell5
	do_fork	ALN_SHOT,-4,-4,255,scr_shrapnell6
	do_fork	ALN_SHOT,-4,0,255,scr_shrapnell7
	do_fork	ALN_SHOT,-4,4,255,scr_shrapnell8
	do_remove

; bomb thrown from tentacle alien
scr_tentacleshot
	do_spd	$00
	mov	#<tune_bomber,trl
	mov	#>tune_bomber,trh
	callf	snd_frameplaytune
	do_wait	8
	do_fork	ALN_SHOT,-4,4,255,scr_shrapnell8
	do_fork	ALN_SHOT,0,4,255,scr_shrapnell1
	do_fork	ALN_SHOT,4,4,255,scr_shrapnell2
	do_remove

; bomb shrapnells
scr_shrapnell1
	do_spd	$02
	do_wait	255
	do_remove
scr_shrapnell2
	do_spd	$22
	do_wait	255
	do_remove
scr_shrapnell3
	do_spd	$20
	do_wait	255
	do_remove
scr_shrapnell4
	do_spd	$2e
	do_wait	255
	do_remove
scr_shrapnell5
	do_spd	$0e
	do_wait	255
	do_remove
scr_shrapnell6
	do_spd	$ee
	do_wait	255
	do_remove
scr_shrapnell7
	do_spd	$e0
	do_wait	255
	do_remove
scr_shrapnell8
	do_spd	$e2
	do_wait	255
	do_remove

; fast shot down
scr_fastdownshot
	do_spd	$02
	do_wait	$ff
	do_remove

; fast shot diagonal left
scr_fastdlshot
	do_spd	$22
	do_wait	$ff
	do_remove

; fast shot diagonal right
scr_fastdrshot
	do_spd	$e2
	do_wait	$ff
	do_remove

; --- several bitmaps ---------------------------------------------------------

; the title screen
; NOTE:
; This was converted with the ppmtovmbitmap.pl script you can get at
; http://soeren.infoserv.de/vm/tools.html
bm_title
	.include "title.i"

; --- several \0 terminated ASCII strings -------------------------------------

str_level	.byte	"LEVEL",0
str_ships	.byte	"SHIPS",0
str_sc		.byte	"SC",0
str_score	.byte	"SCORE",0
str_pressa	.byte	"PRESS@A",0
str_over1	.byte	"GAME",0
str_over2	.byte	"OVER",0
str_title	.byte	"TITLE",0
str_panic	.byte	"PANIC",0
str_newhigh	.byte	"NEW@HI=",0
str_enter	.byte	"ENTER",0
str_name	.byte	"NAME",0

str_sound	.byte	"SOUND",0
str_soundon	.byte	"IS@ON",0
str_soundoff	.byte	"IS@OFF",0

str_calibrate1	.byte	"CALI=",0
str_calibrate2	.byte	"BRATING",0
str_calibrate3	.byte	"DELAY",0
str_calibrate4	.byte	"LOOP;;;",0

menu_main	.byte	"START",0,"OPTIONS",0,"HISCORE",0,"DEMO",0,"DEBUG",0
menu_options	.byte	"SOUND",0
menu_demo	.byte	"LEVEL@1",0,"LEVEL@2",0,"LEVEL@3",0,"LEVEL@4",0
		.ifd	DEBUG
menu_debug	.byte	"LEVEL",0,"TUNE",0
		.endc

; --- sound tunes for RC mode -------------------------------------------------

; ship shoots
tune_shot
	.byte	$01,$10,$88,$02,$00,0

; alien is hit
tune_hit
	.byte	$01,$50,$a8,$02,$00,0

; alien explodes
tune_kill
	.byte	$01,$40,$a0,$02,$01
	.byte	$01,$60,$b0,$02,$01
	.byte	$01,$80,$c0,$02,$02
	.byte	0

; ship explodes
tune_explode
	.byte	$01,$60,$b0,$02,$04
	.byte	$01,$40,$a0,$02,$04
	.byte	$01,$20,$90,$02,$08
	.byte	0

; bomb thrown from tentacle alien
tune_bomber
	.byte	$01,$18,$8c,$02,$01
	.byte	$01,$10,$88,$02,$01
	.byte	$01,$08,$84,$02,$01
	.byte	$01,$00,$80,$02,$01
	.byte	0

tune_test
	.byte	0

; --- sound tunes for sub clock mode ------------------------------------------

; key click
tune_keyclick
	.byte $01,$fc,$fe,$02,$01
	.byte 0

; level starts
tune_level
        .byte $01,$f3,$f9,$02,$02
        .byte $01,$00,$00,$02,$01
        .byte $01,$f3,$f9,$02,$02
        .byte $01,$f0,$f8,$02,$02
        .byte $01,$f3,$f9,$02,$08
        .byte 0

; game over (same as in minesweeper)
tune_over
	.byte $01,$f3,$f9,$02,$02
	.byte $01,$f0,$f8,$02,$02
	.byte $01,$ec,$f6,$02,$08
	.byte 0

; new highscore (same as in minesweeper)
tune_highscore
	.byte $01,$ec,$f6,$02,$04
	.byte $01,$ee,$f7,$02,$04
	.byte $01,$f0,$f8,$02,$08
	.byte $01,$f1,$f8,$02,$04
	.byte $01,$f3,$f9,$02,$08
	.byte $01,$f0,$f8,$02,$04
	.byte $01,$ec,$f6,$02,$08
	.byte $01,$f0,$f8,$02,$04
	.byte $01,$f3,$f9,$02,$04
	.byte $01,$f6,$fb,$02,$10
	.byte 0

; --- highscore table ---------------------------------------------------------

	.cnop	0,$80		; pad to next 128 byte boundary
highscores
	.byte	$00,$01,$00,"SG",0
	.byte	$00,$00,$80,"SG",0
	.byte	$00,$00,$60,"SG",0
	.byte	$00,$00,$40,"SG",0
	.byte	$00,$00,$20,"SG",0

; --- binary version tag ------------------------------------------------------

	.byte	13,10,"$Id: alienfighter.s,v 1.93 2000/12/14 21:13:59 sgust Exp $",13,10,0

	.cnop	0,$200		; pad to an even number of blocks
end				; used in header as pointer to end of code
